// ;-*-C++-*- *  Time-stamp: "2008-10-16 03:49:27 hmmr"

/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-09-02
 *
 * libCN testing grounds evolving into a model runner
 */


#include <iostream>
#include <fstream>
#include <vector>
#include <unordered_set>
#include <boost/regex.hpp>
#include <cfloat>
//#include <cwchar>

#include "cnlib-model.h"
#include "cnlib-integrate-rk65.h"

using namespace std;


static CModel *model;

static const char    *nml_fname;


struct SCNRunnerOptions {
	double	running_time;
	bool	listen_all_neurons:1,
		listen_all_synapses:1,
		dump_params:1,
		list_units:1,
		sort_units:1,
		log_dt:1;
	const char *parm_file,
		   *spike_cnt_stats_file,
		   *single_neuron;
	double	dt_max,
		dt_min,
		dt_max_fac;
	int	verbosely;
};

static SCNRunnerOptions runner_options = {
	100,
	false, false,
	false,
	false,
	false,
	false,
	"CNModelUnitParameters",
	NULL,
	NULL,
	.1, 1e-6, 5.,
	1
};

static vector<string> listeners;
static vector<string> tuned_params;
static vector<string> spiking_neurons;




static int process_param_regex_triplet( string& triplet);
static int prepare_model();

#define CNRUN_PARSE_HELP_REQUEST	-1
#define CNRUN_PARSE_ERROR		-2
#define CNRUN_DUMP_PARAMS		-3

static int parse_options( int argc, char **argv);

static void dump_model_units();
static void report_spike_count_stats();


static list<CNeuron*>::iterator N;
static list<CSynapse*>::iterator Y;




static int
prepare_model()
{
	model->verbosely = runner_options.verbosely;

	int retv = model->import_NetworkML( nml_fname);
	if ( retv < 0 ) {
		cout << "CModel->import_NetworkML failed returning " << retv << endl;
		return 2;
	}

	if ( runner_options.dump_params ) {
		dump_model_units();
		return 1;
	}

	if ( runner_options.single_neuron )
		for_model_neurons (model,N)
			if ( strcmp( runner_options.single_neuron, (*N)->label) == 0 ) {
				(*N)->enable( true);
				(*N)->den.clear();
			} else
				(*N)->enable( false);

	vector<string>::iterator P;

      // set up listeners
	if ( runner_options.listen_all_neurons )
		for_model_neurons (model, N)
			(*N) -> start_listening();
	if ( runner_options.listen_all_synapses )
		for_model_synapses (model, Y)
			(*Y) -> start_listening();
      // individually
	for ( P = listeners.begin(); P != listeners.end(); P++ ) {
		boost::regex pattern( (*P).data());
		boost::cmatch found;
		for_model_neurons (model, N) {
			if ( boost::regex_match( (*N)->label, found, pattern) )
				(*N) -> start_listening();
		}
		for_model_synapses (model, Y)
			if ( boost::regex_match( (*Y)->label, found, pattern) )
				(*Y) -> start_listening();
	}

      // modify named params of named units
	ifstream parm_stream( runner_options.parm_file);
	if ( parm_stream.good() ) {
		if ( runner_options.verbosely > 3 )
			cout << " (from " << runner_options.parm_file << ")\n";
		string buf;
		while ( getline( parm_stream, buf, '\n') || parm_stream.gcount() )
			if ( buf.size() && buf[0] != '#' )
				process_param_regex_triplet( buf);
	}

	if ( tuned_params.size() )
		if ( runner_options.verbosely > 3 )
			cout << " (from cmdline)\n";
	for ( P = tuned_params.begin(); P != tuned_params.end(); P++ )
		process_param_regex_triplet( (*P));

      // set up spike recorders
	for ( P = spiking_neurons.begin(); P != spiking_neurons.end(); P++ ) {
		boost::regex pattern( (*P).data());
		boost::cmatch found;
		for_model_neurons (model,N) {
			(*N)->set_spike_recording( false);
			if ( boost::regex_match( (*N)->label, found, pattern) ) {
				(*N)->set_spike_recording( true);
				if ( runner_options.verbosely > 1 )
					cout << "(enabling spike accounting for neuron \"" << (*N)->label << "\")\n";
			}
		}
	}

	return 0;
}





static void
report_spike_count_stats()
{
	if ( strcmp( runner_options.spike_cnt_stats_file, "-") ) {
		ofstream report_stream( runner_options.spike_cnt_stats_file);
		report_stream <<
			"#Label\tTimes Spiked\n";

		for_model_neurons (model,N)
			if ( (*N)->status() & CN_RECORDINGSPIKES )
				report_stream << (*N)->label << '\t' << (*N)->spike_cnt << endl;
	} else {
		cout <<
			"------- Spiking stats:\n"
			"Label\tTimes Spiked\n";

		for_model_neurons (model,N)
			if ( (*N)->status() & CN_RECORDINGSPIKES )
				printf( "%-20s\t%d\n", (*N)->label, (*N)->spike_cnt);
	}
}









int
main( int argc, char *argv[])
{
	cout << "CNModel runner, compiled " << __TIME__ << " " << __DATE__ << endl << endl;

	switch ( parse_options( argc, argv) ) {
	case CNRUN_PARSE_ERROR:
		cerr << "Problem parsing command line or sanitising values\n";
	case CNRUN_PARSE_HELP_REQUEST:
		cout << "Usage: " << argv[0] << "\n" <<
			" -f filename\t\t(default ./CNModel.nml)\n"
			" -t running_time\t(default 100ms)\n"
			" -1 label\t\tSimulate a single neuron\n"
			"\n"
			" -l regex\t\tMake so labelled unit(s) listen\n"
			" -N \t\t\tEnable listening for all neurons\n"
			" -S \t\t\t\tsame, for all synapses\n"
			"\n"
			" -K filename\t\tReport spike count stats in filename (\"-\" = cout)\n"
			" -o regex\t\t\tinclude only so labelled neurons\n"
			"\n"
			" -p regex:parm_no:value\tSet matching unit(s)' parameter\n"
			" -P filename\t\t\tread these from filename, default ./CNModelUnitParameters\n"
			"\n"
			" -D \t\t\tDump all unit types in the model\n"
			" -U \t\t\tList available unit types\n"
			" \t\t\twith param names and values\n"
			"\n"
			" -T \t\t\tlog dt\n"
			"\n"
			" -i dt_max\t\t(default 0.1)\n"
			" -j dt_min\t\t(default 1e-6)\n"
			" -J dt_max_fac\t\tCap dt by current value x this, default 5\n"
			"\n"
			" -v level\t\tVerbosity\n"
			" -s \t\t\tSort units\n"
			" -h \t\t\tDisplay this help\n"
			"\n";
		return -1;
	}

	if ( runner_options.list_units ) {
		cnmodel_dump_available_units();
		return 0;
	}


	if ( !(model = new CModel( "nameme",
				    new CIntegrateRK65( runner_options.dt_max,
							runner_options.dt_min,
							runner_options.dt_max_fac),
				    0 |
				    (runner_options.log_dt ?CN_MDL_LOGDT :0) |
				    (runner_options.sort_units ?CN_MDL_SORTUNITS :0))) ) {
		cerr << "Failed to create a CModel\n";
		return -2;
	}

	if ( !prepare_model() )
		model->time_step( runner_options.running_time);

	if ( runner_options.spike_cnt_stats_file &&
	     model->cycle() )
		report_spike_count_stats();

	delete model;

	return 0;
}






static int
parse_options( int argc, char **argv)
{
	int	c;
	char*	tail;

	while ( (c = getopt( argc, argv, "f:t:1:NSK:o:l:p:P:i:j:J:DUTsv:h")) != -1 ) {
		switch ( c ) {
		case 'f':
			if ( nml_fname ) {
				cerr << "Please try one nml file at a time\n";
				return CNRUN_PARSE_ERROR;
			} else
				nml_fname = optarg;
		    break;

		case 't':
			runner_options.running_time = strtod( optarg, NULL);
			if ( runner_options.running_time <= 0. ) {
				cerr << "Expecting running time to be a non-negative double\n";
				return CNRUN_PARSE_ERROR;
			}

		    break;

		case '1':	runner_options.single_neuron = optarg;			break;

		case 'i':	runner_options.dt_max = strtod( optarg, NULL);		break;
		case 'j':	runner_options.dt_min = strtod( optarg, NULL);		break;
		case 'J':	runner_options.dt_max_fac = strtod( optarg, NULL);	break;

		case 'N':	runner_options.listen_all_neurons = true;		break;
		case 'S':	runner_options.listen_all_synapses = true;		break;
		case 'l':	listeners.push_back( "");
				listeners.back() = optarg;				break;

		case 'K':	runner_options.spike_cnt_stats_file = optarg;		break;
		case 'o':	spiking_neurons.push_back( "");
				spiking_neurons.back() = optarg;			break;

		case 'p':	tuned_params.push_back( "");
				tuned_params.back() = optarg;		break;

		case 'P':	runner_options.parm_file = optarg;	break;

		case 'D':	runner_options.dump_params = true;	break;
		case 'U':	runner_options.list_units = true;	break;

		case 'T':	runner_options.log_dt = true;		break;

		case 's':	runner_options.sort_units = true;	break;

		case 'v':	runner_options.verbosely = strtol( optarg, NULL, 10);	break;
		case 'h':
			return CNRUN_PARSE_HELP_REQUEST;
		case '?':
		default:
			return CNRUN_PARSE_ERROR;
		}
	}

	if ( !nml_fname )
		nml_fname = "CNModel.nml";

	if ( runner_options.dt_max < DBL_MIN ) {
		cerr << "Expecting a sensible dt_max\n";
		return CNRUN_PARSE_ERROR;
	}
	if ( runner_options.dt_min < DBL_MIN || runner_options.dt_min >= runner_options.dt_max ) {
		cerr << "That's funky\n";
		return CNRUN_PARSE_ERROR;
	}

	return 0;
}





static void
dump_model_units()
{
	cout << "\nUnits in the model:\n";

	unordered_set<int> found_unit_types;
	size_t p = 0;

	cout << "\n===== Neurons:\n";
	for_model_neurons (model, N)
		if ( found_unit_types.count( (*N)->type()) == 0 ) {
			found_unit_types.insert( (*N)->type());

			cout << "--- " << (*N)->unit_class() << endl;
			for ( p = 0; p < (*N)->p_no(); p++ ) {
				printf( "%2d: %-5s\t= % 10g\t // %s\n",
					p, (*N)->param_sym(p),
					(*N)->param_value(p),
					(*N)->param_name(p));
			}

		}
	cout << "\n===== Synapses:\n";
	for_model_synapses (model,Y)
		if ( found_unit_types.count( (*Y)->type()) == 0 ) {
			found_unit_types.insert( (*Y)->type());

			cout << "--- " << (*Y)->unit_class() << endl;
			for ( p = 0; p < (*Y)->p_no(); p++ ) {
				printf( "%2d: %-5s\t= % 10g\t // %s\n",
					p, (*Y)->param_sym(p),
					(*Y)->param_value(p),
					(*Y)->param_name(p));
			}
		}
	cout << endl;
}









static int
process_param_regex_triplet( string& triplet)
{
	char	*label = (char*)triplet.data(),
		*param = (char*)&triplet[triplet.find_first_of(":")],
		*value = (char*)&triplet[triplet.find_last_of (":")];
	if ( label == param || param == value ) {
		cerr << "Incomplete label:param:value set `" << triplet << "'\n";
		return -1;
	}

	bool	assigned = false;

	*(param++) = *(value++) = NULL;
	int	param_param = strtol( param, NULL, 10);
	double	value_value = strtod( value, NULL);

	boost::regex pattern( label);
	boost::cmatch found;

	for_model_neurons (model,N)
		if ( boost::regex_match( (*N)->label, found, pattern) ) {
			if ( param_param >= (*N)->p_no() )
				cerr << "Neuron " << (*N)->label << " (class \""
				     << (*N)->unit_class() << "\")  has only "
				     << (*N)->p_no() << " parameter(s)\n";
			else {
				if ( runner_options.verbosely > 2 )
					cout << " set [" << __CN_DescriptorTable[(*N)->type()].stock_param_syms[param_param]
					     << "] of \"" << (*N)->label << "\" to\t" << value_value << endl;
				(*N)->param_value(param_param) = value_value;
				assigned  = true;
			}
		}
	for_model_synapses (model,Y)
		if ( boost::regex_match( (*Y)->label, found, pattern) )
			if ( param_param >= (*Y)->p_no() )
				cerr << "Synapse " << (*Y)->label << " (class \""
				     << (*Y)->unit_class() << "\") has only "
				     << (*Y)->p_no() << " parameter(s)\n";
			else {
				if ( runner_options.verbosely > 2 )
					cout << " set [" << __CN_DescriptorTable[(*Y)->type()].stock_param_syms[param_param]
					     << "] of \"" << (*Y)->label << "\" to\t" << value_value << endl;
				(*Y)->param_value(param_param) = value_value;
				assigned  = true;
			}

	if ( !assigned )
		cerr << "No unit labelled matching " << label << endl;
}




// EOF
