// ;-*-C++-*- *  Time-stamp: "2008-10-27 12:43:13 hmmr"

/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-09-02
 *
 * NeuroML reading methods for CModel, using libxml2
 */

#include <string>
#include <iostream>
#include <boost/regex.hpp>

#include "cnlib-model.h"

#include "cnlib-neuron.h"
#include "cnlib-oscillator.h"
#include "cnlib-reader-tape.h"
#include "cnlib-synapse.h"


using namespace std;

#ifdef LIBXML_READER_ENABLED



static xmlNode*
find_named_root_child_elem( xmlNode *node,     // node to start search from
			    const char *elem)  // name of the element searched for
{
	xmlNode *n, *ni;

	for ( n = node->children; n; n = n->next ) {
		if ( n->type == XML_ELEMENT_NODE ) {
			if ( xmlStrEqual( n->name, BAD_CAST elem) )
				return n;
// the <populations> and <projections> nodes are expected to appear as
// direct children of the root node; so don't go search deeper than that

//			if ( n->children ) { // go search deeper
//				ni = find_named_elem( n->children, elem);
//				if ( ni )
//					return ni;
//			}
		}
	}
	return NULL;
}



int
CModel::import_NetworkML( const char *fname, bool appending)
{
	int retval = 0;

	if ( !appending )
		reset();

	LIBXML_TEST_VERSION;
	xmlDoc *doc = xmlReadFile( fname, NULL, 0);
	if ( !doc )
		return CN_NMLIN_NOFILE;
	// else our nml file is parsed and passed to us

	// we pass up on validation (for which we might need to keep a
	// .dtd or Schema at hand), and proceed to extracting elements

	xmlNode *root_node = xmlDocGetRootElement( doc),
		*n;

      // read meta:notes and make out a name for the model
	if ( !root_node ) {
		cerr << "Failed to obtain root element\n";
		retval = CN_NMLIN_NOELEM;
		goto out;
	}
	if ( !(n = find_named_root_child_elem( root_node, "notes")) )
		cerr << "<notes> element not found\n";
		// this is not critical, so just keep the user
		// informed and proceed
	else
		if ( n->type == XML_ELEMENT_NODE ) {  // only concern ourselves with nodes of this type
			xmlChar *notes_s = xmlNodeGetContent( n);
			// look for a substring specific to neuroConstruct, which is obviously speculative
			boost::regex pattern( ".*project: (\\w*).*");
			boost::cmatch found;
			name = boost::regex_match( (char*)notes_s, found, pattern)
				? name = found[1]
				: "(unmatched)";
			xmlFree( notes_s);  // cleaning up in our one-shot program may be overscrupulous,
					    // but let's be disciplined nonetheless
		} else
			name = "(unnamed)";

	cout << "Model " << name << ": " << (appending ?"Appending" :"Importing") << " NetworkML from " << fname << endl;

	// In the following calls to _process_{populations,instances}
	// functions, the actual order of appearance of these nodes in
	// the xml file doesn't matter, thanks to the xml contents
	// being already wholly read and availab;e to us as a tree.

      // process <populations>
	if ( !(n = find_named_root_child_elem( root_node, "populations")) ) {
		retval = CN_NMLIN_NOELEM;
		goto out;
	} // assume there is only one <populations> element: don't loop to catch more
	retval = _process_populations( n->children);	// note n->children, which is in fact a pointer to the
							// first child
	if ( retval < 0 )
		goto out;

      // process <projections>
	if ( !(n = find_named_root_child_elem( root_node, "projections")) ) {
		retval = CN_NMLIN_NOELEM;
		goto out;
	}
	retval = _process_projections( n->children);
	if ( retval < 0 )
		goto out;

	// we are done with topology; now put units' variables on a vector
	finalize_additions();
	// can call time_step only after finalize_additions
out:
	xmlFreeDoc( doc);

	cout << endl;

	return retval;
}





int
CModel::_process_populations( xmlNode *n)
{
	TUnitType cell_type;
	int t;
	xmlChar *group_id_s, *cell_type_s;

	int	pop_cnt = 0;

	for ( ; n; n = n->next ) // if is NULL (parent had no children), we won't do a single loop
		if ( n->type == XML_ELEMENT_NODE && xmlStrEqual( n->name, BAD_CAST "population") ) {

			group_id_s = xmlGetProp( n, BAD_CAST "name"); // BAD_CAST is just a cast to xmlChar*
								      // with a catch that libxml functions
								      // expect strings pointed to to be good UTF
			if ( !group_id_s ) {
				cerr << "<population> element missing a \"name\" attribute near line " << n->line << endl;
				return CN_NMLIN_BADATTR;
			}
		      // probably having an unnamed popuation isn't an error so serious as to abort the
		      // operation, but discipline is above all

			cell_type_s = xmlGetProp( n, BAD_CAST "cell_type");
			// now we know the type of cells included in this population; remember it to pass on to
			// _process_population_instances, where it is used to select an appropriate unit type
			// when actually adding a neuron to the model

		      // but well, let's check if we have units of that tyoe in stock
			t = NT_VOID;
			for ( t = NT_FIRST; t <= NT_LAST; t++ )
				if ( xmlStrEqual( cell_type_s, BAD_CAST __CN_DescriptorTable[t].label) ) {
					cell_type = static_cast<TUnitType>(t);
					break;
				}
			if ( t == NT_VOID ) {
				cerr << "Bad cell type (" << (char*)cell_type_s << ") in population " << group_id_s << endl;
				xmlFree( cell_type_s), xmlFree( group_id_s);
				return CN_NMLIN_BADCELLTYPE;
			}

			xmlNode *nin = n->children;  // again, ->children means ->first
			if ( nin )
				for ( ; nin; nin = nin->next )  // deal with multiple <instances> nodes
					if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "instances") ) {
						int subretval = _process_population_instances( nin->children,
											       group_id_s, cell_type);
						if ( subretval < 0 ) {
							xmlFree( group_id_s);
							return subretval;
						}
						if ( verbosely > 1 )
							cout.width( 3), cout << subretval << " instance(s) of type \"" << cell_type_s
							      << "\" in population " << group_id_s << endl;
						pop_cnt++;
					}
			xmlFree( cell_type_s);
			xmlFree( group_id_s);
		}

	if ( verbosely > 1 )
		cout << "\tTotal " << pop_cnt << " population(s)\n";
	return pop_cnt;
}



int
CModel::_process_projections( xmlNode *n)
{
	// much the same code as in _process_populations

	xmlChar *prj_name_s, *prj_src_s, *prj_tgt_s;

	int pop_cnt = 0;

	for ( ; n; n = n->next )
		if ( n->type == XML_ELEMENT_NODE && xmlStrEqual( n->name, BAD_CAST "projection") ) {

			prj_name_s = xmlGetProp( n, BAD_CAST "name");
			if ( !prj_name_s ) {
				cerr << "<projection> element missing a \"name\" attribute near line " << n->line << endl;
				return CN_NMLIN_BADATTR;
			}

			prj_src_s  = xmlGetProp( n, BAD_CAST "source");
			prj_tgt_s  = xmlGetProp( n, BAD_CAST "target");
			if ( !prj_src_s || !prj_tgt_s ) {
				cerr << "Projection \"" << prj_name_s << "\" missing a \"source\" and/or \"target\" attribute near line " << n->line << endl;
				xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
				return CN_NMLIN_BADATTR;
			}

			xmlNode *nin;
			nin = n->children;
			if ( !nin )
				cerr << "Empty <projection> node near line " << n->line << endl;

			xmlChar *synapse_type_s;
			TUnitType synapse_type;
			int t;

			for ( ; nin; nin = nin->next )
				if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "synapse_props") ) {
					synapse_type_s = xmlGetProp( nin, BAD_CAST "synapse_type");
					t = NT_VOID;
					for ( t == YT_FIRST; t <= YT_LAST; t++ )
						if ( xmlStrEqual( synapse_type_s, BAD_CAST __CN_DescriptorTable[t].label) ) {
							synapse_type = static_cast<TUnitType>(t);
							break;
						}
					if ( t == NT_VOID ) {
						cerr << "Bad synapse type (" << (char*)synapse_type_s << ") near line " << nin->line << endl;
						xmlFree( synapse_type_s), xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
						return CN_NMLIN_BADCELLTYPE;
					}
				}

			nin = n->children;
			if ( nin )
				for ( ; nin; nin = nin->next )
					if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "connections") ) {
						int subretval = _process_projection_connections( nin->children,
												 prj_name_s, synapse_type,
												 prj_src_s, prj_tgt_s);
						if ( subretval < 0 ) {
							xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
							return subretval;
						}
						if ( verbosely > 1 )
							cout.width(3), cout << subretval << " connection(s) of type \""
									    << synapse_type_s << "\" in projection " << prj_name_s << endl;
						pop_cnt++;
					}
			xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
		}

	if ( verbosely > 1 )
		cout << "\tTotal " << pop_cnt << " projection(s)\n";
	return pop_cnt;
}







int
CModel::_process_population_instances( xmlNode *n, const xmlChar *group_prefix, TUnitType type)
{
	int	retval = 0;  // also keeps a count of added neurons

	double	x, y, z;
	char	cell_id[CN_LABEL_SIZE];

	xmlNode *nin;

	for ( ; n; n = n->next )
		if ( n->type == XML_ELEMENT_NODE && xmlStrEqual( n->name, BAD_CAST "instance") ) {

			xmlChar *id_s = xmlGetProp( n, BAD_CAST "id");
			if ( !id_s ) {
				// could be less strict here and allow empty ids, which will then be composed
				// from group_prefix + id (say, "LN0", "LN1" and so on); but then, as
				// individual <projection>s would have to reference both endpoints by explicit
				// ids, it is obviously prone to error to have <instance> ids depend solely on
				// their order of appearance.
				// So we bark at empty ids.
				cerr << "<instance> element without an \"id\" attribute near line " << n->line << endl;
				retval = CN_NMLIN_BADATTR;
				goto out;
			}

			size_t total_len = xmlStrlen( group_prefix) + xmlStrlen( id_s);
			if ( total_len >= CN_LABEL_SIZE ) {
				cerr << "<instance> combined label (" << group_prefix << id_s << ") exceeding " << CN_LABEL_SIZE << " near line " << n->line << endl;
				xmlFree( id_s);
				retval = CN_NMLIN_BIGLABEL;
				goto out;
			}
			snprintf( cell_id, CN_LABEL_SIZE-1, "%s%s", group_prefix, id_s);  // here, a new instance is given a name

			xmlFree( id_s);

			if ( (nin = n->children) )
				for ( ; nin; nin = nin->next )
					if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "location") ) {
						xmlChar *x_s = xmlGetProp( nin, BAD_CAST "x"),
							*y_s = xmlGetProp( nin, BAD_CAST "y"),
							*z_s = xmlGetProp( nin, BAD_CAST "z");
					      // here we do actually insert neurons into the model
						if ( x_s && y_s && z_s ) {
							x = strtod( (char*)x_s, NULL), y = strtod( (char*)y_s, NULL), z = strtod( (char*)z_s, NULL);
							int	addition_retval = -1;
							switch ( type ) {
							case NT_HH:
								addition_retval = add_neuron( new CNeuronHH( cell_id, x, y, z), false);
								// false for the last arg above is to delay a call to finalize_additions
							    break;
							case NT_TM:
								addition_retval = add_neuron( new CNeuronTM( cell_id, x, y, z), false);
							    break;
							case NT_EC:
								addition_retval = add_neuron( new CNeuronEC( cell_id, x, y, z), false);
							    break;
							case NT_ECA:
								addition_retval = add_neuron( new CNeuronECA( cell_id, x, y, z), false);
							    break;
							case NT_POISSON:
								addition_retval = add_neuron( new COscillatorPoisson( cell_id, x, y, z), false);
							    break;
							case NT_LV:
								addition_retval = add_neuron( new COscillatorLV( cell_id, x, y, z), false);
							    break;
							case NT_COLPITTS:
								addition_retval = add_neuron( new COscillatorColpitts( cell_id, x, y, z), false);
							    break;
							case NT_VDPOL:
								addition_retval = add_neuron( new COscillatorVdPol( cell_id, x, y, z), false);
							    break;
							case NT_READER:
								string tape_fname( cell_id);
								tape_fname += ".in";
								addition_retval = add_neuron( new CNReader( cell_id, tape_fname.data(), 0.), false);
							    break;
						      // don't provide a default case as the validity of `type'
						      // has been checked by _process_populations
							}

							if ( addition_retval ) {
								cerr << "Failed to add a neuron " << cell_id << endl;
								retval = CN_NMLIN_STRUCTERROR;
							} else
								retval++;

						} else {
							cerr << "<location> element missing full coordinates near line " << nin->line << endl;
							retval = CN_NMLIN_BADATTR;
						}
						xmlFree( x_s), xmlFree( y_s), xmlFree( z_s);

						if ( retval < 0 )
							goto out;
					}
		}
out:
	return retval;
}




int
CModel::_process_projection_connections( xmlNode *n,
					 const xmlChar *synapse_name, TUnitType type,
					 const xmlChar *src_grp_prefix, const xmlChar *tgt_grp_prefix)
{
	// similar to _process_population_instances, except that we read some more attributes (source and
	// target units)

	int	retval = 0;  // is also a counter of synapses
	char	synapse_id [CN_LABEL_SIZE],
		src_cell_id[CN_LABEL_SIZE],
		tgt_cell_id[CN_LABEL_SIZE];

	CNeuron *src, *tgt;

	for ( ; n; n = n->next )
		if ( n->type == XML_ELEMENT_NODE && xmlStrEqual( n->name, BAD_CAST "connection") ) {

			xmlChar *synapse_id_s  = xmlGetProp( n, BAD_CAST "id"),
				*src_cell_id_s = xmlGetProp( n, BAD_CAST "pre_cell_id"),
				*tgt_cell_id_s = xmlGetProp( n, BAD_CAST "post_cell_id");
			if ( !synapse_id_s || !src_cell_id_s || !tgt_cell_id_s ) {
				cerr << "A <connection> element without \"id\", \"pre_cell_id\" and/or \"post_cell_id\" attribute near line " << n->line << endl;
				xmlFree( synapse_id_s), xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s);
				return CN_NMLIN_BADATTR;
			}

			memset( src_cell_id, 0, CN_LABEL_SIZE);
			snprintf( src_cell_id, CN_LABEL_SIZE-1, "%s%s", src_grp_prefix, src_cell_id_s);
			src = neuron_by_label( src_cell_id);

			memset( tgt_cell_id, 0, CN_LABEL_SIZE);
			snprintf( tgt_cell_id, CN_LABEL_SIZE-1, "%s%s", tgt_grp_prefix, tgt_cell_id_s);
			tgt = neuron_by_label( tgt_cell_id);

			if ( xmlStrlen( synapse_name) + xmlStrlen( synapse_id_s) >= CN_LABEL_SIZE ) {
				cerr << "Synapse combined label (" << synapse_name << synapse_id_s << ") exceeding " << CN_LABEL_SIZE << " near line " << n->line << endl;
				xmlFree( synapse_id_s), xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s);
				return CN_NMLIN_BIGLABEL;
			}
			memset( synapse_id, 0, CN_LABEL_SIZE);
			snprintf( synapse_id, CN_LABEL_SIZE-1, "%s%s", synapse_name, synapse_id_s);

			if ( !src || !tgt ) {
				cerr << "Phoney source (" << src_cell_id << ") or target (" << tgt_cell_id << ") for connection near line " << n->line << endl;
				xmlFree( synapse_id_s), xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s);
				return CN_NMLIN_BADATTR;
			}

			int addition_retval = -1;
			switch ( type ) {
			case YT_AB:
				addition_retval = add_synapse( new CSynapseAb( synapse_id, src, tgt), false);
			      break;
			case YT_OSCILTUNER:
				addition_retval = add_synapse( new CSynapseOscilTuner( synapse_id, src, tgt), false);
			      break;
			}
			if ( addition_retval ) {
				cerr << "Failed to add a synapse " << synapse_id << endl;
				retval = CN_NMLIN_STRUCTERROR;
			} else
				retval++;

			xmlFree( synapse_id_s), xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s);
		}

	return retval;
}




#else
#error Need an XMLREADER-enabled libxml2 (>2.6)

#endif // LIBXML_READER_ENABLED
// EOF
