// ;-*-C++-*- *  Time-stamp: "2008-10-16 03:40:57 hmmr"

/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-09-02
 *
 * Class CModel
 */

/*--------------------------------------------------------------------------

The wrapper class which takes lists of pointers to neurons and synapses
which are networked to a neural system and assembles a common state
vector and handles the derivatives. At the same time it serves the neurons
and synapses their state at any given time and allows them to adjust their
parameters.

--------------------------------------------------------------------------*/


#ifndef CN_MODEL_H
#define CN_MODEL_H

#include <list>
#include <vector>
#include <string>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "cnlib-unit.h"
#include "cnlib-integrate-base.h"

using namespace std;


#define CNM_SPIKETHRESH 0.
#define CNM_SPIKELAPSE  5.


// CModel _status bits
#define CN_MDL_LOGDT		(1 << 0)
#define CN_MDL_SORTUNITS	(1 << 2)
#define CN_MDL_ALLOWNOISE	(1 << 3)

class CModel {
    friend class CUnit;	 // accesses V
    friend class CNeuron;
    friend class COscillatorPoisson;
    friend class COscillatorColpitts;
    friend class COscillatorVdPol;
    friend class COscillatorLV;
    friend class CSynapseAb;

	int		_status;

	unsigned long	_cycle;

	vector<double> *V,	// contains catenated var vectors of all constituent neurons and synapses
		       *W;	// V and W alternate in the capacity of the main vector, so avoiding many a
				// memcpy (need to handle them by pointers to allow for easy swapping of what
				// they point to)

	size_t	_var_cnt;	// total # of variables (to be) allocated in V an W, plus one for model_time
	size_t	_unit_cnt;	// unit count

	list<CNeuron*>	neu_list;
	list<CSynapse*>	syn_list;

    public:
	string	name;

      // ctor, dtor
	CModel( const char *inname, CIntegrate_base *inRK65Setup, int instatus);
       ~CModel();

      // total reset of the model
	void reset();

      // this reverts all of V to initial stock values, preserving structure
	void reset_all_units();

	CNeuron  *neuron_by_label( const char *);
	CSynapse *synapse_by_label( const char *);

      // add a unit; if is_last == true, do allocations immediately
      // otherwise defer until addition is done with is_last == true
      // or the user calls finalize_additions
	int add_neuron ( CNeuron  *n, bool is_last = true);  // returns -1 on error
	int add_synapse( CSynapse *y, bool is_last = true);

	void finalize_additions();

	size_t units()			{ return _unit_cnt;	  }
	size_t vars()			{ return _var_cnt;	  }

	unsigned long cycle()		{ return _cycle;          }
	const double& model_time()	{ return (*V)[0];         }
	const double &dt()		{ return _integrator->dt; }

      // go as many cycles as necessary to reach a time dist from now
	unsigned int time_step( double dist, double *cpu_time_p = NULL);


	double	spike_thresh, // above which neurons will detect a spike
		spike_lapse;  // if occurs less than this after n->_last_spike_end

      // NeuroMl interface
	int import_NetworkML( const char *fname, bool appending = false);
//	int export_NetworkML( const char *fname);

      // various convenience fields and members
	list<CNeuron* >::iterator	nlist_begin()	{ return neu_list.begin(); }
	list<CNeuron* >::iterator	nlist_end()	{ return neu_list.end();   }
	list<CSynapse*>::iterator	ylist_begin()	{ return syn_list.begin(); }
	list<CSynapse*>::iterator	ylist_end()	{ return syn_list.end();   }

	int	verbosely;

    private:
    friend class CIntegrate_base;
    friend class CIntegrateRK65;
	CIntegrate_base *_integrator;

	void derivative( vector<double>&, vector<double>&);

	int _process_populations( xmlNode*);
	int _process_population_instances( xmlNode*,
					   const xmlChar *group_prefix,
					   TUnitType);

	int _process_projections( xmlNode*);
	int _process_projection_connections( xmlNode*,
					     const xmlChar *name, TUnitType,
					     const xmlChar *src_grp_prefix, const xmlChar *tgt_grp_prefix);
	ofstream	*_dt_logger;
};


// by popular demand
#define for_all_neurons(iter)   for ( iter = neu_list.begin(); iter != neu_list.end(); iter++ )
#define for_all_synapses(iter)  for ( iter = syn_list.begin(); iter != syn_list.end(); iter++ )

#define for_model_neurons(modelp,iter)   for ( iter = modelp->nlist_begin(); iter != modelp->nlist_end(); iter++ )
#define for_model_synapses(modelp,iter)  for ( iter = modelp->ylist_begin(); iter != modelp->ylist_end(); iter++ )



// return values for import_NetworkML
#define CN_NMLIN_NOFILE		-1
#define CN_NMLIN_NOELEM		-2
#define CN_NMLIN_BADATTR	-3
#define CN_NMLIN_BADCELLTYPE	-4
#define CN_NMLIN_BIGLABEL	-5
#define CN_NMLIN_STRUCTERROR	-6






// various CUnit & CNeuron methods accessing CModel members
// that we want to have inline

inline const double&
CUnit::model_time()
{
	return _parent_model->model_time();
}

inline void
CUnit::reset_state()
{
	memcpy( &(*_parent_model->V)[idx],
		__CN_DescriptorTable[_type].stock_var_values,
		__CN_DescriptorTable[_type].vno * sizeof(double));
}


inline double&
CNeuron::E()
{
	return (*_parent_model->V)[idx+0];
}

inline bool
CNeuron::is_spiking()
{
	return E() >= _parent_model->spike_thresh;
}

inline void
CNeuron::spike_detect()
{
	if ( E() >= _parent_model->spike_thresh ) {
		if ( model_time() - t_last_spike_end > _parent_model->spike_lapse ) {
			t_last_spike_start = model_time();
			spike_cnt++;
		}
		t_last_spike_end = model_time();
	}
}


inline double
CNeuron::Isyn()
{
	return Isyn( *_parent_model->V);
}

#endif

// EOF
