// ;-*-C++-*- *  Time-stamp: "2008-10-16 03:38:26 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny (tnowotny@ucsd.edu)
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-10-16
 *
 */


#include "cnlib-neuron.h"
#include <cmath>



const double __CN_Params_NeuronHH[] = {
	120.0,	// gNa: Na conductance in 1/(mOhms * cm^2)
	 55.0,	// ENa: Na equi potential in mV
	 36.0,	// gK: K conductance in 1/(mOhms * cm^2)
	-72.0,	// EK: K equi potential in mV
	  0.3,	// gl: leak conductance in 1/(mOhms * cm^2)
	-50.0,	// El: leak equi potential in mV
	  1.0,	// Cmem: membr. capacity density in muF/cm^2
	  0.    // Externally applied constant current
};

const char* const __CN_ParamNames_NeuronHH[] = {
	"Na conductance, 1/(m\316\251 * cm\302\262)",
	"Na equi potential, mV",
	"K conductance, 1/(m\316\251 * cm\302\262)",
	"K equi potential in mV",
	"Leak conductance, 1/(m\316\251 * cm\302\262)",
	"Leak equi potential, mV",
	"Membrane capacity density in \302\265F/cm\302\262",
	"Externally applied constant current",
};
const char* const __CN_ParamSyms_NeuronHH[] = {
	"gNa",
	"ENa",
	"gK",
	"EK",
	"gl",
	"El",
	"Cmem",
	"Idc",
};



const double __CN_Vars_NeuronHH[] = {
	-60.0,		// 0 - membrane potential E
	  0.0529324,	// 1 - prob. for Na channel activation m
	  0.3176767,	// 2 - prob. for not Na channel blocking h
	  0.5961207	// 3 - prob. for K channel activation n
};

const char* const __CN_VarNames_NeuronHH[] = {
	"Membrane potential",
	"Prob. of Na channel activation",
	"1-Prob. of Na channel blocking",
	"Prob. of K channel activation",
};
const char* const __CN_VarSyms_NeuronHH[] = {
	"E",
	"m",
	"h",
	"n"
};



inline double ___pow3( double x)
{
	return x * x * x;
}
inline double ___pow4( double x)
{
	return x * x * x * x;
}

void
CNeuronHH::derivative(  vector<double>& x, vector<double>& dx)
{
      // differential eqn for E, the membrane potential
	dE(dx) = -(___pow3(m(x)) * h(x) * P.n.gNa * (E(x) - P.n.ENa) +
		   ___pow4(n(x)) * P.n.gK * (E(x) - P.n.EK)
		   + P.n.gl * (E(x) - P.n.El) - (Isyn(x) + P.n.Idc)) / P.n.Cmem;

	double _a, _b;
      // diferential eqn for m, the probability for Na channel activation
	_a = (3.5 + 0.1 * E(x)) / -expm1( -3.5 - 0.1 * E(x));
	_b = 4.0 * exp( -(E(x) + 60.0) / 18.0);
	dm(dx) = _a * (1.0 - m(x)) - _b * m(x);

      // differential eqn for h, the probability for Na channel inactivation
	_a = 0.07 * exp( -E(x) / 20.0 - 3.0);
	_b = 1.0 / (exp( -3.0 - 0.1 * E(x)) + 1.0);
	dh(dx) = _a * (1.0 - h(x)) -_b * h(x);

      // differential eqn for n, the probability for K channel activation
	_a = (-0.5 - 0.01 * E(x)) / expm1( -5.0 - 0.1 * E(x));
	_b = 0.125 * exp( -(E(x) + 60.0) / 80.0);
	dn(dx) = _a * (1.0 - n(x)) - _b * n(x);
}









const double __CN_Params_NeuronTM[] = {
	7.15,		//  0 - gNa: Na conductance in 1/(mOhms * cm^2)
       50.0,		//  1 - ENa: Na equi potential in mV
	1.43,		//  2 - gK: K conductance in 1/(mOhms * cm^2)
      -95.0,		//  3 - EK: K equi potential in mV
	0.021,		//  4 - gl: leak conductance in 1/(mOhms * cm^2)
      -55.0,		//  5 - El: leak equi potential in mV
	0.143,		//  6 - Cmem: membr. specific capacitance, muF/cm^2
	0.,		//  7 - Idc: constant, externally applied current
	0.00572,	//  8 - gKl: potassium leakage conductivity
      -95.0,		//  9 - EKl: potassium leakage equi pot in mV
       65.0,		// 10 - V0: ~ total equi potential (?)
};
const char* const __CN_ParamNames_NeuronTM[] = {
	"Na conductance, 1/(m\316\251 * cm\302\262)",
	"Na equi potential, mV",
	"K conductance, 1/(m\316\251 * cm\302\262)",
	"K equi potential, mV",
	"Leak conductance, 1/(m\316\251 * cm\302\262)",
	"Leak equi potential, mV",
	"Membrane specific capacitance, \302\265F/cm\302\262",
	"Externally applied constant current",
	"K leakage conductivity",
	"K leakage equi pot, mV",
	"Total equi potential (?)",
};
const char* const __CN_ParamSyms_NeuronTM[] = {
	"gNa",
	"ENa",
	"gK",
	"EK",
	"gl",
	"El",
	"Cmem",
	"Idc",
	"gKl",
	"EKl",
	"V0",
};

const double __CN_Vars_NeuronTM[] = {
// as in a single-neuron run
      -63.0,	// 0 - membrane potential E
	0.0217,	// 1 - prob. for Na channel activation m
	0.993,	// 2 - prob. for not Na channel blocking h
	0.051,	// 3 - prob. for K channel activation n

// previously thought to be resting state values
//      -60.0,		// 0 - membrane potential E
//	0.0529324,	// 1 - prob. for Na channel activation m
//	0.3176767,	// 2 - prob. for not Na channel blocking h
//	0.5961207,	// 3 - prob. for K channel activation n
};
const char* const __CN_VarNames_NeuronTM[] = {
	"Membrane potential",
	"Prob. of Na channel activation",
	"1-Prob. of Na channel blocking",
	"Prob. of K channel activation",
};
const char* const __CN_VarSyms_NeuronTM[] = {
	"E",
	"m",
	"h",
	"n",
};





void
CNeuronTM::derivative( vector<double>& x, vector<double>& dx)
{

      // differential eqn for E, the membrane potential
	dE(dx) = -(___pow3(m(x)) * h(x) * P.n.gNa * (E(x) - P.n.ENa) +
		   ___pow4(n(x)) * P.n.gK * (E(x) - P.n.EK) +
		   P.n.gl * (E(x) - P.n.El) + P.n.gKl * (E(x) - P.n.EKl) - (Isyn(x) + P.n.Idc)) / P.n.Cmem;

	double _a, _b;
      // diferential eqn for m, the probability for one Na channel activation
      // particle
	_a = 0.32 * (13.0 - E(x) - P.n.V0) / expm1( (13.0 - E(x) - P.n.V0) / 4.0);
	_b = 0.28 * (E(x) + P.n.V0 - 40.0) / expm1( (E(x) + P.n.V0 - 40.0) / 5.0);
	dm(dx) = _a * (1.0 - m(x)) - _b * m(x);

      // differential eqn for h, the probability for the Na channel blocking
      // particle to be absent
	_a = 0.128 * exp( (17.0 - E(x) - P.n.V0) / 18.0);
	_b = 4.0 / (exp( (40.0 - E(x) - P.n.V0) / 5.0) + 1.0);
	dh(dx) = _a * (1.0 - h(x)) - _b * h(x);

      // differential eqn for n, the probability for one K channel activation
      // particle
	_a = 0.032 * (15.0 - E(x) - P.n.V0) / (exp( (15.0 - E(x) - P.n.V0) / 5.0) - 1.0);
	_b = 0.5 * exp( (10.0 - E(x) - P.n.V0) / 40.0);
	dn(dx)= _a * (1.0 - n(x)) -_b * n(x);
}









const double __CN_Params_NeuronEC[] = {
	7.15,	//  0 - gNa: Na conductance in 1/(mOhms * cm^2)
       50.0,	//  1 - ENa: Na equi potential in mV
	1.43,	//  2 - gK: K conductance in 1/(mOhms * cm^2)
      -95.0,	//  3 - EK: K equi potential in mV
	0.021,	//  4 - gl: leak conductance in 1/(mOhms * cm^2)
      -55.0,	//  5 - El: leak equi potential in mV
	0.286,	//  6 - Cmem: membr. capacity density in muF/cm^2 // 0.143
	0.,	//  7 - Externally applied constant current
	0.035,	//  8 - gKl: potassium leakage conductivity
      -95.0,	//  9 - EKl: potassium leakage equi pot in mV
       65.0,	// 10 - V0: ~ total equi potential (?)
	0.0185,	// 11 - gh1 // 1.85
	0.01,	// 12 - gh2
      -20.0,	// 13 - Vh
};
const char* const __CN_ParamNames_NeuronEC[] = {
	"Na conductance, 1/(mOhm * cm\302\262)",
	"Na equi potential, mV",
	"K conductance, 1/(mOhm * cm\302\262)",
	"K equi potential, mV",
	"Leak conductance, 1/(mOhm * cm\302\262)",
	"Leak equi potential, mV",
	"Membrane capacity density in \302\265F/cm\302\262",
	"Externally applied constant current",
	"K leakage conductivity",
	"K leakage equi pot in mV",
	"Total equi potential (?)",
	"gh1",
	"gh2",
	"Vh"
};
const char* const __CN_ParamSyms_NeuronEC[] = {
	"gNa",
	"ENa",
	"gK",
	"EK",
	"gl",
	"El",
	"Cmem",
	"Idc",
	"gKl",
	"EKl",
	"V0",
	"gh1",
	"gh2",
	"Vh"
};

const double __CN_Vars_NeuronEC[] = {
      -64.1251,		// 0 - membrane potential E
	0.0176331,	// 1 - prob. for Na channel activation m
	0.994931,	// 2 - prob. for not Na channel blocking h
	0.0433969,	// 3 - prob. for K channel activation n
	0.443961,	// 4 - Ih1 activation
	0.625308	// 5 - Ih2 activation
};
const char* const __CN_VarNames_NeuronEC[] = {
	"Membrane potential",
	"Prob. of Na channel activation",
	"Prob. of not Na channel blocking",
	"Prob. of K channel activation",
	"Ih1 activation",
	"Ih2 activation"
};
const char* const __CN_VarSyms_NeuronEC[] = {
	"E",
	"m",
	"h",
	"n",
	"Ih1",
	"Ih2"
};




#define _xfunc(a,b,k,V)  ((a) * (V) + (b)) / (1.0 - exp(((V)+(b)/(a))/(k)))

void
CNeuronEC::derivative( vector<double>& x, vector<double>& dx)
{
	double _a, _b;
      // differential eqn for E, the membrane potential
	dE(dx) = -(pow( m(x), 3) * h(x) * P.n.gNa * (E(x) - P.n.ENa) +
		 pow( n(x), 4) * P.n.gK * (E(x) - P.n.EK) +
		 (Ih1(x) * P.n.gh1 + Ih2(x) * P.n.gh2) * (E(x) - P.n.Vh)+
		 P.n.gl * (E(x) - P.n.El) + P.n.gKl * (E(x) - P.n.EKl) - Isyn(x)) / P.n.Cmem;

      // diferential eqn for m, the probability for one Na channel activation particle
	_a = 0.32 * (13.0 - E(x) - P.n.V0) / expm1( (13.0 - E(x) - P.n.V0) / 4.0);
	_b = 0.28 * (E(x) + P.n.V0 - 40.0) / expm1( (E(x) + P.n.V0 - 40.0) / 5.0);
	dm(dx) = _a * (1.0 - m(x)) - _b * m(x);

      // differential eqn for h, the probability for the Na channel blocking particle to be absent
	_a = 0.128 * exp( (17.0 - E(x) - P.n.V0) / 18.0);
	_b = 4.0 / (exp( (40.0 - E(x) - P.n.V0) / 5.0) + 1.0);
	dh(dx) = _a * (1.0 - h(x)) - _b * h(x);

      // differential eqn for n, the probability for one K channel activation particle
	_a = 0.032 * (15.0 - E(x) - P.n.V0) / expm1( (15.0 - E(x) - P.n.V0) / 5.0);
	_b = 0.5 * exp( (10.0 - E(x) - P.n.V0) / 40.0);
	dn(dx) = _a * (1.0 - n(x)) - _b * n(x);

      // differential equation for the Ih1 activation variable
	_a = _xfunc (-2.89e-3, -0.445,  24.02, E(x));
	_b = _xfunc ( 2.71e-2, -1.024, -17.40, E(x));
	dIh1(dx) = _a * (1.0 - Ih1(x)) - _b * Ih1(x);

      // differential equation for the Ih2 activation variable
	_a = _xfunc (-3.18e-3, -0.695,  26.72, E(x));
	_b = _xfunc ( 2.16e-2, -1.065, -14.25, E(x));
	dIh2(dx) = _a * (1.0 - Ih2(x)) - _b * Ih2(x);
}

#undef _xfunc












const double __CN_Params_NeuronECA[] = {
	52.0,	//  0 - Na conductance in 1/(mOhms * cm^2)
	55.0,	//  1 - Na equi potential in mV
	11.0,	//  2 - K conductance in 1/(mOhms * cm^2)
       -90.0,	//  3 - K equi potential in mV
	 0.5,	//  4 - Leak conductance in 1/(mOhms * cm^2)
       -65.0,	//  5 - Leak equi potential in mV
	 1.5,	//  6 - Membr. capacity density in muF/cm^2
	 0.,	//  7 - Externally applied constant current
	 0.5,	//  8 - gNap
	 1.5,	//  9 - gh
       -20.0,	// 10 - Vh
};
const char* const __CN_ParamNames_NeuronECA[] = {
	"Na conductance, 1/(m\316\251 * cm\302\262)",
	"Na equi potential, mV",
	"K conductance, 1/(m\316\251 * cm\302\262)",
	"K equi potential, mV",
	"Leak conductance, 1/(m\316\251 * cm\302\262)",
	"Leak equi potential, mV",
	"Membrane capacity density in \302\265F/cm\302\262",
	"Externally applied constant current",
	"gNap",
	"gh",
	"Vh",
};
const char* const __CN_ParamSyms_NeuronECA[] = {
	"gNa",
	"ENa",
	"gK",
	"EK",
	"gl",
	"El",
	"Cmem",
	"Idc",
	"gNap",
	"gh",
	"Vh",
};

const double __CN_Vars_NeuronECA[] = {
      -53.77902178,	// E
	0.0262406368,	// prob. for Na channel activation m
	0.9461831106,	// prob. for not Na channel blocking h
	0.1135915933,	// prob. for K channel activation n
	0.08109646237,	// Nap
	0.06918464221,	// Ih1 activation
	0.09815937825	// Ih2 activation
};
const char* const __CN_VarNames_NeuronECA[] = {
	"Membrane potential",
	"Prob. of Na channel activation",
	"1-Prob. of Na channel blocking",
	"Prob. of K channel activation",
	"mNap",
	"Ih1 activation",
	"Ih2 activation"
};
const char* const __CN_VarSyms_NeuronECA[] = {
	"E",
	"m",
	"h",
	"n",
	"mNap",
	"Ih1",
	"Ih2"
};



void
CNeuronECA::derivative( vector<double>& x, vector<double>& dx)
{
      // differential eqn for E, the membrane potential
	dE(dx) = -((pow( m(x), 3) * h(x) * P.n.gNa + P.n.gNap * mNap(x)) * (E(x) - P.n.ENa) +
		   pow( n(x), 4) * P.n.gK * (E(x) - P.n.EK) +
		   P.n.gh * (Ih1(x) * 0.65 + Ih2(x) * 0.35) * (E(x) - P.n.Vh) +
		   P.n.gl * (E(x) - P.n.El) - (Isyn(x) + P.n.Idc) + 2.85) / P.n.Cmem;

	double _a, _b;
      // diferential eqn for m, the probability for one Na channel activation particle
	_a = -0.1 * (E(x) + 23) / expm1( -0.1 * (E(x) + 23));
	_b =  4.  * exp( -(E(x) + 48) / 18);
	dm(dx) = _a * (1. - m(x)) - _b * m(x);

      // differential eqn for h, the probability for the Na channel blocking particle to be absent
	_a = 0.07 * exp( -(E(x) + 37.0) / 20.0);
	_b = 1. / (exp( -0.1 * (E(x) + 7.)) + 1.0);
	dh(dx) = _a * (1.0 - h(x)) - _b * h(x);

      // differential eqn for n, the probability for one K channel activation particle
	_a = -0.01  * (E(x) + 27) / expm1( -0.1 * (E(x) + 27));
	_b =  0.125 * exp( -(E(x) + 37) / 80);
	dn(dx) = _a * (1.0 - n(x)) - _b * n(x);

	_a = 1. / (0.15 * (1 + exp( -(E(x) + 38) / 6.5)));
	_b = exp( -(E(x) + 38) / 6.5) / (0.15 * (1 + exp( -(E(x) + 38) / 6.5)));
	dmNap(dx) = _a * (1.0 - mNap(x)) - _b * mNap(x);

      // differential equation for the Ihf activation variable
	_a = 1. / (1 + exp( (E(x) + 79.2) / 9.78));
	_b = 0.51 / (exp( (E(x) - 1.7) / 10) + exp( -(E(x) + 340) / 52)) + 1;
	dIh1(dx) = (_a - Ih1(x)) / _b;

      // differential equation for the Ihs activation variable
	_a = 1. / (1 + exp( (E(x) + 71.3) / 7.9));
	_b = 5.6 / (exp( (E(x) - 1.7) / 14) + exp( -(E(x) + 260) / 43)) + 1;
	dIh2(dx) = (_a - Ih2(x)) / _b;
}


// EOF
