// ;-*-C++-*- *  Time-stamp: "2008-10-16 03:45:40 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny (tnowotny@ucsd.edu)
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-08-16
 *
 * Oscillator units
 */


#include "cnlib-oscillator.h"
#include "cnlib-model.h"

const double __CN_Params_OscillatorPoisson[] = {
	2.0,	// spike time of mf_poisson inputneuron
       10.0,	// refractory period + spike time
      -60.0,	// input neuron resting potential
       20.0,	// input neuron potential when firing
	0.02	// firing rate Lambda [1/ms]=[10^3 Hz]
};
const char* const __CN_ParamNames_OscillatorPoisson[] = {
	"Spike time",
	"Spike time + refractory period",
	"Input neuron resting potential",
	"Input neuron potential when firing",
	"Firing rate \316\273",
};
const char* const __CN_ParamSyms_OscillatorPoisson[] = {
	"ts",
	"trts",
	"Vrst",
	"Vfir",
	"lambda",
};

const double __CN_Vars_OscillatorPoisson[] = {
	-60.,	// = Vrst, per initialization code found in ctor
	  0.
};
const char* const __CN_VarNames_OscillatorPoisson[] = {
	"Membrane potential",
	"Time"
};
const char* const __CN_VarSyms_OscillatorPoisson[] = {
	"E",
	"t"
};



void
COscillatorPoisson::possibly_fire()
{
	if ( den.size() > 0 ) {
		P.n.frl = 0.;
		vector<CSynapse*>::iterator Y;
		for ( Y = den.begin(); Y != den.end(); Y++ )
			P.n.frl += (*Y)->source()->E();
//		cerr << " : " << P.n.frl << endl;
	}

	if ( _status & CN_NFIRING ) {
		if ( model_time() - t_firing_started > P.n.tsp ) {
			(_status &= ~CN_NFIRING) |= CN_NREFRACT;
			t_firing_ended = model_time();
		}
	} else {
		if ( _status & CN_NREFRACT ) {
			if ( model_time() - t_firing_started > P.n.tref )
				_status &= ~CN_NREFRACT;
		} else {
			if ( __R_std.n() <= P.n.frl * _parent_model->dt() ) {
				_status |= CN_NFIRING;
				t_firing_started = model_time() + _parent_model->dt();
			}
		}
	}

	E() = next_state_E;
	next_state_E = (_status & CN_NFIRING) ?P.n.Vfir :P.n.Vrst;
}










const double __CN_Params_OscillatorColpitts[] = {
	1.0,	// a
	0.0797,	// g
	0.6898,	// q
	6.2723	// eta
};

const char* const __CN_ParamNames_OscillatorColpitts[] = {
	"a",
	"g",
	"q",
	"\316\267"
};
const char* const __CN_ParamSyms_OscillatorColpitts[] = {
	"a",
	"g",
	"q",
	"eta"
};


const double __CN_Vars_OscillatorColpitts[] = {
	0.02,
	0.69,
       -0.53
};

const char* const __CN_VarNames_OscillatorColpitts[] = {
	"x0",
	"x1",
	"x2"
};
const char* const __CN_VarSyms_OscillatorColpitts[] = {
	"x0",
	"x1",
	"x2"
};


void
COscillatorColpitts::derivative( vector<double>& x, vector<double>& dx)
{
	dx0(dx) =  P.n.a   *  x1(x) + Isyn(x);
	dx1(dx) = -P.n.g   * (x0(x) + x2(x)) - P.n.q * x1(x);
	dx2(dx) =  P.n.eta * (x1(x) + 1.0 - exp( -x0(x)));
//	dx[idx  ] =  p[0] *  x[idx+1] + Isyn;
//	dx[idx+1] = -p[1] * (x[idx  ] + x[idx+2]) - p[2] * x[idx+1];
//	dx[idx+2] =  p[3] * (x[idx+1] + 1.0 - exp(-x[idx]));
}












const double __CN_Params_OscillatorLV[] = {
	1.0,	// 0 - rho_ii: "self inhibition"
	0.0	// 1 - sigma: noise level
};

const char* const __CN_ParamNames_OscillatorLV[] = {
	"Self inhibition",
	"\317\203"
};
const char* const __CN_ParamSyms_OscillatorLV[] = {
	"rho_ii",
	"sigma"
};


const double __CN_Vars_OscillatorLV[] = {
	0.,	// 0 - added a place for E
	0.1	// 1 - firing rate
};

const char* const __CN_VarNames_OscillatorLV[] = {
	"Membrane potential",
	"Firing rate"
};
const char* const __CN_VarSyms_OscillatorLV[] = {
	"E",
	"fr"
};








const double __CN_Params_OscillatorVdPol[] = {
	1.0,	// eta
	0.1,	// omega^2
	0.0	// noise level
};
const char* const __CN_ParamNames_OscillatorVdPol[] = {
	"\316\267",
	"\317\211\302\262",
	"\317\203"
};
const char* const __CN_ParamSyms_OscillatorVdPol[] = {
	"eta",
	"omegasq", // omega^2
	"sigma"
};

const double __CN_Vars_OscillatorVdPol[] = {
	0.1,	// amplitude
	0.0	// internal var
};
const char* const __CN_VarNames_OscillatorVdPol[] = {
	"Amplitude",
	"Internal var"
};
const char* const __CN_VarSyms_OscillatorVdPol[] = {
	"A",
	"v"
};


void
COscillatorVdPol::derivative( vector<double>& x, vector<double> &dx)
{
	damp(dx) = _x(x);
	d_x(dx) = (P.n.eta - pow( amp(x), 2)) * _x(x) - P.n.omega2 * amp(x) + Isyn(x);
}

// EOF
