// ;-*-C++-*- *  Time-stamp: "2008-10-15 16:59:51 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny (tnowotny@ucsd.edu)
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-08-02
 *
 * Oscillator units: Poisson, Lottka-Volterra, Colpitts, Van der Pol, and a synapse
 * to drive these (CSynapseOscilTuner)
 */



#ifndef CN_OSCILLATOR_H
#define CN_OSCILLATOR_H

#include "cnlib-unit.h"
#include "cnlib-model.h"
#include "RandomGen.h"



extern const double __CN_Params_OscillatorPoisson[];
extern const char* const __CN_ParamNames_OscillatorPoisson[];
extern const char* const __CN_ParamSyms_OscillatorPoisson[];
extern const double __CN_Vars_OscillatorPoisson[];
extern const char* const __CN_VarNames_OscillatorPoisson[];
extern const char* const __CN_VarSyms_OscillatorPoisson[];

struct SParametersOscilPoisson {
	double	tsp,
		tref,
		Vrst, Vfir,
		frl;
};

class COscillatorPoisson
  : public CNeuron {

    public:
	union {
		double i[sizeof(SParametersOscilPoisson)];
		SParametersOscilPoisson n;
	} P;
	virtual double &param_value( size_t idx)	{ return P.i[idx]; }
	virtual void reset_params()
		{
			memcpy( P.i, __CN_DescriptorTable[NT_POISSON].stock_param_values,
				sizeof(double) * p_no());
		}

	COscillatorPoisson( const char *inlabel, double x, double y, double z)
	      : CNeuron( NT_POISSON, inlabel, x, y, z)
		{
			reset_params();
			_status |= CN_NCONSCIOUS;
			t_firing_started = -P.n.tref;
			next_state_E = P.n.Vrst;
		}

	virtual void possibly_fire();

	virtual void derivative( vector<double>& u, vector<double>& uu)
		{}
	virtual void noise()
		{}
    private:
	double	next_state_E;
};






extern const double __CN_Params_OscillatorColpitts[];
extern const char* const __CN_ParamNames_OscillatorColpitts[];
extern const char* const __CN_ParamSyms_OscillatorColpitts[];
extern const double __CN_Vars_OscillatorColpitts[];
extern const char* const __CN_VarNames_OscillatorColpitts[];
extern const char* const __CN_VarSyms_OscillatorColpitts[];

struct SParametersOscilColpitts {
	double	a, g, q,
		eta;
};

class COscillatorColpitts
  : public CNeuron {

    public:
	union {
		double i[sizeof(SParametersOscilColpitts)];
		SParametersOscilColpitts n;
	} P;
	virtual double &param_value( size_t idx)	{ return P.i[idx]; }
	virtual void reset_params()
		{
			memcpy( P.i, __CN_DescriptorTable[NT_COLPITTS].stock_param_values,
				sizeof(double) * p_no());
		}

	double&  x0( vector<double>& b)	{ return b[idx+0]; }  // there's no E() for this one
	double&  x1( vector<double>& b)	{ return b[idx+1]; }
	double&  x2( vector<double>& b)	{ return b[idx+2]; }
	double& dx0( vector<double>& b)	{ return b[idx+0]; }
	double& dx1( vector<double>& b)	{ return b[idx+1]; }
	double& dx2( vector<double>& b)	{ return b[idx+2]; }

	COscillatorColpitts( const char *inlabel, double x, double y, double z)
	      : CNeuron( NT_COLPITTS, inlabel, x, y, z)
		{	reset_params();
		}

	virtual void possibly_fire()
		{}
	virtual void derivative( vector<double>&, vector<double>&);  // move code to possibly_fire()?
	virtual void noise()
		{}
};








extern const double __CN_Params_OscillatorLV[];
extern const char* const __CN_ParamNames_OscillatorLV[];
extern const char* const __CN_ParamSyms_OscillatorLV[];
extern const double __CN_Vars_OscillatorLV[];
extern const char* const __CN_VarNames_OscillatorLV[];
extern const char* const __CN_VarSyms_OscillatorLV[];

struct SParametersOscilLV {
	double	rho, sigma;
};

class COscillatorLV
  : public CNeuron {

    public:
	union {
		double i[sizeof(SParametersOscilLV)];
		SParametersOscilLV n;
	} P;
	virtual double &param_value( size_t idx)	{ return P.i[idx]; }
	virtual void reset_params()
		{
			memcpy( P.i, __CN_DescriptorTable[YT_AB].stock_param_values,
				sizeof(double) * p_no());
		}

	double&  fr( vector<double>& b)	{ return b[idx+1]; }
	double& dfr( vector<double>& b)	{ return b[idx+1]; }

	COscillatorLV( const char *inlabel, double x, double y, double z)
	      : CNeuron( NT_LV, inlabel, x, y, z)
		{	reset_params();
		}

	virtual void derivative( vector<double>& x, vector<double>& dx)
		{
			dE(dx) = fr(x) * (1.0 - P.n.rho * fr(x)) - Isyn(x);
		}

	virtual void noise()
		{
			dE(*_parent_model->V) = P.n.sigma * __R_std.n();
		}
	virtual void possibly_fire()
		{}
};





extern const double __CN_Params_OscillatorVdPol[];
extern const char* const __CN_ParamNames_OscillatorVdPol[];
extern const char* const __CN_ParamSyms_OscillatorVdPol[];
extern const double __CN_Vars_OscillatorVdPol[];
extern const char* const __CN_VarNames_OscillatorVdPol[];
extern const char* const __CN_VarSyms_OscillatorVdPol[];

struct SParametersOscilVdPol {
	double	eta, omega2, sigma;
};

class COscillatorVdPol
  : public CNeuron {

     public:
	union {
		double i[sizeof(SParametersOscilVdPol)];
		SParametersOscilVdPol n;
	} P;
	virtual double &param_value( size_t idx)	{ return P.i[idx]; }
	virtual void reset_params()
		{
			memcpy( P.i, __CN_DescriptorTable[NT_VDPOL].stock_param_values,
				sizeof(double) * p_no());
		}

	double&  amp( vector<double>& b)	{ return b[idx+0]; }
	double&   _x( vector<double>& b)	{ return b[idx+1]; }
	double& damp( vector<double>& b)	{ return b[idx+0]; }
	double&  d_x( vector<double>& b)	{ return b[idx+1]; }

	COscillatorVdPol( const char *inlabel, double x, double y, double z)
	      : CNeuron( NT_VDPOL, inlabel, x, y, z)
		{	reset_params();
		}

	virtual void possibly_fire()
		{}
	virtual void derivative( vector<double>&, vector<double>&);
	virtual void noise()
		{
			damp(*_parent_model->V) = P.n.sigma * __R.n();
		}
};





// An accessory synapse that transmits its unaltered Isyn to
// oscillator units, which may use it to set their firing rate or
// other parameter

class CSynapseOscilTuner
  : public CSynapse {

    public:
	CSynapseOscilTuner( const char *inlabel,
			    CNeuron *insource, CNeuron *intarget)
	      : CSynapse( YT_OSCILTUNER, inlabel, insource, intarget)
		{}

	virtual double &param_value( size_t idx)
		{
			return __cn_dummy_double;
		}
	virtual void reset_params()
		{
		}

	virtual double Isyn( vector<double>&)
		{}
	virtual void derivative( vector<double>&, vector<double>&)
		{}
	virtual void noise()
		{}
};


#endif

// EOF
