// ;-*-C++-*- *  Time-stamp: "2010-01-19 19:32:34 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny (tnowotny@ucsd.edu)
 *
 * License: GPL
 *
 * Initial version: 2008-08-02
 *
 */


#include <iostream>
#include <limits>

#include <gsl/gsl_statistics_double.h>

#include "base-unit.h"
#include "model.h"

using namespace std;



double __cn_dummy_double;

unsigned short __cn_default_unit_precision = 8;

int __cn_verbosely = 2;

char __cn_qlabelp[CN_MAX_LABEL_SIZE];

// ----- CUnit

int
__unit_compare( const void *l, const void *r)
{
	size_t dotpos = strrchr( (*(C_BaseUnit**)(l))->label, '.') - (*(C_BaseUnit**)(l))->label;
	int cmp_predot = strncmp( (*(C_BaseUnit**)(l))->label, (*(C_BaseUnit**)(r))->label,
				  dotpos);
	if ( cmp_predot )
		return cmp_predot;
	else {
		unsigned l1 = strtol( (*(C_BaseUnit**)(l))->label+dotpos+1, NULL, 10),
			 r1 = strtol( (*(C_BaseUnit**)(r))->label+dotpos+1, NULL, 10);
		return (l1 < r1) ?-1 :(l1 > r1) ?1 :0;
	}
}








// mind that child

C_BaseUnit::C_BaseUnit( TUnitType intype, const char *inlabel,
			CModel* inM, int s_mask)
      : _type (intype), _status (0 |/* CN_UENABLED |*/ s_mask),
	M (inM),
	param_schedulers (NULL), param_readers (NULL), param_ranges (NULL),
	_listener_disk (NULL), _listener_mem (NULL),
	precision (__cn_default_unit_precision)
{
	memset( label, 0, CN_MAX_LABEL_SIZE);
	if ( inlabel )
		strncpy( label, inlabel, CN_MAX_LABEL_SIZE);
	else
		snprintf( label, CN_MAX_LABEL_SIZE-1, "fafa%p", this);

	reset_params();
	// don't have field idx to do reset_vars() safely
}



void
C_BaseUnit::reset_state()
{
	if ( M && M->verbosely > 3 )
		fprintf( stderr, "Resetting \"%s\"\n", label);
	reset_vars();
	if ( is_listening() )
		restart_listening();
}


int
C_BaseUnit::param_idx_by_sym( const char *sym) const
{
	for ( int i = 0; i < p_no(); i++ )
		if ( strcmp( sym, __CNUDT[_type].stock_param_syms[i]) == 0 )
			return i;
	return -1;
}

int
C_BaseUnit::var_idx_by_sym( const char *sym) const
{
	for ( unsigned short i = 0; i < v_no(); i++ )
		if ( strcmp( sym, __CNUDT[_type].stock_var_syms[i]) == 0 )
			return i;
	return -1;
}




int
C_BaseUnit::attach_scheduler_device( const char *fname, bool is_looping,
				     TDeviceAssignee assignee_kind,
				     unsigned short idx)
{
	if ( !param_schedulers )
		param_schedulers = new vector<CSchedulerDevice>;

	CSchedulerDevice device;
	if ( device.setup( fname, is_looping, assignee_kind, idx) )
		return -1;
	param_schedulers->push_back( device);

	M->register_scheduler( this);

	return 0;
}



int
C_BaseUnit::attach_reader_device( const char *fname, bool is_looping,
				  TDeviceAssignee assignee_kind,
				  unsigned short idx)
{
	if ( !param_readers )
		param_readers = new vector<CReaderDevice>;

	CReaderDevice device;
	if ( device.setup( fname, is_looping, assignee_kind, idx) )
		return -1;
	if ( M->verbosely > 2 )
		fprintf( stderr, "%zu values in \"%s\" for %s{%s}\n", device.values.size(), fname, label, param_sym(idx));
	param_readers->push_back( device);

	M->register_reader( this);

	return 0;
}


int
C_BaseUnit::attach_param_range( double _min, double _span, double _max,
				TDeviceAssignee assignee_kind, unsigned short idx)
{
	if ( !param_ranges )
		param_ranges = new vector<CRangeSpec>;

	param_ranges->push_back( CRangeSpec( _min, _span, _max, assignee_kind, idx));

	M->register_param_range_unit( this);

	return 0;
}






void
C_BaseUnit::start_listening( int mask)
{
	if ( !M ) {
		fprintf( stderr, "start_listening() called for an unattached unit \"%s\"\n", label);
		return;
	}
	if ( _listener_disk || _listener_mem ) { // listening already; check if user wants us to listen differently
		if ( (_status | (mask & (CN_ULISTENING_DISK | CN_ULISTENING_MEM | CN_ULISTENING_1VARONLY | CN_ULISTENING_DEFERWRITE)))
		     != mask ) {
			stop_listening();  // this will NULLify _listener_{mem,disk}, avoiding recursion
			start_listening( mask);
			if ( M->verbosely > 4 )
				fprintf( stderr, "Unit \"%s\" was already listening\n", label);
			return;
		}
	}

      // deferred write implies a mem listener
	if ( mask & CN_ULISTENING_DEFERWRITE && !(mask & CN_ULISTENING_MEM) )
		mask |= CN_ULISTENING_MEM;

	if ( mask & CN_ULISTENING_MEM )
		_listener_mem = new vector<double>;

	if ( mask & CN_ULISTENING_DISK ) {
		if ( M->_status & CN_MDL_DISKLESS )
			fprintf( stderr, "Cannot get Unit \"%s\" to listen to disk in a diskless model\n", label);
		else {
			_listener_disk = new ofstream( (string(label)+".var").c_str(), ios_base::trunc);
			_listener_disk->precision( precision);

			*_listener_disk << "# " << label << " variables\n#<time>";
			if ( mask & CN_ULISTENING_1VARONLY )
				*_listener_disk << "\t<" << var_sym(0) << ">";
			else
				for ( unsigned short v = 0; v < v_no(); v++ )
					*_listener_disk << "\t<" << var_sym(v) << ">";
			*_listener_disk << endl;
			if ( M->verbosely > 4 )
				fprintf( stderr, "Unit \"%s\" now listening\n", label);
		}
	}

	_status |= (mask & (CN_ULISTENING_DISK | CN_ULISTENING_MEM
			    | CN_ULISTENING_1VARONLY | CN_ULISTENING_DEFERWRITE));

      // inform the model
	M->register_listener( this);
}


void
C_BaseUnit::stop_listening()
{
	if ( M->verbosely > 4 )
		fprintf( stderr, "Unit \"%s\" not listening now\n", label);

	if ( _listener_disk ) {
	      // do deferred write
		if ( _status & CN_ULISTENING_DEFERWRITE && _listener_mem )
			for ( vector<double>::iterator V = _listener_mem->begin();
			      V != _listener_mem->end(); ) {
				*_listener_disk << *(V++);
				if ( _status & CN_ULISTENING_1VARONLY )
					*_listener_disk << "\t" << (*V++);
				else
					for ( size_t v = 0; v < v_no(); v++ )
						*_listener_disk << "\t" << *(V++);
				*_listener_disk << endl;
			}

		_listener_disk->close();
		delete _listener_disk;
		_listener_disk = NULL;
	}
	if ( _listener_mem ) {
		delete _listener_mem;
		_listener_mem = NULL;
	}

	_status &= ~(CN_ULISTENING_MEM | CN_ULISTENING_DISK);

	if ( M )
		M->unregister_listener( this);
}




void
C_BaseUnit::tell()
{
	if ( _listener_disk && !(_status & CN_ULISTENING_DEFERWRITE) ) {
		*_listener_disk << model_time();
		if ( _status & CN_ULISTENING_1VARONLY )
			*_listener_disk << "\t" << var_value(0);
		else
			for ( size_t v = 0; v < v_no(); v++ )
				*_listener_disk << "\t" << var_value(v);
		*_listener_disk << endl;
	}

	if ( _listener_mem ) {
//		_listener_mem->push_back( 999);
		_listener_mem->push_back( model_time());
		if ( _status & CN_ULISTENING_1VARONLY )
			_listener_mem->push_back( var_value(0));
		else
			for ( size_t v = 0; v < v_no(); v++ )
				_listener_mem->push_back( var_value(v));
	}
}





// parameter reader, scheduler and range
int
C_BaseUnit::CReaderDevice::setup( const char *fname, bool inis_looping,
				  TDeviceAssignee inassignee_kind,
				  unsigned short inidx)
{
	assignee_kind = inassignee_kind,  idx = inidx,  is_looping = inis_looping;

	ifstream ins( fname);
	if ( !ins.good() )
		return -1;
	skipws(ins);

	while ( !ins.eof() && ins.good() ) {
		while ( ins.peek() == '#' || ins.peek() == '\n' )
			ins.ignore( numeric_limits<streamsize>::max(), '\n');
		double timestamp, datum;
		ins >> timestamp >> datum;
		values.push_back( pair<double,double>(timestamp, datum));
	}

	if ( values.size() == 0 ) {
		fprintf( stderr, "No usable values in \"%s\"\n", fname);
		return -1;
	}

	I = values.begin();

	return 0;
}

int
C_BaseUnit::CSchedulerDevice::setup( const char *fname, bool inis_looping,
				     TDeviceAssignee inassignee_kind,
				     unsigned short inidx)
{
	assignee_kind = inassignee_kind,  idx = inidx,  is_looping = inis_looping;

	ifstream ins( fname);
	if ( !ins.good() )
		return -1;
	skipws(ins);

	while ( ins.peek() == '#' || ins.peek() == '\n' )
		ins.ignore( numeric_limits<streamsize>::max(), '\n');
	ins >> dt;

	while ( true ) {
		while ( ins.peek() == '#' || ins.peek() == '\n' )
			ins.ignore( numeric_limits<streamsize>::max(), '\n');
		double datum;
		ins >> datum;
		if ( ins.eof() || !ins.good() )
			break;
		values.push_back( datum);
	}

	if ( dt == 0 ) {
		fprintf( stderr, "dt is 0 in \"%s\"\n", fname);
		return -1;
	}
	if ( values.size() < 2 ) {
		fprintf( stderr, "Need at least 2 scheduled values in \"%s\"\n", fname);
		return -1;
	}
	return 0;
}




void
C_BaseUnit::CReaderDevice::update_host_state( C_BaseUnit *U)
{
      // position I such that *I < model_time < *(I+1)
	while ( I+1 != values.end() && (I+1)->first < U->model_time() )
		I++;

	switch ( assignee_kind ) {
	case RDA_PARAM:
		U->param_value(idx) = I->second;
		U -> param_changed_hook();
	    break;
	case RDA_VAR:
		U->var_value(idx) = I->second;
	    break;
	default:
	    break;
	}

	if ( I == values.end() && is_looping )
		I = values.begin();
}


void
C_BaseUnit::CSchedulerDevice::update_host_state( C_BaseUnit *U)
{
	size_t	i_abs = (size_t)(U->model_time() / dt),
		i_eff = is_looping
			? i_abs % values.size()
			: min (i_abs, values.size() - 1);

//	cerr << "@" << U->model_time() << ' ' << U->label << " p[" << idx << "] = " << "values[" << i_eff << "] = " << values[i_eff] << endl;

	switch ( assignee_kind ) {
	case RDA_PARAM:	U->param_value(idx) = values[i_eff];	break;
	case RDA_VAR:	U->  var_value(idx) = values[i_eff];	break;
	default:;
	}

	U -> param_changed_hook();
}




void
C_BaseUnit::CRangeSpec::update_host_state( C_BaseUnit *U)
{
	double	range_frac = (fmod( U->model_time(), _span)) / _span * (_max - _min);

	switch ( assignee_kind ) {
	case RDA_PARAM:	U->param_value(idx) = _min + range_frac;	break;
	case RDA_VAR:		U->  var_value(idx) = _min + range_frac;	break;
	default:;
	}

	U -> param_changed_hook();
}







void
C_BaseUnit::dump( bool with_params) const
{
	printf( "[%lu]%s (%s)\n", serial_id, label, type_name());

	if ( with_params ) {
		printf( "Pp: ");
		for ( size_t p = 0; p < p_no(); p++ )
			printf( "%s = %g; ", param_sym(p), get_param_value(p));
		printf( "\n");
	}
	printf( "Vv: ");
	for ( size_t v = 0; v < v_no(); v++ )
		printf( "%s = %g; ", var_sym(v), get_var_value(v));

	printf( "\n(%slistening", is_listening() ? "" : "not ");
	if ( is_listening() )
		printf( " to %s%s%s", _listener_mem ? "mem" : "",
			_listener_mem && _listener_disk ? ", " : "",
			_listener_disk ? "disk" : "");
	printf( ")\n");
}




C_BaseUnit::~C_BaseUnit()
{
	if ( M && M->verbosely > 5 )
		fprintf( stderr, "   deleting base unit \"%s\"\n", label);

	if ( is_listening() ) {
		stop_listening();
		if ( M && M->model_time() == 0. )
		      // nothing has been written yet, delete the files on disk
			unlink( (string(label) + ".var").c_str());
	}
	if ( M )
		M->exclude_unit( this, false);

	if ( param_schedulers )
		delete param_schedulers;
	if ( param_readers )
		delete param_readers;
	if ( param_ranges )
		delete param_ranges;
}








// ----- C_BaseNeuron


C_BaseNeuron::~C_BaseNeuron()
{
	if ( M && M->verbosely > 4 )
		fprintf( stderr, "  deleting base neuron \"%s\"\n", label);
	for ( vector<C_BaseSynapse*>::reverse_iterator Y = _dendrites.rbegin(); Y != _dendrites.rend(); Y++ ) {
		(*Y) -> _target = NULL;  // prevent target from touching our vector
		delete (*Y);
	}
	for ( vector<C_BaseSynapse*>::reverse_iterator Y = _axonal_harbour.rbegin(); Y != _axonal_harbour.rend(); Y++ ) {
		(*Y) -> _source = NULL;
//		delete (*Y);  // rely on cull_deaf_absynapses outright
	}

	if ( _spikelogger_agent ) {
		if ( M && !(_spikelogger_agent->_status & CN_KL_IDLE) )
			M->unregister_spikelogger( this);
		delete _spikelogger_agent;
	}
}





double
C_BaseNeuron::Isyn()
{
	double I = 0.;
	for ( vector<C_BaseSynapse*>::iterator Y = _dendrites.begin(); Y != _dendrites.end(); Y++ )
		I += (*Y) -> Isyn();
	return I;
}

double
C_BaseNeuron::Isyn( vector<double> &b)
{
	double I = 0.;
	for ( vector<C_BaseSynapse*>::iterator Y = _dendrites.begin(); Y != _dendrites.end(); Y++ )
		I += (*Y) -> Isyn( b);
	return I;
}



void
C_BaseNeuron::dump( bool with_params) const
{
	C_BaseUnit::dump( with_params);
	if ( _spikelogger_agent )
		printf( "(logging spikes)\n");

}









// --- SSpikeloggerService

double
SSpikeloggerService::sdf( double at, double sample_width, double sigma, unsigned *nspikes)
{
	if ( nspikes )
		*nspikes = 0;

	double	dt,
		result = 0.;
	for ( vector<double>::iterator T = spike_history.begin(); T != spike_history.end(); T++ ) {
		dt = *T - at;
		if ( dt < -sample_width/2. )
			continue;
		if ( dt >  sample_width/2. )
			break;
		if ( nspikes )
			(*nspikes)++;
		result += exp( -dt*dt/(sigma * sigma));
	}
	return result;
}



double
SSpikeloggerService::shf( double at, double sample_width)
{
	double	dt,
		last_spike_at;
	vector<double> intervals;
	bool	counted_one = false;
	for ( vector<double>::iterator T = spike_history.begin(); T != spike_history.end(); T++ ) {
		dt = *T - at;
		if ( dt < -sample_width/2. )
			continue;
		if ( dt >  sample_width/2. )
			break;

		if ( counted_one )
			intervals.push_back( last_spike_at - *T);
		else
			counted_one = true;

		last_spike_at = *T;
	}

	return gsl_stats_sd( intervals.data(), 1, intervals.size());
//	return stdev( intervals);
}





size_t
SSpikeloggerService::get_sxf_vector( vector<double> *sdf_buffer, vector<double> *shf_buffer,
				     vector<unsigned> *nspikes_buffer,
				     double sample_period, double sigma,
				     double from, double to)
{
	if ( from < 0 + sample_period/2 )
		from = sample_period/2;

	if ( to == 0. || to > _client->M->model_time() - sample_period/2 )
		to = _client->M->model_time() - sample_period/2;

	if ( sdf_buffer )	sdf_buffer->clear();
	if ( shf_buffer )	shf_buffer->clear();
	if ( nspikes_buffer)	nspikes_buffer->clear();

	if ( from + sample_period/2 > to - sample_period/2 )
		return 0;

	// cerr << "get_sxf_vector( " << from << ", " << to << ") for "
	//      << label << ", _spike_history.size() = " << _spike_history->size() << endl;

	for ( double t = from; t <= to; t += sample_period ) {
		unsigned nspikes = 0;
		double	sdf_value = sdf( t, sample_period, sigma, &nspikes);
		if ( sdf_buffer )	sdf_buffer->push_back( sdf_value);
		if ( shf_buffer )	shf_buffer->push_back( shf( t, sample_period));
		if ( nspikes_buffer )	nspikes_buffer->push_back( nspikes);
	}

	return (to - from) / sample_period;
}







void
SSpikeloggerService::sync_history()
{
	if ( !_client->M || (_client->M && _client->M->_status & CN_MDL_DISKLESS) )
		return;

//	if ( M && M->verbosely > 4 )
//		cout << "  writing spiking history for " << label
//		     << " (" << _spike_history->size() << " spikes)\n";
	ofstream spikecnt_strm( (string(_client->label) + ".spikes").c_str());
	spikecnt_strm.precision( _client->precision);
	spikecnt_strm << "#spike time\n";

	for ( vector<double>::iterator V = spike_history.begin(); V != spike_history.end(); V++ )
		spikecnt_strm << *V << endl;

	if ( _status & CN_KL_COMPUTESDF ) {
		ofstream sdf_strm( (string(_client->label) + ".sxf").c_str());
		sdf_strm.precision( _client->precision);
		sdf_strm << "#<time>\t<sdf>\t<shf>\t<nspikes>\n";

		vector<double> sdf_vector, shf_vector;
		vector<unsigned> nspikes_vector;
		get_sxf_vector( &sdf_vector, &shf_vector, &nspikes_vector);

		double t = 0.;
		for ( size_t i = 0; i < sdf_vector.size(); i++, t += sample_period )
			sdf_strm << t << "\t"
				 << sdf_vector[i] << "\t"
				 << shf_vector[i] << "\t"
				 << nspikes_vector[i] << endl;
	}
}





size_t
SSpikeloggerService::n_spikes_since( double since)
{
	size_t i = 0;
	for ( vector<double>::iterator K = spike_history.begin(); K != spike_history.end(); K++, i++ )
		if ( *K > since )
			return spike_history.size() - i;
	return 0;
}





// ----- CSynapse


C_BaseSynapse::C_BaseSynapse( TUnitType intype, const char *inlabel,
			      C_BaseNeuron *insource, C_BaseNeuron *intarget,
			      CModel *inM, int s_mask)
      : C_BaseUnit (intype, inlabel, inM, s_mask),
	_source (insource), _target (intarget),
	t_last_release_started (-INFINITY)
{
//	cout << _target->label << "->dendrites.push_back( " << label << "\n";
	_target->_dendrites.push_back( this);
	_source->_axonal_harbour.push_back( this);
}



C_BaseSynapse::~C_BaseSynapse()
{
	if ( M && M->verbosely > 4 )
		fprintf( stderr, "  deleting base synapse \"%s\"\n", label);

	if ( _target )
		for ( vector<C_BaseSynapse*>::iterator Y = _target->_dendrites.begin(); Y != _target->_dendrites.end(); Y++ )
			if ( (*Y) == this ) {
//				cout << "    removing ourselves from " << _target->label << " dendrites\n";
				_target -> _dendrites.erase( Y);
				break;
			}
	if ( _source )
		for ( vector<C_BaseSynapse*>::iterator Y = _source->_axonal_harbour.begin(); Y != _source->_axonal_harbour.end(); Y++ )
			if ( (*Y) == this ) {
//				cout << "    removing ourselves from " << _target->label << " dendrites\n";
				_source -> _axonal_harbour.erase( Y);
				break;
			}
}


// EOF
