// ;-*-C++-*- *  Time-stamp: "2009-12-11 01:31:04 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny (tnowotny@ucsd.edu)
 *
 * License: GPL
 *
 * Initial version: 2008-08-02
 *
 */


#ifndef LIBCN_BASE_UNIT_H
#define LIBCN_BASE_UNIT_H

#include "config.h"

#include <fstream>
#include <cstring>
#include <vector>

#include "types.h"


using namespace std;

// this gets referenced in case of out-of-bounds idx or misspelled sym
// in param accessors
extern double __cn_dummy_double;



// forward decls
class CModel;
class C_BaseUnit;









// for all units
//#define CN_UENABLED			(1 << 0)
#define CN_UOWNED			(1 << 1)
#define CN_UHASPARAMRANGE		(1 << 2)
#define CN_ULISTENING_MEM		(1 << 3)
#define CN_ULISTENING_DISK		(1 << 4)
#define CN_ULISTENING_1VARONLY		(1 << 5)
#define CN_ULISTENING_DEFERWRITE	(1 << 6)
//#define CN_SMPLTAPE_LOOP		(1 << 7)
//#define CN_NDYNPARAMS			(1 << 8)

// only for neurons
#define CN_NFIRING			(1 << 9)  // firing now
#define CN_NREFRACT			(1 << 10)  // in refractory phase now


#define CN_MAX_LABEL_SIZE 40


// the base unit provides the methods for the following:
// * classification;
// * access to parameters, tape reader and range interface;
// * attachment to the mother model;
// * listening, i.e., keeping a history of vars along a timeline;
class C_BaseUnit {

    private:
	C_BaseUnit();  // not callable

    protected:
	TUnitType
		_type;  // will look up p, pno and vno from __CNUDT using _type as index
    public:
	TUnitType
		type()		{  return _type;   }

      // classification
	const int	traits()	const {  return __CNUDT[_type].traits;			}
	const bool	is_hostable()	const {  return __CNUDT[_type].traits & UT_HOSTED;		}
	const bool	is_ddtbound()	const {  return __CNUDT[_type].traits & UT_DDTSET;		}
	const bool	is_neuron()	const {  return _type >= NT_FIRST && _type <= NT_LAST;	}
	const bool	is_synapse()	const {  return _type >= YT_FIRST && _type <= YT_LAST;	}
	const bool	is_oscillator()	const {  return __CNUDT[_type].traits & UT_OSCILLATOR;	}
	const bool	is_conscious()	const {  return is_oscillator();				}

	const char *class_name() const
		{  return is_neuron() ? "Neuron" : "Synapse";		}
	const char *type_name() const
		{  return __CNUDT[_type].reporting_name;		}
	const char *type_id() const
		{  return __CNUDT[_type].id;				}
	const char *type_description() const
		{  return __CNUDT[_type].description;			}

      // parameter & variable names and symbols
	const char *const param_name( size_t i)	const { return __CNUDT[_type].stock_param_names[i]; }
	const char *const param_sym( size_t i)	const { return __CNUDT[_type].stock_param_syms[i];  }
	int param_idx_by_sym( const char*) const;
	const char *const var_name( size_t i)	const { return __CNUDT[_type].stock_var_names[i];   }
	const char *const var_sym( size_t i)	const { return __CNUDT[_type].stock_var_syms[i];    }
	int var_idx_by_sym( const char*) const;
	unsigned short v_no() const	{ return __CNUDT[_type].vno; }
	unsigned short p_no() const	{ return __CNUDT[_type].pno; }

      // identification
	unsigned long
		serial_id;  // assigned incrementally as read by import_NetworkML
	char	label[CN_MAX_LABEL_SIZE];
	const char* const qlabelp();  // right-padded to the longest label in the model

      // status bitfield & properties
    protected:
	int	_status;
    public:
	int	status()	{  return _status; }

      // ctor & dtor
    protected:
	C_BaseUnit( TUnitType, const char *label,
		    CModel*, int s_mask);
    public:
	virtual ~C_BaseUnit();  // surely virtual

      // parent model
	friend class CModel;
	friend class SSpikeloggerService;
    protected:
	CModel	*M;
    public:
	const CModel&	parent_model() const	{ return *M; }
	bool	is_owned() const		{ return _status & CN_UOWNED; }
	const double&
		model_time();  // defined in model.h

    public:
      // private copy of params
	vector<double> P;
	double get_param_value( size_t p) const
		{  return P[p];  }
	double get_param_value( const char *sym) const
		{
			int id = param_idx_by_sym( sym);
			return (id == -1) ? __cn_dummy_double : P[id];
		}
	double &param_value( size_t p)	{  return P[p];  }
	double &param_value( const char *sym)
		{
			int id = param_idx_by_sym( sym);
			return (id == -1) ? __cn_dummy_double : P[id];
		}
	void reset_params()
		{
			P.resize( p_no());
			memcpy( P.data(), __CNUDT[_type].stock_param_values,
				sizeof(double) * p_no());
			param_changed_hook();
		}
	virtual void dump( bool with_params = false) const;

      // Tape reader interface
	enum TDeviceAssignee { RDA_VOID, RDA_PARAM, RDA_VAR };

	class CReaderDevice {
	    public:
		vector< pair<double, double> > values;
		bool is_looping;
		TDeviceAssignee assignee_kind;
		unsigned short idx;

		CReaderDevice()
		      : assignee_kind (RDA_VOID)
			{}
		int setup( const char *fname, bool is_looping,
			   TDeviceAssignee, unsigned short idx);
		void update_host_state( C_BaseUnit*);
		vector< pair< double, double> >::iterator I;
	};
	class CSchedulerDevice {
	    public:
		vector<double> values;
		double dt;
		bool is_looping;
		TDeviceAssignee assignee_kind;
		unsigned short idx;

		CSchedulerDevice()
		      : assignee_kind (RDA_VOID)
			{}
		int setup( const char *fname, bool is_looping,
			   TDeviceAssignee, unsigned short idx);
		void update_host_state( C_BaseUnit*);
	};

      // Ranges are similar to tape readers, but with calculated values instead
	class CRangeSpec {
	    public:
		double _min, _span, _max;
		TDeviceAssignee assignee_kind;
		unsigned short idx;

		CRangeSpec()
		      : assignee_kind (RDA_VOID)
			{}
		CRangeSpec( double __min, double __span, double __max,
			    TDeviceAssignee _assignee_kind, unsigned short _idx)
		      : _min (__min), _span (__span), _max (__max),
			assignee_kind (_assignee_kind),
			idx (_idx)
			{}

		int setup( double _min, double _span, double _max,
			   TDeviceAssignee, unsigned short idx);
		void update_host_state( C_BaseUnit*);
	};

	vector<CSchedulerDevice> *param_schedulers;
	vector<CReaderDevice> *param_readers;
	vector<CRangeSpec> *param_ranges;
	int attach_scheduler_device( const char *fname, bool is_looping,
				     TDeviceAssignee, unsigned short idx);
	int attach_reader_device( const char *fname, bool is_looping,
				  TDeviceAssignee, unsigned short idx);
	int attach_param_range( double _min, double _duration, double _max,
				TDeviceAssignee, unsigned short idx);
	bool is_doped()
		{
			return param_ranges || param_readers || param_schedulers;
		}
	virtual void param_changed_hook()
		{}

      // access to vars: differs per hosted or standalone
	virtual double &var_value( size_t) = 0;
	virtual const double &get_var_value( size_t) const = 0;
	virtual void reset_vars() = 0;
	virtual void reset_state();

      // state history
	bool is_listening() const
		{
			return _status & (CN_ULISTENING_DISK | CN_ULISTENING_MEM);
		}
	void start_listening( int mask = 0 | CN_ULISTENING_DISK);
	void stop_listening();
	void restart_listening()
		{
			int lbits = _status & (CN_ULISTENING_DISK | CN_ULISTENING_MEM
					       | CN_ULISTENING_1VARONLY | CN_ULISTENING_DEFERWRITE);
			stop_listening();
			start_listening( lbits);
		}
	void pause_listening();
	void resume_listening();

    private:
      // where vars are written on each time step
	ofstream *_listener_disk;
      // ... and/or stored, in a diskless model
	vector<double> *_listener_mem;
    public:
      // by this method
	void tell();
	const vector<double> *listener_mem() const	{ return _listener_mem; }

	unsigned short precision;

      // one common method for all descendants
    public:
	virtual void noise()
		{}
};



extern unsigned short __cn_default_unit_precision;

extern int __cn_verbosely;

extern char __cn_qlabelp[CN_MAX_LABEL_SIZE];




int __unit_compare( const void*, const void*);




#endif

// EOF
