// ;-*-C++-*- *  Time-stamp: "2010-02-01 18:21:56 hmmr"

/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny
 *
 * License: GPL
 *
 * Initial version: 2008-09-02
 *
 * Class CModel
 */

/*--------------------------------------------------------------------------

The wrapper class which takes lists of pointers to neurons and synapses
which are networked to a neural system and assembles a common state
vector and handles the derivatives. At the same time it serves the neurons
and synapses their state at any given time and allows them to adjust their
parameters.

--------------------------------------------------------------------------*/


#ifndef LIBCN_MODEL_H
#define LIBCN_MODEL_H

#include "config.h"

#include <list>
#include <vector>
#include <string>

#include <libxml/parser.h>
#include <libxml/tree.h>

#ifdef USE_GSL_RNG
#include <gsl/gsl_rng.h>
#else
#include "newran.h"
#endif

#include "base-neuron.h"
#include "base-synapse.h"
#include "hosted-neurons.h"
#include "hosted-synapses.h"
#include "standalone-neurons.h"
#include "standalone-synapses.h"
#include "integrate-rk65.h"


using namespace std;


// CModel _status bits
#define CN_MDL_LOGDT		(1 << 0)
#define CN_MDL_LOGSPIKERS	(1 << 1)
#define CN_MDL_LOGUSINGID	(1 << 2)
#define CN_MDL_SORTUNITS	(1 << 3)
#define CN_MDL_ALLOWNOISE	(1 << 4)
#define CN_MDL_NOTREADY		(1 << 5)
#define CN_MDL_DISKLESS		(1 << 6)
#define CN_MDL_HAS_DDTB_UNITS	(1 << 7)
#define CN_MDL_DISPLAY_PROGRESS_PERCENT	(1 << 8)
#define CN_MDL_DISPLAY_PROGRESS_TIME	(1 << 9)

class CModel {

    public:
	string	name;

    private:
	int	_status;
    public:
	int status()	{ return _status; }

      // structure ------
    private:
	friend class C_HostedNeuron;
	friend class C_HostedConductanceBasedNeuron;
	friend class C_HostedRateBasedNeuron;
	friend class C_HostedSynapse;
    public:
	size_t units()			{ return unit_list.size(); }

    private:
	unsigned long
		_global_unit_id_reservoir;
    private:
	vector<C_BaseUnit*>		unit_list; // all units together
      // these have derivative(), are churned in _integrator->cycle()
	vector<C_HostedNeuron*>		hosted_neu_list;
	vector<C_HostedSynapse*>	hosted_syn_list;
      // these need preadvance() and fixate()
	vector<C_StandaloneNeuron*>	standalone_neu_list;
	vector<C_StandaloneSynapse*>	standalone_syn_list;
      // ... also these, but at discrete dt only
      // (only the standalone map units currently)
	vector<C_StandaloneNeuron*>	ddtbound_neu_list;
	vector<C_StandaloneSynapse*>	ddtbound_syn_list;

      // neurons that can possibly_fire() (various oscillators), and
      // have no inputs, and hence not dependent on anything else
	vector<C_BaseNeuron*>		conscious_neu_list;

      // various lists to avoid traversing all of them in unit_list:
      // listeners, spikeloggers & readers
	vector<C_BaseUnit*>		lisn_unit_list;
      // uses a meaningful do_spikelogging_or_whatever
	vector<C_BaseNeuron*>		spikelogging_neu_list;
      // `Multiplexing AB' synapses are treated very specially
	vector<C_BaseSynapse*>		mx_syn_list;
	vector<C_BaseUnit*>		reader_unit_list;
	vector<C_BaseUnit*>		scheduler_unit_list;
	vector<C_BaseUnit*>		param_range_unit_list;

	list<double>		scheduler_update_periods;
	list<unsigned>		scheduler_update_periods_last_checked;

    public:
	const C_BaseUnit    *unit_by_label( const char *) const;
	const C_BaseNeuron  *neuron_by_label( const char *) const;
	const C_BaseSynapse *synapse_by_label( const char *) const;

	size_t hosted_unit_cnt()
		{  return hosted_neu_list.size() + hosted_syn_list.size();		}
	size_t standalone_unit_cnt()
		{  return standalone_neu_list.size() + standalone_syn_list.size();	}
	size_t ddtbound_unit_cnt()
		{  return ddtbound_neu_list.size() + ddtbound_syn_list.size();		}
	size_t total_neuron_cnt()
		{  return hosted_neu_list.size()
				+ standalone_neu_list.size()
				+ ddtbound_neu_list.size();		}
	size_t total_synapse_cnt()
		{  return hosted_syn_list.size()
				+ standalone_syn_list.size()
				+ ddtbound_syn_list.size();		}
	size_t conscious_neuron_cnt()
		{  return conscious_neu_list.size();	}
	size_t listening_unit_cnt()
		{  return lisn_unit_list.size();	}
	size_t spikelogging_neuron_cnt()
		{  return spikelogging_neu_list.size();	}
	size_t reader_unit_cnt()
		{  return reader_unit_list.size();	}
	size_t scheduler_unit_cnt()
		{  return scheduler_unit_list.size();	}
	size_t param_range_unit_cnt()
		{  return param_range_unit_list.size();	}

      // if is_last == true, do allocations of hosted units' vars immediately
      // otherwise defer until addition is done with is_last == true
      // or the user calls finalize_additions
	int include_unit( C_HostedNeuron*, bool is_last = true);
	int include_unit( C_HostedSynapse*, bool is_last = true);
	void finalize_additions();
	int include_unit( C_StandaloneNeuron*);
	int include_unit( C_StandaloneSynapse*);
    private:
	void _include_base_unit( C_BaseUnit*);
	int _check_new_synapse( C_BaseSynapse*);
    public:
	C_BaseUnit* exclude_unit( C_BaseUnit*, bool do_delete = false);
	// return NULL if do_delete == true, the excluded unit otherwise, even if it was not owned
	void delete_unit( C_BaseUnit* u)
		{  exclude_unit( u, true);  }
    private:
	friend class C_BaseUnit;
	void register_listener( C_BaseUnit*);
	void unregister_listener( C_BaseUnit*);
	friend class C_BaseNeuron;
	friend class SSpikeloggerService;
	void register_spikelogger( C_BaseNeuron*);
	void unregister_spikelogger( C_BaseNeuron*);
	void register_mx_synapse( C_BaseSynapse*);
	void unregister_mx_synapse( C_BaseSynapse*);
	void register_reader( C_BaseUnit*);
	void unregister_reader( C_BaseUnit*);
	void register_scheduler( C_BaseUnit*);
	void unregister_scheduler( C_BaseUnit*);
	void register_param_range_unit( C_BaseUnit*);
	void unregister_param_range_unit( C_BaseUnit*);

    private:
	unsigned short	_longest_label;
    public:
	unsigned short longest_label()  { return _longest_label; }

    public:
      // ctor, dtor
	CModel( const char *inname, CIntegrate_base *inRK65Setup, int instatus);
       ~CModel();

	void reset( bool also_reset_params = false);
	void reset_state_all_units();

    public:
      // NeuroMl interface
	int import_NetworkML( const char *fname, bool appending = false);
	int import_NetworkML( xmlDoc *doc, const char *fname, bool appending = false);  // fname is merely informational here
	void cull_deaf_synapses();  // those with gsyn == 0
	void cull_blind_synapses(); // those with _source == NULL
	int export_NetworkML( const char *fname);
	int export_NetworkML( xmlDoc *doc);
	void dump_metrics();
    private:
	int _process_populations( xmlNode*);
	int _process_population_instances( xmlNode*, const xmlChar*, TUnitType);

	int _process_projections( xmlNode*);
	int _process_projection_connections( xmlNode*, const xmlChar*, TUnitType,
					     const xmlChar *src_grp_prefix, const xmlChar *tgt_grp_prefix);

      // the essential mechanical parts: ----
      // hosted unit variables
    private:
	vector<double> V,	// contains catenated var vectors of all constituent neurons and synapses
		       W;	// V and W alternate in the capacity of the main vector, so avoiding many a memcpy
	size_t	_var_cnt;	// total # of variables (to be) allocated in V an W, plus one for model_time
    public:
	size_t vars()			{ return _var_cnt;         }

      // integrator interface
    private:
	friend class CIntegrate_base;
	friend class CIntegrateRK65;
	CIntegrate_base
		*_integrator;
    public:
	const double& model_time()			{ return V[0];            }
	const double& dt()				{ return _integrator->dt;	}
	const double& dt( const double&);
	const double& dt_min()				{ return _integrator->_dt_min;	}
	const double& dt_max()				{ return _integrator->_dt_max;	}
      // this one is internal
	friend class CSynapseMxAB_dd;
    private:
	const double& model_time( vector<double> &x)	{ return x[0];            }

    private:
	unsigned long
		_cycle;
	double	_discrete_time;
	double	_discrete_dt;
    public:
	unsigned long cycle()				{ return _cycle;          }
	const double& model_discrete_time()		{ return _discrete_time;  }
	const double& discrete_dt()			{ return _discrete_dt;    }
	const double& discrete_dt( const double& new_ddt);

      // simulation
    private:
	void prepare_advance();
	unsigned _do_advance_on_pure_hosted( double, double*);
	unsigned _do_advance_on_pure_standalone( double, double*);
	unsigned _do_advance_on_pure_ddtbound( double, double*);
	unsigned _do_advance_on_mixed( double, double*);
    public:
	unsigned advance( double dist, double *cpu_time_p = NULL);

    public:
	double	spike_threshold, // above which neurons will detect a spike
		spike_lapse;  // if occurs less than this after the unit's _last_spike_end
	// (begs to be moved to SSpikeloggerService)

    public:
	double	logging_dt;
    private:
	ofstream
		*_dt_logger, *_spike_logger;

    public:
	int process_listener_tags( vector<string> &regex_tags, int lbits);
	int process_spikelogger_tags( vector<string> &regex_tags,
				      double sample_period = 0., double sigma = 0.);  // defaults disable sdf computation
	int process_putout_tags( vector<string> &regex_tags);
	struct SDecimationGroup {
		string name;
		float fraction;
	};
	int process_decimate_tags( vector<SDecimationGroup> &regex_tags);

    public:
	int	verbosely;

#ifdef USE_GSL_RNG
	gsl_rng	*_rng;
#else
	Uniform __newran_U;
#endif

	double rng_sample()
		{
#ifdef USE_GSL_RNG
			return gsl_rng_uniform_pos( _rng);
#else
			return __newran_U.Next();
#endif
		}

      // various convenience fields and members
	vector<C_BaseUnit*>::		iterator ulist_begin()	{ return unit_list.begin();		}
	vector<C_BaseUnit*>::		iterator ulist_end()	{ return unit_list.end();		}
	vector<C_HostedNeuron*>::	iterator hnlist_begin()	{ return hosted_neu_list.begin();	}
	vector<C_HostedNeuron*>::	iterator hnlist_end()	{ return hosted_neu_list.end();		}
	vector<C_HostedSynapse*>::	iterator hylist_begin()	{ return hosted_syn_list.begin();	}
	vector<C_HostedSynapse*>::	iterator hylist_end()	{ return hosted_syn_list.end();		}
	vector<C_StandaloneNeuron*>::	iterator snlist_begin()	{ return standalone_neu_list.begin();	}
	vector<C_StandaloneNeuron*>::	iterator snlist_end()	{ return standalone_neu_list.end();	}
	vector<C_StandaloneSynapse*>::	iterator sylist_begin()	{ return standalone_syn_list.begin();	}
	vector<C_StandaloneSynapse*>::	iterator sylist_end()	{ return standalone_syn_list.end();	}
	vector<C_StandaloneNeuron*>::	iterator dnlist_begin()	{ return ddtbound_neu_list.begin();	}
	vector<C_StandaloneNeuron*>::	iterator dnlist_end()	{ return ddtbound_neu_list.end();	}
	vector<C_StandaloneSynapse*>::	iterator dylist_begin()	{ return ddtbound_syn_list.begin();	}
	vector<C_StandaloneSynapse*>::	iterator dylist_end()	{ return ddtbound_syn_list.end();	}

	vector<C_BaseUnit*>::		reverse_iterator ulist_rbegin()	{ return unit_list.rbegin();		}
	vector<C_BaseUnit*>::		reverse_iterator ulist_rend()	{ return unit_list.rend();		}
	vector<C_HostedNeuron*>::	reverse_iterator hnlist_rbegin(){ return hosted_neu_list.rbegin();	}
	vector<C_HostedNeuron*>::	reverse_iterator hnlist_rend()	{ return hosted_neu_list.rend();	}
	vector<C_HostedSynapse*>::	reverse_iterator hylist_rbegin(){ return hosted_syn_list.rbegin();	}
	vector<C_HostedSynapse*>::	reverse_iterator hylist_rend()	{ return hosted_syn_list.rend();	}
	vector<C_StandaloneNeuron*>::	reverse_iterator snlist_rbegin(){ return standalone_neu_list.rbegin();	}
	vector<C_StandaloneNeuron*>::	reverse_iterator snlist_rend()	{ return standalone_neu_list.rend();	}
	vector<C_StandaloneSynapse*>::	reverse_iterator sylist_rbegin(){ return standalone_syn_list.rbegin();	}
	vector<C_StandaloneSynapse*>::	reverse_iterator sylist_rend()	{ return standalone_syn_list.rend();	}

	vector<C_BaseUnit*>::		iterator lulist_begin()	{ return lisn_unit_list.begin();	}
	vector<C_BaseUnit*>::		iterator lulist_end()	{ return lisn_unit_list.end();		}
	vector<C_BaseUnit*>::		iterator rlist_begin()	{ return reader_unit_list.begin();	}
	vector<C_BaseUnit*>::		iterator rlist_end()	{ return reader_unit_list.end();	}
	vector<C_BaseNeuron*>::		iterator cnlist_begin()	{ return conscious_neu_list.begin();	}
	vector<C_BaseNeuron*>::		iterator cnlist_end()	{ return conscious_neu_list.end();	}
	vector<C_BaseNeuron*>::		iterator knlist_begin()	{ return spikelogging_neu_list.begin();	}
	vector<C_BaseNeuron*>::		iterator knlist_end()	{ return spikelogging_neu_list.end();	}
	vector<C_BaseSynapse*>::	iterator mxylist_begin(){ return mx_syn_list.begin();		}
	vector<C_BaseSynapse*>::	iterator mxylist_end()	{ return mx_syn_list.end();		}

	// vector<C_BaseUnit*>::		reverse_iterator rlist_rbegin()	{ return reader_unit_list.rbegin();	}
	// vector<C_BaseUnit*>::		reverse_iterator rlist_rend()	{ return reader_unit_list.rend();	}
};




// by popular demand
#define for_all_units(U) \
	for ( vector<C_BaseUnit*>::iterator U = ulist_begin(); U != ulist_end(); U++ )
#define for_all_neurons(U) \
	for ( vector<C_BaseUnit*>::iterator U = ulist_begin(); U != ulist_end(); U++ ) if ( (*U)->is_neuron() )
#define for_all_synapses(U) \
	for ( vector<C_BaseUnit*>::iterator U = ulist_begin(); U != ulist_end(); U++ ) if ( (*U)->is_synapse() )
#define for_all_neurons_reversed(U) \
	for ( vector<C_BaseUnit*>::reverse_iterator U = ulist_rbegin(); U != ulist_rend(); U++ ) if ( (*U)->is_neuron() )
#define for_all_synapses_reversed(U) \
	for ( vector<C_BaseUnit*>::reverse_iterator U = ulist_rbegin(); U != ulist_rend(); U++ ) if ( (*U)->is_synapse() )

#define for_all_hosted_neurons(U) \
	for ( vector<C_HostedNeuron*>::iterator U = hnlist_begin(); U != hnlist_end(); U++ )
#define for_all_hosted_synapses(U) \
	for ( vector<C_HostedSynapse*>::iterator U = hylist_begin(); U != hylist_end(); U++ )
#define for_all_standalone_neurons(U) \
	for ( vector<C_StandaloneNeuron*>::iterator U = snlist_begin(); U != snlist_end(); U++ )
#define for_all_standalone_synapses(U) \
	for ( vector<C_StandaloneSynapse*>::iterator U = sylist_begin(); U != sylist_end(); U++ )
#define for_all_ddtbound_neurons(U) \
	for ( vector<C_StandaloneNeuron*>::iterator U = dnlist_begin(); U != dnlist_end(); U++ )
#define for_all_ddtbound_synapses(U) \
	for ( vector<C_StandaloneSynapse*>::iterator U = dylist_begin(); U != dylist_end(); U++ )

#define for_all_readers(U) \
	for ( vector<C_BaseUnit*>::iterator U = reader_unit_list.begin(); U != reader_unit_list.end(); U++ )
#define for_all_schedulers(U) \
	for ( vector<C_BaseUnit*>::iterator U = scheduler_unit_list.begin(); U != scheduler_unit_list.end(); U++ )
#define for_all_units_with_param_ranges(U) \
	for ( vector<C_BaseUnit*>::iterator U = param_range_unit_list.begin(); U != param_range_unit_list.end(); U++ )

#define for_all_units_reversed(U) \
	for ( vector<C_BaseUnit*>::reverse_iterator U = ulist_rbegin(); U != ulist_rend(); U++ )
#define for_all_readers_reversed(U) \
	for ( vector<C_BaseUnit*>::reverse_iterator U = rlist_rbegin(); U != rlist_rend(); U++ )

#define for_all_hosted_neurons_reversed(U)   for ( vector<C_HostedNeuron*> ::reverse_iterator U = hosted_neu_list.rbegin(); U != hosted_neu_list.rend(); U++ )
#define for_all_hosted_synapses_reversed(U)  for ( vector<C_HostedSynapse*>::reverse_iterator U = hosted_syn_list.rbegin(); U != hosted_syn_list.rend(); U++ )
#define for_all_standalone_synapses_reversed(U)  for ( vector<C_StandaloneSynapse*>::reverse_iterator U = standalone_syn_list.rbegin(); U != standalone_syn_list.rend(); U++ )

#define for_all_listening_units(U) \
	for ( vector<C_BaseUnit*>::iterator U = lulist_begin(); U != lulist_end(); U++ )
#define for_all_conscious_neurons(N) \
	for ( vector<C_BaseNeuron*>::iterator N = cnlist_begin(); N != cnlist_end(); N++ )
#define for_all_spikelogging_neurons(N)	\
	for ( vector<C_BaseNeuron*>::iterator N = knlist_begin(); N != knlist_end(); N++ )
#define for_all_mx_synapses(N)		for ( vector<C_BaseSynapse*>::iterator Y = mxylist_begin(); Y != mxylist_end(); Y++ )


#define for_model_units(M,U)   for ( vector<C_BaseUnit*> ::iterator U = M->ulist_begin(); U != M->ulist_end(); U++ )
#define for_model_units_reversed(M,U)   for ( vector<C_BaseUnit*> ::reverse_iterator U = M->ulist_rbegin(); U != M->ulist_rend(); U++ )

#define for_model_hosted_neurons(M,U)   for ( vector<C_HostedNeuron*> ::iterator U = M->hnlist_begin(); U != M->hnlist_end(); U++ )
#define for_model_hosted_synapses(M,U)  for ( vector<C_HostedSynapse*>::iterator U = M->hylist_begin(); U != M->hylist_end(); U++ )
#define for_model_hosted_neurons_reversed(M,U)   for ( vector<C_HostedNeuron*> ::reverse_iterator U = M->hnlist_rbegin(); U != M->hnlist_rend(); U++ )
#define for_model_hosted_synapses_reversed(M,U)  for ( vector<C_HostedSynapse*>::reverse_iterator U = M->hylist_rbegin(); U != M->hylist_rend(); U++ )

#define for_model_standalone_neurons(M,U)   for ( vector<C_StandaloneNeuron*> ::iterator U = M->snlist_begin(); U != M->snlist_end(); U++ )
#define for_model_standalone_synapses(M,U)  for ( vector<C_StandaloneSynapse*>::iterator U = M->sylist_begin(); U != M->sylist_end(); U++ )
#define for_model_standalone_neurons_reversed(M,U)   for ( vector<C_StandaloneNeuron*> ::reverse_iterator U = M->snlist_rbegin(); U != M->snlist_rend(); U++ )
#define for_model_standalone_synapses_reversed(M,U)  for ( vector<C_StandaloneSynapse*>::reverse_iterator U = M->sylist_rbegin(); U != M->sylist_rend(); U++ )

#define for_model_neuron_units(M,U)  for_model_units(M,U) if ( (*U)->is_neuron() )
#define for_model_synapse_units(M,U)  for_model_units(M,U) if ( (*U)->is_synapse() )

#define for_model_spikelogging_neurons(M,N)	for ( vector<C_BaseNeuron*>::iterator N = M->knlist_begin(); N != M->knlist_end(); N++ )


// return values for import_NetworkML
#define CN_NMLIN_NOFILE		-1
#define CN_NMLIN_NOELEM		-2
#define CN_NMLIN_BADATTR	-3
#define CN_NMLIN_BADCELLTYPE	-4
#define CN_NMLIN_BIGLABEL	-5
#define CN_NMLIN_STRUCTERROR	-6






inline void
CIntegrateRK65::fixate()
{
	swap( model->V, model->W);
}


// various CUnit & CNeuron methods accessing CModel members
// that we want to have inline

inline const double&
C_BaseUnit::model_time()
{
	return M->model_time();
}

inline void
C_BaseUnit::pause_listening()
{
	if ( M )
		M->unregister_listener( this);
}

inline void
C_BaseUnit::resume_listening()
{
	if ( M )
		M->register_listener( this);
}





inline const char* const
C_BaseUnit::qlabelp()
{
	snprintf( __cn_qlabelp, CN_MAX_LABEL_SIZE-1,
		  "%-*s", M->longest_label(), label);
	return __cn_qlabelp;
}





inline SSpikeloggerService*
C_BaseNeuron::enable_spikelogging_service( int s_mask)
{
	if ( !_spikelogger_agent )
		_spikelogger_agent = new SSpikeloggerService( this, s_mask);
	M->register_spikelogger( this);
	return _spikelogger_agent;
}
inline SSpikeloggerService*
C_BaseNeuron::enable_spikelogging_service( double sample_period, double sigma, int s_mask)
{
	if ( !_spikelogger_agent )
		_spikelogger_agent = new SSpikeloggerService( this, sample_period, sigma, s_mask);
	M->register_spikelogger( this);
	return _spikelogger_agent;
}

inline void
C_BaseNeuron::disable_spikelogging_service()
{
	if ( _spikelogger_agent && !(_spikelogger_agent->_status & CN_KL_PERSIST)) {
		_spikelogger_agent->sync_history();
		M->unregister_spikelogger( this);

		delete _spikelogger_agent;
		_spikelogger_agent = NULL;
	}
}






inline void
C_HostedNeuron::reset_vars()
{
//	cout << "reset_vars() on " << label << " (idx " << idx << ")\n";
	if ( M && idx < M->_var_cnt )
		memcpy( &M->V[idx],
			__CNUDT[_type].stock_var_values,
			__CNUDT[_type].vno * sizeof(double));
}

inline double&
C_HostedNeuron::var_value( size_t v)
{
	return M->V[idx + v];
}

inline const double&
C_HostedNeuron::get_var_value( size_t v) const
{
	return M->V[idx + v];
}



inline unsigned
C_HostedConductanceBasedNeuron::n_spikes_in_last_dt()
{
	return E() >= M->spike_threshold;
}

inline unsigned
C_HostedRateBasedNeuron::n_spikes_in_last_dt()
{
	return round(E() * M->dt() * M->rng_sample());
}


inline unsigned
C_StandaloneConductanceBasedNeuron::n_spikes_in_last_dt()
{
	return E() >= M->spike_threshold;
}

inline unsigned
C_StandaloneRateBasedNeuron::n_spikes_in_last_dt()
{
	return round(E() * M->dt() * M->rng_sample());
}








inline void
C_HostedSynapse::reset_vars()
{
//	cout << "reset_vars() on " << label << " (idx " << idx << ")\n";
	if ( M && M->_var_cnt > idx )
		memcpy( &M->V[idx],
			__CNUDT[_type].stock_var_values,
			__CNUDT[_type].vno * sizeof(double));
}



inline double&
C_HostedSynapse::var_value( size_t v)
{
	return M->V[idx + v];
}

inline const double&
C_HostedSynapse::get_var_value( size_t v) const
{
	return M->V[idx + v];
}



inline double&
C_HostedConductanceBasedNeuron::E()
{
	return M->V[idx+0];
}

// F is computed on the fly, so far usually


inline double&
C_HostedSynapse::S()
{
	return M->V[idx+0];
}










inline void
CSynapseMap::preadvance()
{
//	printf( "fafa %s\n", label);
	V_next[0] = S() * exp( -M->discrete_dt() / P[_tau_])
		+ (_source->n_spikes_in_last_dt() ? P[_delta_] : 0);

//	V_next[1] = ;
}



inline void
CSynapseMxMap::preadvance()
{
	V_next[0] = S() * exp( -M->discrete_dt() / P[_tau_]) + q() * P[_delta_];
}


#endif

// EOF
