// ;-*-C++-*- *  Time-stamp: "2009-12-11 01:31:04 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *
 * License: GPL
 *
 * Initial version: 2008-09-02
 *
 * NeuroML import/export methods for CModel
 */

#include <string>
#include <iostream>
#include <boost/regex.hpp>

#include "model.h"


using namespace std;

#ifdef LIBXML_READER_ENABLED





int
CModel::import_NetworkML( const char *fname, bool appending)
{
	LIBXML_TEST_VERSION;

	xmlDoc *doc = xmlReadFile( fname, NULL, 0);
	if ( !doc )
		return CN_NMLIN_NOFILE;

	int retval = import_NetworkML( doc, fname, appending);

	xmlFreeDoc( doc);

	return retval;
}




static xmlNode* find_named_root_child_elem( xmlNode *node, const char *elem);

int
CModel::import_NetworkML( xmlDoc *doc, const char *fname, bool appending)
{
	int retval = 0;

	if ( !appending )
		reset();

	// we pass up on validation (for which we would need to keep a
	// .dtd or Schema at hand), and proceed to extracting elements

	xmlNode *root_node = xmlDocGetRootElement( doc),
		*n;

      // read meta:notes and make out a name for the model
	if ( !root_node ) {
		cerr << "Failed to obtain root element\n";
		retval = CN_NMLIN_NOELEM;
		goto out;
	}
	if ( !(n = find_named_root_child_elem( root_node, "notes")) )
		cerr << "<notes> element not found\n";
		// this is not critical, so just keep the user
		// informed and proceed
	else
		if ( n->type == XML_ELEMENT_NODE ) {  // only concern ourselves with nodes of this type
			xmlChar *notes_s = xmlNodeGetContent( n);
			// look for a substring specific to neuroConstruct, which is obviously speculative
			boost::regex pattern( ".*project: (\\w*).*");
			boost::cmatch found;
			name = boost::regex_match( (char*)notes_s, found, pattern)
				? name = found[1]
				: "(unmatched)";
			xmlFree( notes_s);  // cleaning up in our one-shot program may be overscrupulous,
					    // but let's be disciplined nonetheless
		} else
			name = "(unnamed)";

	if ( verbosely > 1 )
		printf( "Model \"%s\": %sing topology from %s\n",
			name.data(), (appending ?"Append" :"Import"), fname);

	// In the following calls to _process_{populations,instances}
	// functions, the actual order of appearance of these nodes in
	// the xml file doesn't matter, thanks to the xml contents
	// being already wholly read and available to us as a tree.

      // process <populations>
	if ( !(n = find_named_root_child_elem( root_node, "populations")) ) {
		retval = CN_NMLIN_NOELEM;
		goto out;
	} // assume there is only one <populations> element: don't loop to catch more
	if ( (retval = _process_populations( n->children)) < 0)	// note n->children, which is in fact a pointer to the first child
		goto out;

      // process <projections>
      // don't strictly require any projections as long as there are some neurons
	if ( (n = find_named_root_child_elem( root_node, "projections")) ) {
		if ( (retval = _process_projections( n->children)) < 0 )
			goto out;
	} else
		if ( verbosely > 2 )
			cout << "No projections found\n";

out:
	// we are done with topology; now put units' variables on a vector
	finalize_additions();
	// can call time_step only after finalize_additions

	cout << endl;

	return retval;
}


static xmlNode*
find_named_root_child_elem( xmlNode *node,     // node to start search from
			    const char *elem)  // name of the element searched for
{
	xmlNode *n;

	for ( n = node->children; n; n = n->next ) {
		if ( n->type == XML_ELEMENT_NODE ) {
			if ( xmlStrEqual( n->name, BAD_CAST elem) )
				return n;
// the <populations> and <projections> nodes are expected to appear as
// direct children of the root node; so don't go search deeper than that

//			if ( n->children ) { // go search deeper
//				ni = find_named_elem( n->children, elem);
//				if ( ni )
//					return ni;
//			}
		}
	}
	return NULL;
}






int
CModel::_process_populations( xmlNode *n)
{
	TUnitType cell_type = NT_VOID;
	int t;
	xmlChar *group_id_s, *cell_type_s;

	int	pop_cnt = 0;

	for ( ; n; n = n->next ) // if is NULL (parent had no children), we won't do a single loop
		if ( n->type == XML_ELEMENT_NODE && xmlStrEqual( n->name, BAD_CAST "population") ) {

			group_id_s = xmlGetProp( n, BAD_CAST "name"); // BAD_CAST is just a cast to xmlChar*
								      // with a catch that libxml functions
								      // expect strings pointed to to be good UTF
			if ( !group_id_s ) {
				cerr << "<population> element missing a \"name\" attribute near line " << n->line << endl;
				return CN_NMLIN_BADATTR;
			}
		      // probably having an unnamed popuation isn't an error so serious as to abort the
		      // operation, but discipline is above all

			cell_type_s = xmlGetProp( n, BAD_CAST "cell_type");
			// now we know the type of cells included in this population; remember it to pass on to
			// _process_population_instances, where it is used to select an appropriate unit type
			// when actually adding a neuron to the model

		      // but well, let's check if we have units of that type in stock
			for ( t = NT_FIRST; t <= NT_LAST; t++ )
				if ( xmlStrEqual( cell_type_s, BAD_CAST __CNUDT[t].id) ) {
					cell_type = static_cast<TUnitType>(t);
					break;
				}
			if ( t > NT_LAST ) {
				cerr << "Bad cell type (" << (char*)cell_type_s << ") in population " << group_id_s << endl;
				xmlFree( cell_type_s), xmlFree( group_id_s);
				return CN_NMLIN_BADCELLTYPE;
			}

			xmlNode *nin = n->children;  // again, ->children means ->first
			if ( nin )
				for ( ; nin; nin = nin->next )  // deal with multiple <instances> nodes
					if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "instances") ) {
						int subretval = _process_population_instances( nin->children,
											       group_id_s, cell_type);
						if ( subretval < 0 ) {
							xmlFree( group_id_s);
							return subretval;
						}
						if ( verbosely > 2 )
							printf( " %5d instance(s) of type \"%s\" in population \"%s\"\n",
								subretval, cell_type_s,  group_id_s);
						pop_cnt++;
					}
			xmlFree( cell_type_s);
			xmlFree( group_id_s);
		}

	if ( verbosely > 1 )
		cout << "\tTotal " << pop_cnt << " population(s)\n";
	return pop_cnt;
}






int
CModel::_process_projections( xmlNode *n)
{
	// much the same code as in _process_populations

	xmlChar *prj_name_s, *prj_src_s, *prj_tgt_s;

	size_t pop_cnt = 0;

	for ( ; n; n = n->next ) {
		if ( n->type != XML_ELEMENT_NODE || !xmlStrEqual( n->name, BAD_CAST "projection") )
			continue;

		prj_name_s = xmlGetProp( n, BAD_CAST "name");
		if ( !prj_name_s ) {
			cerr << "<projection> element missing a \"name\" attribute near line " << n->line << endl;
			return CN_NMLIN_BADATTR;
		}

		prj_src_s  = xmlGetProp( n, BAD_CAST "source");
		prj_tgt_s  = xmlGetProp( n, BAD_CAST "target");
		if ( !prj_src_s || !prj_tgt_s ) {
			cerr << "Projection \"" << prj_name_s << "\" missing a \"source\" and/or \"target\" attribute near line " << n->line << endl;
			xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
			return CN_NMLIN_BADATTR;
		}

		xmlNode *nin;
		nin = n->children;
		if ( !nin )
			cerr << "Empty <projection> node near line " << n->line << endl;

		xmlChar *synapse_type_s = NULL;
		TUnitType synapse_type = NT_VOID;
		int t;

		for ( ; nin; nin = nin->next )
			if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "synapse_props") ) {
				synapse_type_s = xmlGetProp( nin, BAD_CAST "synapse_type");
				for ( t = YT_FIRST; t <= YT_LAST; t++ )
					if ( xmlStrEqual( synapse_type_s, BAD_CAST __CNUDT[t].id) ) {
						synapse_type = static_cast<TUnitType>(t);
						break;
					}
				if ( t > YT_LAST ) {
					fprintf( stderr, "Bad synapse type \"%s\" near line %u\n",
						 (char*)synapse_type_s, nin->line);
					xmlFree( synapse_type_s), xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
					return CN_NMLIN_BADCELLTYPE;
				}
			}

		for ( nin = n->children; nin; nin = nin->next )
			if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "connections") ) {
				int subretval = _process_projection_connections( nin->children,
										 prj_name_s, synapse_type,
										 prj_src_s, prj_tgt_s);
				if ( subretval < 0 ) {
					xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
					return subretval;
				}
				if ( verbosely > 2 )
					printf( " %5d connection(s) of type \"%s\" in projection \"%s\"\n",
						subretval, synapse_type_s,  prj_name_s);
				pop_cnt++;
			}
		xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
	}

	if ( verbosely > 1 )
		printf( "\tTotal %zd projection(s)\n", pop_cnt);

	return (int)pop_cnt;
}







int
CModel::_process_population_instances( xmlNode *n, const xmlChar *group_prefix, TUnitType type)
{
	int	retval = 0;  // also keeps a count of added neurons

	double	x, y, z;
	char	cell_id[CN_MAX_LABEL_SIZE];

	xmlNode *nin;

	for ( ; n; n = n->next ) {
		if ( n->type != XML_ELEMENT_NODE || !xmlStrEqual( n->name, BAD_CAST "instance") )
			continue;

		xmlChar *id_s = xmlGetProp( n, BAD_CAST "id");
		if ( !id_s ) {
		      // could be less strict here and allow empty ids, which will then be composed
		      // from group_prefix + id (say, "LN0", "LN1" and so on); but then, as
		      // individual <projection>s would have to reference both endpoints by explicit
		      // ids, it is obviously prone to error to have <instance> ids depend solely on
		      // their order of appearance.
		      // So we bark at empty ids.
			cerr << "<instance> element without an \"id\" attribute near line " << n->line << endl;
			retval = CN_NMLIN_BADATTR;
			goto out;
		}

		size_t total_len = xmlStrlen( group_prefix) + xmlStrlen( id_s);
		if ( total_len >= CN_MAX_LABEL_SIZE ) {
			cerr << "<instance> combined label (" << group_prefix << id_s << ") exceeding " << CN_MAX_LABEL_SIZE << " near line " << n->line << endl;
			xmlFree( id_s);
			retval = CN_NMLIN_BIGLABEL;
			goto out;
		}
		_longest_label = max( _longest_label,
				      (short unsigned)snprintf( cell_id, CN_MAX_LABEL_SIZE-1, "%s.%s",
								group_prefix, id_s));  // here, a new instance is given a name

		xmlFree( id_s);

		if ( !(nin = n->children) )
			goto out;

		for ( ; nin; nin = nin->next ) {
			if ( !(nin->type == XML_ELEMENT_NODE &&
			       xmlStrEqual( nin->name, BAD_CAST "location")) )
				continue;

			xmlChar *x_s = xmlGetProp( nin, BAD_CAST "x"),
				*y_s = xmlGetProp( nin, BAD_CAST "y"),
			*z_s = xmlGetProp( nin, BAD_CAST "z");
		      // here we do actually insert neurons into the model
			if ( !(x_s && y_s && z_s) ) {
				cerr << "<location> element missing full coordinates near line " << nin->line << endl;
				retval = CN_NMLIN_BADATTR;
			} else {
				x = strtod( (char*)x_s, NULL), y = strtod( (char*)y_s, NULL), z = strtod( (char*)z_s, NULL);
				int	add_retval = -1;
				C_BaseNeuron *n = NULL;
				switch ( type ) {
				case NT_HH_D:
					n = new CNeuronHH_d( cell_id, x, y, z, this, CN_UOWNED, false);
				    break;
				case NT_HH_R:
					n = new CNeuronHH_r( cell_id, x, y, z, this, CN_UOWNED);
				    break;

				case NT_HH2_D:
					n = new CNeuronHH2_d( cell_id, x, y, z, this, CN_UOWNED, false);
				    break;
				// case NT_HH2_R:
				// 	n = new CNeuronHH2_r( cell_id, x, y, z, this, CN_UOWNED, false);
				//     break;
#ifdef CN_WANT_MORE_NEURONS
				case NT_EC_D:
					n = new CNeuronEC_d( cell_id, x, y, z, this, CN_UOWNED, false);
				    break;
				case NT_ECA_D:
					n = new CNeuronECA_d( cell_id, x, y, z, this, CN_UOWNED, false);
				    break;
				case NT_LV:
					n = new COscillatorLV( cell_id, x, y, z, this, CN_UOWNED, false);
				    break;
				case NT_COLPITTS:
					n = new COscillatorColpitts( cell_id, x, y, z, this, CN_UOWNED, false);
				    break;
				case NT_VDPOL:
					n = new COscillatorVdPol( cell_id, x, y, z, this, CN_UOWNED, false);
				    break;
#endif
				case NT_DOTPOISSON:
					n = new COscillatorDotPoisson( cell_id, x, y, z, this, CN_UOWNED);
				    break;
				case NT_POISSON:
					n = new COscillatorPoisson( cell_id, x, y, z, this, CN_UOWNED);
				    break;

				case NT_DOTPULSE:
					n = new CNeuronDotPulse( cell_id, x, y, z, this, CN_UOWNED);
				    break;

				case NT_MAP:
					n = new CNeuronMap( cell_id, x, y, z, this, CN_UOWNED);
				    break;
				default:
				    break;
			      // don't care about a default case as the validity of `type'
			      // has been checked by _process_populations
				}

				if ( !n /*|| !n->is_enabled() */) {
					if ( n )
						delete n;
					add_retval = -1;
					cerr << "Failed to add a neuron " << cell_id << endl;
					retval = CN_NMLIN_STRUCTERROR;
				} else {
					n->serial_id = _global_unit_id_reservoir++;
					retval++;
				}
			}
			xmlFree( x_s), xmlFree( y_s), xmlFree( z_s);

			if ( retval < 0 )
				goto out;
		}
	}
out:
	return retval;
}




int
CModel::_process_projection_connections( xmlNode *n,
					 const xmlChar *synapse_name, TUnitType type,
					 const xmlChar *src_grp_prefix, const xmlChar *tgt_grp_prefix)
{
	// similar to _process_population_instances, except that we read some more attributes (source and
	// target units)

	int	retval = 0;  // is also a counter of synapses
	char	synapse_id [CN_MAX_LABEL_SIZE],
		src_cell_id[CN_MAX_LABEL_SIZE],
		tgt_cell_id[CN_MAX_LABEL_SIZE];

	C_BaseNeuron *src, *tgt;

	for ( ; n; n = n->next ) {
		if ( n->type != XML_ELEMENT_NODE || !xmlStrEqual( n->name, BAD_CAST "connection") )
			continue;

		C_BaseSynapse *y = NULL;

		xmlChar *synapse_id_s  = xmlGetProp( n, BAD_CAST "id"),
			*src_cell_id_s = xmlGetProp( n, BAD_CAST "pre_cell_id"),
			*tgt_cell_id_s = xmlGetProp( n, BAD_CAST "post_cell_id");
		if ( !synapse_id_s || !src_cell_id_s || !tgt_cell_id_s ) {
			cerr << "A <connection> element without \"id\", \"pre_cell_id\" and/or \"post_cell_id\" attribute near line " << n->line << endl;
			retval = CN_NMLIN_BADATTR;
			goto out;
		}

		memset( src_cell_id, 0, CN_MAX_LABEL_SIZE);
		snprintf( src_cell_id, CN_MAX_LABEL_SIZE-1, "%s.%s", src_grp_prefix, src_cell_id_s);
		src = const_cast<C_BaseNeuron*>(neuron_by_label( src_cell_id));

		memset( tgt_cell_id, 0, CN_MAX_LABEL_SIZE);
		snprintf( tgt_cell_id, CN_MAX_LABEL_SIZE-1, "%s.%s", tgt_grp_prefix, tgt_cell_id_s);
		tgt = const_cast<C_BaseNeuron*>(neuron_by_label( tgt_cell_id));

		if ( xmlStrlen( synapse_name) + xmlStrlen( synapse_id_s) >= CN_MAX_LABEL_SIZE ) {
			fprintf( stderr, "Synapse combined label (\"%s%s\") exceeding %u chars near line %u\n",
				 synapse_name, synapse_id_s, CN_MAX_LABEL_SIZE, n->line);
			retval = CN_NMLIN_BIGLABEL;
			goto out;
		}
		memset( synapse_id, 0, CN_MAX_LABEL_SIZE);
		_longest_label = max( _longest_label,
				      (unsigned short)snprintf( synapse_id, CN_MAX_LABEL_SIZE-1,
								"%s.%s", synapse_name, synapse_id_s));

		if ( !src || !tgt ) {
			fprintf( stderr, "Phoney source (%s) or target (%s) for connection near line %u\n",
				 src_cell_id, tgt_cell_id, n->line);
			retval = CN_NMLIN_BADATTR;
			goto out;
		}

		switch ( type ) {
	      // the __CNUDT entry at first TUnitType element whose
	      // 'name' matches the type id supplied, captures all cases for a given synapse family
		case YT_AB_DD:
			if ( src->traits() & UT_RATEBASED && tgt->traits() & UT_RATEBASED )
				y = new CSynapseAB_rr( synapse_id, src, tgt, this, CN_UOWNED, false);
			else if ( src->traits() & UT_RATEBASED && !(tgt->traits() & UT_RATEBASED) )
				// y = new CSynapseAB_rd( synapse_id, src, tgt, this, CN_UOWNED, false);
				fprintf( stderr, "AB_rd not implemented\n");
			else if ( !(src->traits() & UT_RATEBASED) && tgt->traits() & UT_RATEBASED )
				if ( src->traits() & UT_DOT )
					y = new CSynapseMxAB_dr( synapse_id, src, tgt, this, CN_UOWNED, false);
				else
					// y = new CSynapseAB_dr( synapse_id, src, tgt, this, CN_UOWNED, false);
					fprintf( stderr, "AB_dr not implemented\n");
			else
				if ( src->traits() & UT_DOT )
					y = new CSynapseMxAB_dd( synapse_id, src, tgt, this, CN_UOWNED, false);
				else
					y = new CSynapseAB_dd( synapse_id, src, tgt, this, CN_UOWNED, false);
		    break;

		case YT_ABMINS_DD:
			if ( src->traits() & UT_RATEBASED && tgt->traits() & UT_RATEBASED )
				// y = new CSynapseABMinus_rr( synapse_id, src, tgt, this, CN_UOWNED, false);
				fprintf( stderr, "ABMinus_rr not implemented\n");
			else if ( src->traits() & UT_RATEBASED && !(tgt->traits() & UT_RATEBASED) )
				// y = new CSynapseABMinus_rd( synapse_id, src, tgt, this, CN_UOWNED, false);
				fprintf( stderr, "ABMinus_rd not implemented\n");
			else if ( !(src->traits() & UT_RATEBASED) && tgt->traits() & UT_RATEBASED )
				// y = new CSynapseABMinus_dr( synapse_id, src, tgt, this, CN_UOWNED, false);
				fprintf( stderr, "ABMinus_dr and MxABMinus_dr not implemented\n");
			else
				y = new CSynapseABMinus_dd( synapse_id, src, tgt, this, CN_UOWNED, false);
		    break;

		case YT_RALL_DD:
			if ( src->traits() & UT_RATEBASED && tgt->traits() & UT_RATEBASED )
				// y = new CSynapseRall_rr( synapse_id, src, tgt, this, CN_UOWNED, false);
				fprintf( stderr, "Rall_rr not implemented\n");
			else if ( src->traits() & UT_RATEBASED && !(tgt->traits() & UT_RATEBASED) )
				// y = new CSynapseRall_rd( synapse_id, src, tgt, this, CN_UOWNED, false);
				fprintf( stderr, "Rall_rd not implemented\n");
			else if ( !(src->traits() & UT_RATEBASED) && tgt->traits() & UT_RATEBASED )
				// y = new CSynapseRall_dr( synapse_id, src, tgt, this, CN_UOWNED, false);
				fprintf( stderr, "Rall_dr and MxRall_dr not implemented\n");
			else
				y = new CSynapseRall_dd( synapse_id, src, tgt, this, CN_UOWNED, false);
		    break;

		case YT_MAP:
			if ( src->traits() & UT_DOT )
				y = new CSynapseMxMap( synapse_id, src, tgt, this, CN_UOWNED);
			else
				y = new CSynapseMap( synapse_id, src, tgt, this, CN_UOWNED);
		    break;
		default:
		    break;
		}
		if ( !y /*|| !y->is_enabled() */) {
			if ( y )
				delete y;
			fprintf( stderr, "Failed to add synapse \"%s\"\n", synapse_id);
			retval = CN_NMLIN_STRUCTERROR;
			goto out;
		} else
			retval++;


		xmlFree( synapse_id_s), xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s);
		continue;
	out:
		xmlFree( synapse_id_s), xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s);
		break;

	}

	return retval;
}



int
CModel::export_NetworkML( const char *fname)
{
	int retval = 0;

	LIBXML_TEST_VERSION;

	fprintf( stderr, "export_NetworkML() not implemented yet\n");

	return retval;
}







#else
#error Need an XMLREADER-enabled libxml2 (>2.6)

#endif // LIBXML_READER_ENABLED
// EOF
