// ;-*-C++-*- *  Time-stamp: "2010-01-21 21:30:50 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny
 *
 * License: GPL
 *
 * Initial version: 2008-09-02
 *
 * CModel household
 */

#include <sys/time.h>
#include <iostream>
#include <algorithm>

#include <boost/regex.hpp>
#include "model.h"

using namespace std;



CModel::CModel( const char *inname, CIntegrate_base *inintegrator, int instatus)
      : name (inname),
	_status (instatus | CN_MDL_NOTREADY),
	_global_unit_id_reservoir (0l),
	_longest_label (1),
	_var_cnt (1),			// reserve [0] for model_time
	_cycle (0),
	_discrete_time (0.),  _discrete_dt (1.),
	spike_threshold (0.),
	spike_lapse (5.),
	logging_dt (0),
	_dt_logger (NULL),
	_spike_logger (NULL),	// open these streams at first write instead in prepare_advance()
	verbosely (1)
{
	V.resize( _var_cnt), W.resize( _var_cnt);
	V[0] = 0.;

	(_integrator = inintegrator) -> model = this;

	{
		struct timeval tp = { 0L, 0L };
		gettimeofday( &tp, NULL);

#ifdef USE_GSL_RNG
		const gsl_rng_type * T;
		gsl_rng_env_setup();
		T = gsl_rng_default;
		if ( gsl_rng_default_seed == 0 )
			gsl_rng_default_seed = tp.tv_usec;
		_rng = gsl_rng_alloc( T);

#else
		Random::Set( (double)tp.tv_usec / 1e6);
#endif
//		Random::Set( 3);
	}
}


CModel::~CModel()
{
	if ( verbosely > 4 )
		fprintf( stdout, "Deleting all units...\n");
	for_all_units_reversed (U)
		if ( (*U)->is_owned() )
			delete (*U);

	if ( _integrator->is_owned )
		delete _integrator;

	if ( _dt_logger )
		delete _dt_logger;
	if ( _spike_logger )
		delete _spike_logger;

#ifdef USE_GSL_RNG
	gsl_rng_free( _rng);
#endif
}


void
CModel::reset( bool also_reset_params)
{
	_cycle = 0, V[0] = 0.;

	_integrator->dt = _integrator->_dt_min;

	reset_state_all_units();
	if ( also_reset_params )
		for_all_units (U)
			(*U)->reset_params();

	scheduler_update_periods.clear();
	scheduler_update_periods_last_checked.clear();

	if ( _status & CN_MDL_LOGDT ) {
		if ( _dt_logger )
			delete _dt_logger;
		string	fname = name + ".dtlog";
		_dt_logger = new ofstream( fname.data());
	}
	if ( _status & CN_MDL_LOGSPIKERS ) {
		if ( _spike_logger )
			delete _spike_logger;
		string	fname = name + ".spikes";
		_spike_logger = new ofstream( fname.data());
	}
}





const double&
CModel::dt( const double& new_dt)
{
	if ( _integrator ) {
		if ( _integrator->_dt_min < new_dt )
			_integrator->dt = _integrator->_dt_min;
		else if ( _integrator->_dt_max > new_dt )
			_integrator->dt = _integrator->_dt_max;
		else
			_integrator->dt = new_dt;
	}
	return _integrator->dt;
}

const double&
CModel::discrete_dt( const double& new_ddt)
{
      // set ddt as requested, but make sure the integration dt
      // will not reach discrete_dt, and by a good margin
	if ( _integrator && _integrator->_dt_max > new_ddt / 5 )
		_integrator->_dt_max = new_ddt / 5;
	return _discrete_dt = new_ddt;
}




const C_BaseUnit*
CModel::unit_by_label( const char *inlabel) const
{
	for ( vector<C_BaseUnit*>::const_iterator U = unit_list.begin(); U != unit_list.end(); U++ )
		if ( strcmp( (*U)->label, inlabel) == 0 )
			return *U;
	return NULL;
}

const C_BaseNeuron*
CModel::neuron_by_label( const char *inlabel) const
{
	for ( vector<C_BaseUnit*>::const_iterator U = unit_list.begin(); U != unit_list.end(); U++ )
		if ( (*U)->is_neuron() && strcmp( (*U)->label, inlabel) == 0 )
			return static_cast<C_BaseNeuron*>(*U);
	return NULL;
}

const C_BaseSynapse*
CModel::synapse_by_label( const char *inlabel) const
{
	for ( vector<C_BaseUnit*>::const_iterator U = unit_list.begin(); U != unit_list.end(); U++ )
		if ( (*U)->is_synapse() && strcmp( (*U)->label, inlabel) == 0 )
			return static_cast<C_BaseSynapse*>(*U);
	return NULL;
}





// ----- adding units
void
CModel::_include_base_unit( C_BaseUnit* u)
{
	for_all_units (U)
		if ( (*U) == u ) {
			fprintf( stderr, "Unit %s found already included in model %s\n", u->label, name.c_str());
			goto skip_ul_pushback;
		}
	unit_list.push_back( u);
skip_ul_pushback:

	if ( verbosely > 5 )
		fprintf( stdout, "  registered base unit %s\n", u->label);

	if ( u->param_readers )
		register_reader( u);
	if ( u->param_ranges )
		register_param_range_unit( u);

	if ( u->is_listening() ) {
		for_all_listening_units (U)
			if ( (*U) == u ) {
				cerr << "Unit " << u->label << " already on listening list\n";
				goto skip_lisn_reg;
			}
		lisn_unit_list.push_back( u);
	}
skip_lisn_reg:

	u->M = this;

	u->serial_id = _global_unit_id_reservoir++;
}


int
CModel::_check_new_synapse( C_BaseSynapse* y)
{
	bool	src_found = false,
		tgt_found = false;
	for_all_neurons (N) {
		if      ( *N == y->_source )
			src_found = true;
		else if ( *N == y->_target )
			tgt_found = true;
		if ( src_found && tgt_found )
			break;
	}

	if ( verbosely > 0 ) {
		if ( !src_found )
			fprintf( stderr, "Warning: For synapse \"%s\", source not found in model \"%s\"\n",
				 y->label, name.c_str());
		if ( !tgt_found )
			fprintf( stderr, "Warning: For synapse \"%s\", target not found in model \"%s\"\n",
				 y->label, name.c_str());
	}

	if ( y->traits() & UT_MULTIPLEXING || y->_source->traits() & UT_DOT ) {
		if ( !(y->traits() & UT_MULTIPLEXING) ) {
			fprintf( stderr, "Dot unit \"%s\" can only have a Multiplexing synapse as axon\n",
				 y->source()->label);
			return -1;
		}
		if ( !(y->_source->traits() & UT_DOT) ) {
			fprintf( stderr, "Multiplexing synapse \"%s\" can only have a Dot unit as source\n",
				 y->label);
			return -1;
		}
	}

	if ( y->traits() & UT_RATEBASED &&
	    (!(y->_target->traits() & UT_RATEBASED)) ) {
		fprintf( stderr,
			 "Rate-based synapse \"%s\" connects to \"%s\",\n"
			 "which is not of a rate-based kind as it should\n",
			y->label, y->_target->label);
		return -1;
	}

	return 0;
}


int
CModel::include_unit( C_HostedNeuron *u, bool is_last)
{
	_include_base_unit( u);

	u->idx = _var_cnt;
	_var_cnt += u->v_no();

	hosted_neu_list.push_back( u);

	// if ( u->_spikelogger_agent  &&  !(u->_spikelogger_agent->_status & CN_KL_IDLE) )
	// 	spikelogging_neu_list.push_back( u);

	if ( u->is_conscious() )
		conscious_neu_list.push_back( u);

	if ( is_last )
		finalize_additions();

	return 0;
}

int
CModel::include_unit( C_HostedSynapse *u, bool is_last)
{
	if ( _check_new_synapse( u) ) {
//		u->enable( false);
		u->M = NULL;
		return -1;
	}

	_include_base_unit( u);

	u->idx = _var_cnt;
	_var_cnt += u->v_no();

	hosted_syn_list.push_back( u);

	if ( u->traits() & UT_MULTIPLEXING )
		mx_syn_list.push_back( u);

	if ( is_last )
		finalize_additions();

	return 0;
}



int
CModel::include_unit( C_StandaloneNeuron *u)
{
	_include_base_unit( u);

	// if ( u->_spikelogger_agent  &&  !(u->_spikelogger_agent->_status & CN_KL_IDLE) )
	// 	spikelogging_neu_list.push_back( u);

	if ( u->is_conscious() )
		conscious_neu_list.push_back( u);

	if ( u->is_ddtbound() )
		ddtbound_neu_list.push_back( u);
	else
		standalone_neu_list.push_back( u);

	return 0;
}


int
CModel::include_unit( C_StandaloneSynapse *u)
{
	if ( _check_new_synapse( u) ) {
//		u->enable( false);
		u->M = NULL;
		return -1;
	}

	_include_base_unit( u);

	if ( u->is_ddtbound() )
		ddtbound_syn_list.push_back( u);
	else
		standalone_syn_list.push_back( u);

	if ( u->traits() & UT_MULTIPLEXING )
		mx_syn_list.push_back( u);

	return 0;
}



// preserve the unit if !do_delete, so it can be re-included again
C_BaseUnit*
CModel::exclude_unit( C_BaseUnit *u, bool do_delete)
{
	for_all_units (U)
		if ( *U == u ) {

			if ( __cn_verbosely > 5 )
				fprintf( stderr, "-excluding unit \"%s\"\n", u->label);

			if ( u->param_readers )
				unregister_reader( u);

			if ( u->param_ranges )
				unregister_param_range_unit( u);

			if ( u->is_listening() )
				u->stop_listening();  // also calls unregister_listener

			if ( u->traits() & UT_MULTIPLEXING )
				for_all_mx_synapses (Y)
					if ( (*Y) == u ) {
						mx_syn_list.erase( Y);
						break;
					}

			if ( u->is_conscious() )
				for_all_conscious_neurons (N)
					if ( *N == u ) {
						conscious_neu_list.erase( N);
						break;
					}

			if ( u->is_hostable() ) {
				if ( u->is_neuron() )
					for_all_hosted_neurons (N) {
						if ( *N == u ) {
							hosted_neu_list.erase( N);
							break;
						}
					}
				else
					for_all_hosted_synapses (Y) {
						if ( *Y == u ) {
							hosted_syn_list.erase( Y);
							break;
						}
					}
			} else if ( u->is_ddtbound() ) {
				if ( u->is_neuron() )
					for_all_ddtbound_neurons (N) {
						if ( *N == u ) {
							ddtbound_neu_list.erase( N);
							break;
						}
					}
				else
					for_all_ddtbound_synapses (Y) {
						if ( *Y == u ) {
							ddtbound_syn_list.erase( Y);
							break;
						}
					}
			} else {  // !is_hostable && !is_ddtbound
				if ( u->is_neuron() )
					for_all_standalone_neurons (N) {
						if ( *N == u ) {
							standalone_neu_list.erase( N);
							break;
						}
					}
				else
					for_all_standalone_synapses (Y) {
						if ( *Y == u ) {
							standalone_syn_list.erase( Y);
							break;
						}
					}
			}

			unit_list.erase( U);

			if ( do_delete ) {
				delete u;
				return NULL;
			} else {
				u->M = NULL;
//				_status &= ~CN_UENABLED;
				return u;
			}
		}

	return u;
}



void
CModel::register_listener( C_BaseUnit *u)
{
	for_all_listening_units (U)
		if ( (*U) == u )
			return;
	lisn_unit_list.push_back( u);
}

void
CModel::unregister_listener( C_BaseUnit *u)
{
	for_all_listening_units (U)
		if ( (*U) == u ) {
			lisn_unit_list.erase( U);
			return;
		}
}





void
CModel::register_reader( C_BaseUnit *u)
{
      // a unit reading more than one param will have multiple entries
      // in this list
	reader_unit_list.push_back( u);
}

void
CModel::unregister_reader( C_BaseUnit *u)
{
      // a unit reading more than one param will have multiple entries in this list
start_over:
	for_all_readers (U)
		if ( (*U) == u ) {
			reader_unit_list.erase( U);
			if ( verbosely > 5 )
				cout << " (removed " << u->label << " instance from reader list)\n";
			goto start_over;
		}
}

void
CModel::register_scheduler( C_BaseUnit *u)
{
	scheduler_unit_list.push_back( u);
}

void
CModel::unregister_scheduler( C_BaseUnit *u)
{
start_over:
	for_all_schedulers (U)
		if ( (*U) == u ) {
			scheduler_unit_list.erase( U);
			if ( verbosely > 5 )
				cout << " (removed " << u->label << " instance from scheduler list)\n";
			goto start_over;
		}
}



void
CModel::register_param_range_unit( C_BaseUnit *u)
{
	for ( vector<C_BaseUnit*>::iterator U = param_range_unit_list.begin();
	      U != param_range_unit_list.end(); U++ )
		if ( *U == u )
			return;
	param_range_unit_list.push_back( u);
}


void
CModel::unregister_param_range_unit( C_BaseUnit *u)
{
	for_all_units_with_param_ranges (U)
		if ( (*U) == u ) {
			param_range_unit_list.erase( U);
			return;
		}
}







void
CModel::register_spikelogger( C_BaseNeuron *n)
{
	for_all_spikelogging_neurons (N)
		if ( (*N) == n ) {
			cerr << "Neuron " << n->label << " already on spikelogger list\n";
			return;
		}
	spikelogging_neu_list.push_back( n);
}

void
CModel::unregister_spikelogger( C_BaseNeuron *n)
{
	for_all_spikelogging_neurons (N)
		if ( (*N) == n ) {
			spikelogging_neu_list.erase( N);
			if ( verbosely > 5 )
				cout << " (removed " << n->label << " from spikelogger list)\n";
			return;
		}
	fprintf( stderr, "Neuron \"%s\" was not on spikelogger list\n", n->label);
}





void
CModel::finalize_additions()
{
	V.resize( _var_cnt),  W.resize( _var_cnt);

	for_all_hosted_neurons (N)
		(*N) -> reset_vars();
	for_all_hosted_synapses (Y)
		(*Y) -> reset_vars();

	if ( _status & CN_MDL_SORTUNITS ) {
		qsort( unit_list.data(), unit_list.size(), sizeof(C_BaseUnit*),  __unit_compare);
		qsort( hosted_neu_list.data(), hosted_neu_list.size(), sizeof(C_HostedNeuron*),  __unit_compare);
		qsort( hosted_syn_list.data(), hosted_syn_list.size(), sizeof(C_HostedSynapse*), __unit_compare);
		qsort( standalone_neu_list.data(), standalone_neu_list.size(), sizeof(C_StandaloneNeuron*),  __unit_compare);
		qsort( standalone_syn_list.data(), standalone_syn_list.size(), sizeof(C_StandaloneSynapse*), __unit_compare);
	}

	_integrator->prepare();
}







template <class T>
static bool
device_list_concerns_parm( vector<T> *dv, unsigned short whom)
{
	typename vector<T>::iterator D;
	D = dv->begin();
	for ( ; D != dv->end(); D++ )
		if ( D->assignee_kind == C_BaseUnit::RDA_PARAM && D->idx == whom )
			return true;
	return false;
}

void
CModel::cull_deaf_synapses()
{
	size_t cnt = 0;
      // 1. Need to traverse syn_list backwards due to shifts its vector will undergo on element deletions;
      // 2. Omit those with a param reader, scheduler or range, but only if it is connected to parameter "gsyn"
	for_all_synapses_reversed (Y) {
		int gsyn_pidx = (*Y) -> param_idx_by_sym( "gsyn");
		if ( ((*Y)->param_schedulers && device_list_concerns_parm( (*Y)->param_schedulers, gsyn_pidx)) ||
		     ((*Y)->param_readers    && device_list_concerns_parm( (*Y)->param_readers,    gsyn_pidx)) ||
		     ((*Y)->param_ranges     && device_list_concerns_parm( (*Y)->param_ranges,     gsyn_pidx)) ) {
			if ( verbosely > 2 )
				printf( " (preserving doped synapse with zero gsyn: \"%s\")\n", (*Y)->label);
			continue;
		}
		if ( gsyn_pidx > -1 && (*Y)->param_value( gsyn_pidx) == 0. ) {
			if ( verbosely > 2 )
				printf( " (deleting synapse with zero gsyn: \"%s\")\n", (*Y)->label);
			delete (*Y);
			cnt++;
		}
	}
	if ( verbosely > 0 && cnt )
		printf( "Deleted %zd deaf synapses\n", cnt);
}



void
CModel::cull_blind_synapses()
{
	for_all_hosted_synapses_reversed (Y)
		if ( (*Y)->_source == NULL && !(*Y)->is_doped() ) {
			if ( verbosely > 3 )
				printf( " (deleting synapse with NULL source: \"%s\")\n", (*Y)->label);
			delete (*Y);
		}
	for_all_standalone_synapses_reversed (Y)
		if ( (*Y)->_source == NULL && !(*Y)->is_doped() ) {
			if ( verbosely > 3 )
				printf( " (deleting synapse with NULL source: \"%s\")\n", (*Y)->label);
			delete (*Y);
		}
}



void
CModel::reset_state_all_units()
{
	for_all_units (U)
		(*U) -> reset_state();
}







// tags

int
CModel::process_listener_tags( vector<string> &Listeners, int lbits)
{
	for ( vector<string>::iterator P = Listeners.begin(); P != Listeners.end(); P++ ) {
		boost::regex pattern( (*P).data());
		boost::cmatch found;
		for_all_units (U) {
			if ( boost::regex_match( (*U)->label, found, pattern) ) {
				(*U) -> start_listening( lbits);
				if ( verbosely > 3 )
					printf( " (unit \"%s\" listening%s)\n",
						(*U)->label, lbits & CN_ULISTENING_1VARONLY ? ", to one var only" :"");
			}
		}
	}

	return 0;
}


int
CModel::process_spikelogger_tags( vector<string> &Spikeloggers, double sxf_sample, double sdf_sigma)
{
	bool log_sdf = !(sxf_sample == 0. || sdf_sigma == 0.);

	for ( vector<string>::iterator P = Spikeloggers.begin(); P != Spikeloggers.end(); P++ ) {
		boost::regex pattern( P->c_str());
		boost::cmatch found;
		for_all_standalone_neurons (N) {
			if ( boost::regex_match( (*N)->label, found, pattern) ) {
				if ( ( log_sdf && !(*N)->enable_spikelogging_service( sxf_sample, sdf_sigma))
				     ||
				     (!log_sdf && !(*N)->enable_spikelogging_service()) ) {
					fprintf( stderr, "Cannot have \"%s\" log spikes because it is not a conductance-based neuron (of type %s)\n",
						 (*N)->label, (*N)->type_name());
					return -1;
				}
				if ( verbosely > 3 )
					printf( " (enabling spike logging for standalone neuron \"%s\")\n", (*N)->label);
			}
		}
		for_all_hosted_neurons (N) {
			if ( boost::regex_match( (*N)->label, found, pattern) ) {
				if ( ( log_sdf &&
				      !(*N)->enable_spikelogging_service( sxf_sample, sdf_sigma))
				     ||
				     (!log_sdf &&
				      !(*N)->enable_spikelogging_service()) ) {
					fprintf( stderr, "Cannot have \"%s\" log spikes because it is not a conductance-based neuron (of type %s)\n",
						 (*N)->label, (*N)->type_name());
					return -1;
				}
				if ( verbosely > 3 )
					printf( " (enabling spike logging for hosted neuron \"%s\")\n", (*N)->label);
			}
		}
	}

	return 0;
}


int
CModel::process_putout_tags( vector<string> &ToRemove)
{
      // execute some
	for ( vector<string>::iterator P = ToRemove.begin(); P != ToRemove.end(); P++ ) {
		boost::regex pattern( P->c_str());
		boost::cmatch found;
		for_all_units (U) {
			if ( boost::regex_match( (*U)->label, found, pattern) ) {
				if ( verbosely > 2 )
					printf( " (put out unit \"%s\")\n",
						(*U)->label);
				delete (*U);
				if ( units() > 0 )
					U = ulist_begin();
				else
					break;
			}
		}
	}

//	cull_blind_synapses();
	// don't call cull_blind_synapses before process_spikelogger_tags and process_listener_tags have
	// completed because Spikeloggers and Listeners will then point to deleted objects

	return 0;
}


int
CModel::process_decimate_tags( vector<SDecimationGroup> &ToDecimate)
{
      // decimate others
	size_t n_executed = 0;
	for ( vector<SDecimationGroup>::iterator P = ToDecimate.begin(); P != ToDecimate.end(); P++ ) {
		boost::regex pattern( P->name.c_str());
		boost::cmatch found;

	      // collect group
		vector<C_BaseUnit*> dcmgroup;
		for_all_units (U)
			if ( boost::regex_match( (*U)->label, found, pattern) )
				dcmgroup.push_back( *U);
	      // execute
		unsigned to_execute = rint( dcmgroup.size() * P->fraction),
			 u;
		while ( n_executed < to_execute ) {
			u = rng_sample() * (dcmgroup.size()-1);  // cout << "u = " << u << "/" << dcmgroup.size() << endl;
			delete dcmgroup[u];
			unsigned uu = 0;
			for ( vector<C_BaseUnit*>::iterator I = dcmgroup.begin(); I != dcmgroup.end(); I++ )
				if ( uu++ == u )
					dcmgroup.erase( I);
			n_executed++;
		}
		if ( verbosely > 3 )
			printf( " (decimated %4.1f%% (%zu units) of %s)\n", P->fraction*100, n_executed, P->name.c_str());

	}

	return 0;
}














static inline const char*
pl_ending( size_t cnt)
{
	return cnt == 1 ? "" : "s";
}

void
CModel::dump_metrics()
{
	printf( "\nModel \"%s\"%s:\n"
		"  %5zd unit%s total (%zd Neuron%s, %zd Synapse%s):\n"
		"    %5zd hosted,\n"
		"    %5zd standalone\n"
		"    %5zd discrete dt-bound\n"
		"  %5zd Listening unit%s\n"
		"  %5zd Spikelogging neuron%s\n"
		"  %5zd Tape reader%s\n"
		"  %5zd Unit%s with scheduled parameter changes\n"
		"  %5zd Unit%s with parameter ranges\n"
		"  %5zd Spontaneously firing neuron%s\n"
		"  %5zd Multiplexing synapse%s\n"
		" %6zd vars on integration vector\n\n",
		name.c_str(), (_status & CN_MDL_DISKLESS) ? " (diskless)" : "",
		units(), pl_ending(units()),
		total_neuron_cnt(), pl_ending(total_neuron_cnt()),
		total_synapse_cnt(), pl_ending(total_synapse_cnt()),
		hosted_unit_cnt(),
		standalone_unit_cnt(),
		ddtbound_unit_cnt(),
		listening_unit_cnt(), pl_ending(listening_unit_cnt()),
		spikelogging_neuron_cnt(), pl_ending(spikelogging_neuron_cnt()),
		reader_unit_cnt(), pl_ending(reader_unit_cnt()),
		scheduler_unit_cnt(), pl_ending(scheduler_unit_cnt()),
		param_range_unit_cnt(), pl_ending(param_range_unit_cnt()),
		conscious_neuron_cnt(), pl_ending(conscious_neuron_cnt()),
		mx_syn_list.size(), pl_ending(mx_syn_list.size()),
		_var_cnt-1);
}


// EOF
