// ;-*-C++-*- *  Time-stamp: "2010-01-21 21:32:11 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *
 * License: GPL
 *
 * Initial version: 2008-09-02
 *
 * CNModel runner (main, cmdline parser)
 */


#include <sys/stat.h>
#include <cfloat>
#include "runner.h"

#include "config.h"

CModel *Model;



SCNRunnerOptions Options;

#define CNRUN_DEFAULT_PARAMFNAME "./Default.cnp"
#define CNRUN_DEFAULT_NMLFNAME	 "./Default.nml"

vector<const char*> nml_fnames;
vector<string> ParameterFiles;

vector<string> Listeners;
vector<string> CmdLineUnitSpecs;
vector<string> Spikeloggers;
vector<string> ToRemove;
vector<CModel::SDecimationGroup> ToDecimate;




#define CNRUN_CLPARSE_HELP_REQUEST	-1
#define CNRUN_CLPARSE_ERROR		-2
#define CNRUN_DUMP_PARAMS		-3


static int parse_options( int argc, char **argv);
static void usage( const char *argv0);


#define CNRUN_EARGS		-1
#define CNRUN_ESETUP		-2
#define CNRUN_ETRIALFAIL	-3



int
main( int argc, char *argv[])
{
	if ( argc == 1 ) {
		cout << "No operational options on command line; use -h for help\n";
		return 0;
	}

	int retval = 0;

	switch ( parse_options( argc, argv) ) {
	case CNRUN_CLPARSE_ERROR:
		cerr << "Problem parsing command line or sanitising values; try -h for help\n";
		return CNRUN_EARGS;
	case CNRUN_CLPARSE_HELP_REQUEST:
		usage( argv[0]);
		return 0;
	}

	if ( Options.verbosely > 0 )
		cout << "CNRun v. " << VERSION << " compiled " << __TIME__ << " " << __DATE__ << endl;

      // purely informational, requires no model
	if ( Options.list_units ) {
		cnmodel_dump_available_units();
		return 0;
	}

      // check for some required options
	if ( !Options.dump_params ) {
		if ( Options.running_time == 0 ) {
			cerr << "Running time (-t) not specified\n";
			return CNRUN_EARGS;
		}
		if ( Options.dt_max < DBL_MIN ) {
			cerr << "dt_max is too small (<" << DBL_MIN << ")\n";
			return CNRUN_EARGS;
		}
		if ( Options.dt_min < DBL_MIN || Options.dt_min > Options.dt_max ) {
			cerr << "Supplied dt_min either too small (<" << DBL_MIN << ") or greater than dt_max\n";
			return CNRUN_EARGS;
		}
	}

	if ( nml_fnames.size() == 0 ) {
		cout << "(Using default model topology file: " CNRUN_DEFAULT_NMLFNAME ")\n";
		nml_fnames.push_back( CNRUN_DEFAULT_NMLFNAME);
	}

      // cd as requested
	char *pwd = NULL;
	if ( Options.working_dir ) {
		pwd = getcwd( NULL, 0);
		if ( chdir( Options.working_dir) ) {
			cerr << "Failed to cd to \"" << Options.working_dir << "\"\n";
			return CNRUN_EARGS;
		}
	}


      // create and set up the model
	__cn_verbosely = Options.verbosely;
	__cn_default_unit_precision = Options.precision;
	if ( !(Model = new CModel( "(unnamed)",
				   new CIntegrateRK65( Options.dt_min,
						       Options.dt_max,
						       Options.dt_max_fac),
				   0 |
				   (Options.log_dt ? CN_MDL_LOGDT : 0) |
				   (Options.log_spikers ? CN_MDL_LOGSPIKERS : 0) |
				   (Options.sort_units ? CN_MDL_SORTUNITS : 0) |
				   (Options.log_spikers_use_serial_id ? CN_MDL_LOGUSINGID : 0) |
				   (Options.display_progress_percent ? CN_MDL_DISPLAY_PROGRESS_PERCENT : 0) |
				   (Options.display_progress_time ? CN_MDL_DISPLAY_PROGRESS_TIME : 0))) ) {
		cerr << "Failed to create a model\n";
		return CNRUN_ESETUP;
	}
	Model->verbosely = Options.verbosely;
	Model->discrete_dt( Options.discrete_dt);
	Model->logging_dt = Options.logging_dt;
	Model->spike_threshold = Options.spike_threshold /*,  Model->spike_lapse = Options.spike_lapse */;
#ifdef USE_GSL_RNG
	if ( Options.verbosely > 3 ) {
		printf( "generator type: %s\n"
			"         seed = %lu\n"
			"  first value = %lu\n", gsl_rng_name (Model->_rng), gsl_rng_default_seed, gsl_rng_get (Model->_rng));
	}
#endif

      // load nml, remember param fnames
	{
		vector<string> MoreParamFiles;  // in addition to what user supplied with -P

		struct stat s;
		if ( stat( CNRUN_DEFAULT_PARAMFNAME, &s) == 0 )
			MoreParamFiles.push_back( CNRUN_DEFAULT_PARAMFNAME);

		for ( vector<const char*>::iterator F = nml_fnames.begin(); F != nml_fnames.end(); F++ ) {
			if ( Model->import_NetworkML( *F, true) < 0 ) {
				fprintf( stderr, "Failed to import topology from `%s'\n", *F);
				retval = CNRUN_ESETUP;
				goto out;
			}

			string try_param_fname( *F);
			if ( try_param_fname.size() > 4 &&
			     strcmp( &try_param_fname[try_param_fname.size()-4], ".nml") == 0 )
				try_param_fname.resize( try_param_fname.size()-4);
			try_param_fname += ".cnp";

			if ( stat( try_param_fname.data(), &s) == 0 )
				MoreParamFiles.push_back( try_param_fname);
		}

		for ( vector<string>::reverse_iterator P = MoreParamFiles.rbegin(); P != MoreParamFiles.rend(); P++ )
			ParameterFiles.insert( ParameterFiles.begin(), *P);
	}


	if ( Options.dump_params ) {
		dump_model_units();
		goto out;
	}

      // goes after the model creation since params are checked against a live model
	if ( do_unit_specs() ) {
		cerr << "Failed parsing unit specs\n";
		retval = CNRUN_ESETUP;
		goto out;
	}


      // ensure tuned values appear in dump_model_units instead of stock ones
	if ( assign_unit_static_params() ) {
		cerr << "Failed to set up unit parameters\n";
		retval = CNRUN_ESETUP;
		goto out;
	}

      // loggers are remembered by pointers, so tag them now (before
      // reducing the unit headcount in cull_deaf_absynapses
	if ( Model->process_listener_tags( Listeners, CN_ULISTENING_DISK
					   | (Options.listen_1varonly ? CN_ULISTENING_1VARONLY : 0)
					   | (Options.listen_deferwrite ? CN_ULISTENING_DEFERWRITE : 0)) ||
	     Model->process_spikelogger_tags( Spikeloggers, Options.sxf_sample, Options.sdf_sigma) ||
	     Model->process_putout_tags( ToRemove) ||
	     Model->process_decimate_tags( ToDecimate) ) {
		retval = CNRUN_ESETUP;
		goto out;
	}
	Model->cull_deaf_synapses();


	if ( Options.verbosely > 1 )
		Model->dump_metrics();

	if ( !Options.allow_silent_run &&
	     Model->listening_unit_cnt() + Model->spikelogging_neuron_cnt() == 0 ) {
		cout << "Not starting simulation because no output will be generated from this run.\n"
			"Use -y if this is intentional, or specify some spikeloggers (-o) or listeners (-l)\n";
		retval = 0;
		goto out;
	}


	if ( Options.verbosely > 2 ) {
		if ( Model->status() & CN_MDL_HAS_DDTB_UNITS )
			printf( "Discrete dt: %g msec\n", Model->discrete_dt());
		printf( "Integrator dt_min: %g msec, dt_max: %g msec\n"
			"Logging at: %g msec\n\n",
			Model->dt_min(), Model->dt_max(),
			Model->logging_dt);
	}


      // run!
	if ( Batches.size() > 0 ) {
		if ( Options.verbosely )
			cout << "\nIterating over " << Batches.size() << " ranges\n";
		do_param_batch( 0);
	} else {
		Model->advance( Options.running_time);
		for_model_spikelogging_neurons (Model,N)
			(*N)->sync_spikelogging_history();
	}

out:
	delete Model;

	if ( pwd )
		if ( chdir( pwd) )
			;

	return retval;
}








static int
parse_options( int argc, char **argv)
{
	int	c;

	while ( (c = getopt( argc, argv, "M:t:L:q:E:e:K:k:o:S:d:l:yp:P:i:DUTsC:v:h")) != -1 )
		switch ( c ) {
		case 'M':
			nml_fnames.push_back( optarg);
		    break;

		case 't':	if ( sscanf( optarg, "%lu", &Options.running_time) != 1 ) {
					cerr << "-t takes an unsigned long\n";
					return CNRUN_CLPARSE_ERROR;
				}						break;
//		case '1':	Options.single_neuron = optarg;			break;

		case 'i':	switch ( optarg[0] ) {
				case 'T':	if ( sscanf( optarg+1, "%lg", &Options.dt_max) != 1 ) {
							cerr << "-i takes a double\n";
							return CNRUN_CLPARSE_ERROR;
						}						break;
				case 't':	if ( sscanf( optarg+1, "%lg", &Options.dt_min) != 1 ) {
							cerr << "-j takes a double\n";
							return CNRUN_CLPARSE_ERROR;
						}						break;
				case 'x':	if ( sscanf( optarg+1, "%lg", &Options.dt_max_fac) != 1 ) {
							cerr << "-J takes a double\n";
							return CNRUN_CLPARSE_ERROR;
						}						break;
				default:	cerr << "Unrecognised option modifier for -i\n";
						return CNRUN_CLPARSE_ERROR;
				}						break;

		case 'L':	if ( strchr( optarg, 'd') != NULL )
					Options.listen_deferwrite = true;
				if ( strchr( optarg, '1') != NULL )
					Options.listen_1varonly = true;		break;

		case 'l':	Listeners.push_back( "");
				Listeners.back() = optarg;			break;
		case 'q':	if ( sscanf( optarg, "%lg", &Options.discrete_dt) != 1 ) {
					cerr << "-q takes a double\n";
					return CNRUN_CLPARSE_ERROR;
				}						break;
		case 'E':	if ( sscanf( optarg, "%lg", &Options.logging_dt) != 1 ) {
					cerr << "-E takes a double\n";
					return CNRUN_CLPARSE_ERROR;
				}						break;
		case 'y':	Options.allow_silent_run = true;		break;

		case 'e':	if ( sscanf( optarg, "%u", &Options.precision) != 1 ) {
					cerr << "-e takes a short unsigned int\n";
					return CNRUN_CLPARSE_ERROR;
				}						break;

		case 'S':	switch ( optarg[0] ) {
				case 't':	if ( sscanf( optarg+1, "%lg", &Options.spike_threshold) != 1 ) {
							cerr << "-St takes a double\n";
							return CNRUN_CLPARSE_ERROR;
						}						break;
				// case 'l':	if ( sscanf( optarg+1, "%lg", &Options.spike_lapse) != 1 ) {
				// 			cerr << "-Sl takes a double\n";
				// 			return CNRUN_CLPARSE_ERROR;
				// 		}						break;
				default:	cerr << "Unrecognised option modifier for -S\n";
						return CNRUN_CLPARSE_ERROR;
				}						break;

		case 'o':	Spikeloggers.push_back( "");
				Spikeloggers.back() = optarg;			break;
//		case 'K':	Options.spike_cnt_stats_file = optarg;		break;
		case 'k':	Options.log_spikers = true;
				switch ( *optarg ) {
				case '0':	Options.log_spikers_use_serial_id = true;	break;
				case 'l':	Options.log_spikers_use_serial_id = false;	break;
				default:	cerr << "Expecting either -k0 or -kl\n";
						return CNRUN_CLPARSE_ERROR;
				}						break;
		case 'd':	Options.log_sdf = true;
				if ( sscanf( optarg, "%d:%d",
					     &Options.sxf_sample, &Options.sdf_sigma) < 2 ) {
					cerr << "Expecting two parameters with -d (sxf_period:sdf_sigma)\n";
					return CNRUN_CLPARSE_ERROR;
				}						break;

		case 'p':	CmdLineUnitSpecs.push_back( "");
				CmdLineUnitSpecs.back() = optarg;		break;

		case 'P':	ParameterFiles.push_back( optarg);		break;

		case 'D':	Options.dump_params = true;			break;
		case 'U':	Options.list_units = true;			break;

		case 'T':	Options.log_dt = true;				break;

		case 's':	Options.sort_units = true;			break;

		case 'C':	Options.working_dir = optarg;			break;

		case 'v':	Options.verbosely = strtol( optarg, NULL, 10);
				{	char *p;
					if ( (p = strchr( optarg, '%')) )
						Options.display_progress_percent = (*(p+1) != '-');
					if ( (p = strchr( optarg, 't')) )
						Options.display_progress_time    = (*(p+1) != '-');
				}
										break;
		case 'h':
			return CNRUN_CLPARSE_HELP_REQUEST;
		case '?':
		default:
			return CNRUN_CLPARSE_ERROR;
		}

	return 0;
}




static void
usage( const char *argv0)
{
	cout << "Usage: " << argv0 << "\n" <<
		" -M <filename>\t\tModel topology (can be repeated; default " CNRUN_DEFAULT_NMLFNAME ")\n"
		" -D \t\t\tDump all unit types in the model and exit\n"
		" -C <dir>\t\tWork in dir\n"
		" -t <ulong>\t\tRunning time in msec\n"
		" -q <double>\t\tTime step for discrete units\n"
		" -y \t\t\tDon't refuse to run if no output requested\n"
		"\t\t\t  (i.e., no spikeloggers nor listeners)\n"
		" -s \t\t\tSort units\n"
		"\n"
		" -l <regex>\t\tMake so labelled unit(s) listen\n"
		" -L[d1] \t\tFor all listeners:\n"
		"\t\t\t    d, defer writing to disk until done rather than write continuously\n"
		"\t\t\t    1, only log the first variable\n"
		" -E <double>\t\tListen at this interval (default " << Options.logging_dt << ";\n"
		"\t\t\t   set to 0 to listen every cycle)\n"
		"\n"
		" -o <regex>\t\tLog spiking times for these neurons\n"
		" -kl \t\t\tWrite a model-wide log of spiking neurons, using labels\n"
		" -k0 \t\t\t... use unit id instead\n"
		" -K <filename>\t\tReport spike count stats in filename (\"-\" = stdout)\n"
		" -d <double>:<double>\tEstimate spike density every specified period:sigma\n"
		" -St <double>\t\tSpike detection threshold (default " << Options.spike_threshold << ")\n"
//		" -Sl <double>\t\tSpike lapse (default " << Options.spike_lapse << ")\n"
		"\n"
		" -p <regex>{<parm>}<expr>\n"
		"\t\t\tSet matching unit(s)' parameter\n"
		"\t\t\t    expr can include +, -, *, /, and ()\n"
		" -P <filename>\t\tRead these from filename, default " CNRUN_DEFAULT_PARAMFNAME "\n"
		"\t\t\t    and files matching model_name s/.nml$/.cnp/\n"
		"\n"
		" -e <uint>\t\tSet precision for all output (default " << Options.precision << ")\n"
		"\n"
		" -iT <double>\t\tdt_max (default " << Options.dt_max << ")\n"
		" -it <double>\t\tdt_min (default " << Options.dt_min << ")\n"
		" -ix <double>\t\tCap dt by current dt value x this (default " << Options.dt_max_fac << ")\n"
		" -T \t\t\tLog dt\n"
		"\n"
		" -v <int>\t\tSet verbosity level (default " << Options.verbosely << ")\n"
		"\t\t\t Use a negative value to show the progress percentage only,\n"
		" -v[%[-]t[-]]\t\tDisplay (with -, suppress) progress percent and/or time\n"
		"\t\t\t indented on the line at 8 x (minus) this value.\n"
		"\n"
		" -U \t\t\tList available unit types with parameter names\n"
		"\t\t\t  and values, and exit\n"
		" -h \t\t\tDisplay this help\n"
		"\n";
}



// EOF
