// ;-*-C++-*- *  Time-stamp: "2009-12-20 17:38:18 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *
 * License: GPL
 *
 * Initial version: 2008-09-02
 *
 * CNModel runner (triplet and parameter parser)
 */

#include <cfloat>
#include <boost/regex.hpp>
#include "exprparser.h"

#include "config.h"

#include "runner.h"


vector<SParmValuePair> Singles;
vector<SParmValueRange> Ranges;
vector<SParmReaderOrScheduler> Readers;
vector<SParmReaderOrScheduler> Schedulers;




#define CNRUN_PARSE_HELP_REQUEST	-1
#define CNRUN_PARSE_ERROR		-2
#define CNRUN_DUMP_PARAMS		-3

#define CNRUN_SPEC_ILLFORMED -1
#define CNRUN_SPEC_BADVALUE  -2
#define CNRUN_SPEC_BADPARAM  -3
#define CNRUN_SPEC_BADUNIT   -4
#define CNRUN_SPEC_BADCMD    -5


static int do_single_unit_spec( const char*, unsigned lineno = 0);
static int do_triplet( const char*, unsigned lineno = 0);




int
do_unit_specs()
{
	for ( vector<string>::iterator PF = ParameterFiles.begin(); PF != ParameterFiles.end(); PF++ ) {
		ifstream parm_stream( PF->c_str());
		if ( !parm_stream.good() ) {
			fprintf( stderr, "Failed to open parameter file \"%s\"\n", PF->c_str());
			return -5;
		}

		if ( Options.verbosely > 3 )
			printf( "\n--------- Parameters assigned from %s:\n", PF->c_str());
		unsigned lineno = 0;
		string buf;
		while ( getline( parm_stream, buf, '\n') || parm_stream.gcount() ) {
			lineno++;
			if ( buf.size() && buf[0] != '#' ) {
				int retval = do_single_unit_spec( buf.c_str(), lineno);
				if ( retval )
					return retval;
			}
		}
	}

	if ( CmdLineUnitSpecs.size() )
		if ( Options.verbosely > 3 )
			cout << "------------ Parameters assigned from cmdline:\n";
	vector<string>::iterator P;
	for ( P = CmdLineUnitSpecs.begin(); P != CmdLineUnitSpecs.end(); P++ ) {
		int retval = do_single_unit_spec( P->c_str());
		if ( retval )
			return retval;
	}

	if ( Options.verbosely > 2 )
		cout << endl;

	return 0;
}



static const char* whitespace = " \t";


static int
do_single_unit_spec( const char *spec, unsigned lineno)
{
	char lineno_s[20] = { '\0' };
	if ( lineno )
		snprintf( lineno_s, 19, " at line %u", lineno);

	if ( strchr( spec, '{') )
		return do_triplet( spec);

	string	spec_copy( spec);
	char	*cmd = strtok( &spec_copy[0], whitespace),
		*label = strtok( NULL, whitespace);
	if ( !label ) {
		if ( lineno )
			fprintf( stderr, "No unit label given in \"%s\" at line %u\n", spec, lineno);
		else
			fprintf( stderr, "No unit label given in \"%s\"\n", spec);
		return CNRUN_SPEC_ILLFORMED;
	}

	if ( strcmp( cmd, "listen") == 0 )
		Listeners.push_back( string(label));

	else if ( strcmp( cmd, "log_spikes") == 0 )
		Spikeloggers.push_back( string(label));

	else if ( strcmp( cmd, "delete") == 0 )
		ToRemove.push_back( string(label));

	else if ( strcmp( cmd, "decimate") == 0 ) {
		CModel::SDecimationGroup dcmspec;
		char *fraction_s = strtok( NULL, whitespace);
		if ( !fraction_s || sscanf( fraction_s, "%g", &dcmspec.fraction) < 1 ) {
			fprintf( stderr, "Missing decimation fraction in \"%s\"%s\n", spec, lineno_s);
			return CNRUN_SPEC_ILLFORMED;
		}
		if ( dcmspec.fraction < 0. || dcmspec.fraction > 1. ) {
			fprintf( stderr, "Decimation fraction in \"%s\" outside [0..1]%s\n", spec, lineno_s);
			return CNRUN_SPEC_ILLFORMED;
		}

		dcmspec.name.assign( label);
		ToDecimate.push_back( dcmspec);

	} else {
		fprintf( stderr, "Unrecognised command: \"%s\"%s\n", cmd, lineno_s);
		return CNRUN_SPEC_BADCMD;
	}

	return 0;
}




enum { THING_VALUE, THING_RANGE, THING_READER, THING_SCHEDULER, THING_BATCH };

static int
classify_and_parse_argument( char* what, char*& e1, char*& e2, char*& e3,
			     bool &reader_is_looping)
{
	int have_thing = -1;

	e1 = what;

	while ( isspace( *e1) )
		e1++;

	if ( *e1 == ':' ) {
		e1++;
		if ( *e1 == '<' ) {
			reader_is_looping = true;
			e1++;
		}
		e2 = e3 = NULL;
		return have_thing = THING_READER;
	}

	if ( *e1 == '<' ) {
		e1++;
		if ( *e1 == '<' ) {
			reader_is_looping = true;
			e1++;
		}
		e2 = e3 = NULL;
		return have_thing = THING_SCHEDULER;
	}

      // is it a BATCH?
	if ( (e2 = strstr( e1, "::")) ) {
		*e2 = '\0';
		e2 += 2;
		if ( (e3 = strstr( e2, "::")) ) {
			*e3 = '\0';
			e3 += 2;
			return have_thing = THING_BATCH;
		}
	}

      // or a RANGE?
	if ( have_thing == -1 )  // no "::" found to select a BATCH: try ':' for a RANGE
		if ( (e2 = strchr( e1, ':')) ) {
			*e2 = '\0';
			e2++;
			if ( (e3 = strchr( e2, ':')) ) {
				*e3 = '\0';
				e3++;
				return have_thing = THING_RANGE;
			}
		}

	return THING_VALUE;
}



static int
do_triplet( const char* triplet, unsigned lineno)
{
	char lineno_s[20] = { '\0' };
	if ( lineno )
		snprintf( lineno_s, 19, " at line %u", lineno);

	string triplet_copy( triplet);
	char	*l_s = strtok( &triplet_copy[0], "{"),
		*p_s = strtok( NULL, "}"),
		*v_s = strtok( NULL, "\n");
	l_s = strtok( l_s, whitespace);
	p_s = strtok( p_s, whitespace);
	if ( !l_s || !p_s || !v_s ) {
		fprintf( stderr, "Ill-formed label{param}value triplet: \"%s\"%s\n",
			 triplet, lineno_s);
		return CNRUN_SPEC_ILLFORMED;
	}

	bool	assigned = false;

	int p_d;

      // now deal with the argument
	char	*v1s = v_s, *v2s, *v3s;
	double	v1,  // serves as a value for THING_VALUE, as v_min for THING_{RANGE,BATCH}
		v2 = 0., v3;  // used for both BATCH and RANGE

	bool	reader_is_looping = false;
	CExpression expr1, expr2, expr3, expr4;
	int	have_thing = classify_and_parse_argument( v_s, v1s, v2s, v3s, reader_is_looping);

	switch ( have_thing ) {

	case THING_BATCH:
		if ( expr1( v1s, v1) || expr2( v2s, v2) || expr3( v3s, v3) ) {
			fprintf( stderr, "Unparsable expression in batch specification in triplet \"%s\"%s\n", triplet, lineno_s);
			return CNRUN_SPEC_BADVALUE;
		}
		if ( v1 + v2 >= v3 ) {
			fprintf( stderr, "Step value (%g) not within [%g:%g] in triplet \"%s\"%s\n",
				 v2, v1, v3, triplet, lineno_s);
			return CNRUN_SPEC_BADVALUE;
		}
	      break;

	case THING_RANGE:
		if ( expr1( v1s, v1) || expr2( v2s, v2) || expr3( v3s, v3) ) {
			fprintf( stderr, "Unparsable expression in range specification in triplet \"%s\"%s\n", triplet, lineno_s);
			return CNRUN_SPEC_BADVALUE;
		}
	    break;

	case THING_VALUE:
		if ( expr1( v1s, v1) ) {
			fprintf( stderr, "Unparsable expression in triplet \"%s\"%s\n", triplet, lineno_s);
			return CNRUN_SPEC_BADVALUE;
		}
	    break;
	}


	boost::regex pattern( l_s);
	boost::cmatch found;

	vector<string> assigned_labels;

	for_model_units (Model,U) {
		if ( !boost::regex_match( (*U)->label, found, pattern) )
			continue;
	      // because a named parameter can map to a different param_id in different units, rather
	      // do lookup every time
		C_BaseUnit::TDeviceAssignee kind = C_BaseUnit::RDA_VOID;
		if ( (p_d = (*U)->param_idx_by_sym( p_s)) > -1 )
			kind = C_BaseUnit::RDA_PARAM;
		if ( kind == C_BaseUnit::RDA_VOID && (p_d = (*U)->var_idx_by_sym( p_s)) > -1 )
			kind = C_BaseUnit::RDA_VAR;
		if ( kind == C_BaseUnit::RDA_VOID ) {
			fprintf( stderr,
				 "%s \"%s\" (type \"%s\") has no parameter or variable named \"%s\"%s\n",
				 (*U)->class_name(), (*U)->label, (*U)->type_name(), p_s, lineno_s);
			return CNRUN_SPEC_BADPARAM;
		}

		if ( kind == C_BaseUnit::RDA_PARAM && (p_d < 0 || p_d >= (int)(*U)->p_no()) ) {
			fprintf( stderr, "%s \"%s\" (type \"%s\") has only %d parameter(s)%s\n",
				 (*U)->class_name(), (*U)->label, (*U)->type_name(), (*U)->p_no(), lineno_s);
			return CNRUN_SPEC_BADPARAM;
		}

		switch ( have_thing ) {
		case THING_VALUE:
		{
			SParmValuePair pvp; // = { *N, (unsigned short)p_d, v_d };
			pvp.unit = *U,  pvp.kind = kind,  pvp.parm = (unsigned short)p_d,  pvp.value = v1;
			Singles.push_back( pvp);

			assigned  = true;
			assigned_labels.push_back( (*U)->label);
		} break;

		case THING_BATCH:
		{
			SParmValueBatch pvr; // = { *N, (unsigned short)p_d, v_d, v_inc, v_max };
			pvr.unit = *U,  pvr.kind = kind,  pvr.parm = (unsigned short)p_d,
				pvr.min = v1,  pvr.step = v2,  pvr.max = v3;
			Batches.push_back( pvr);

			assigned  = true;
			assigned_labels.push_back( (*U)->label);
		} break;

		case THING_RANGE:
		{
			SParmValueRange pvr; // = { *N, (unsigned short)p_d, v_d, v_inc, v_max };
			pvr.unit = *U,  pvr.kind = kind,  pvr.parm = (unsigned short)p_d,
				pvr.min = v1,  pvr.span = v2,  pvr.max = v3;
			Ranges.push_back( pvr);

			assigned  = true;
			assigned_labels.push_back( (*U)->label);
		} break;

		case THING_READER:
		{
			SParmReaderOrScheduler pr;
			pr.unit = *U,  pr.kind = kind,  pr.parm = (unsigned short)p_d,
				pr.fname = v1s, pr.is_looping = reader_is_looping;
			Readers.push_back( pr);

			assigned = true;
			assigned_labels.push_back( (*U)->label);
		} break;
		case THING_SCHEDULER:
		{
			SParmReaderOrScheduler pr;
			pr.unit = *U,  pr.kind = kind,  pr.parm = (unsigned short)p_d,
				pr.fname = v1s, pr.is_looping = reader_is_looping;
			Schedulers.push_back( pr);

			assigned = true;
			assigned_labels.push_back( (*U)->label);
		} break;

	      // cause p_d to be looked up again from p_s
		p_d = -1;
		}
	}

	if ( !assigned ) {
		fprintf( stderr, "No unit labelled matching \"%s\"%s\n", l_s, lineno_s);
		return CNRUN_SPEC_BADUNIT;
	}

	if ( Options.verbosely > 3 ) {
		printf( " set ");
		for ( vector<string>::iterator S = assigned_labels.begin(); S != assigned_labels.end(); S++ )
			printf( "%s%s",
				(S == assigned_labels.begin()) ? "" : ", ",
				/* Model->longest_label(), */ S->c_str());
		switch ( have_thing ) {
		case THING_VALUE:
			printf( " {%s} = %g\n", p_s, v1);
		    break;
		case THING_SCHEDULER:
			printf( " {%s}<%s\"%s\"\n", p_s, reader_is_looping ? "<" : "", v1s);
		    break;
		case THING_READER:
			printf( " {%s}:%s\"%s\"\n", p_s, reader_is_looping ? "<" : "", v1s);
		    break;
		case THING_RANGE:
			printf( " {%s} = [%g..%g in %g msec]\n", p_s, v1, v3, v2);
		    break;
		}
	}
	return 0;
}







int
assign_unit_static_params()
{
	for ( vector<SParmValuePair>::iterator S = Singles.begin(); S != Singles.end(); S++ ) {
		switch ( S->kind ) {
		case C_BaseUnit::RDA_PARAM:
			S->unit->param_value(S->parm) = S->value;
		    break;
		case C_BaseUnit::RDA_VAR:
			S->unit->  var_value(S->parm) = S->value;
		    break;
		default:
//			fprintf( stderr, "What kind of a unit is that?\n");
			;
		}
		S->unit->param_changed_hook();
	}

	for ( vector<SParmReaderOrScheduler>::iterator R = Readers.begin(); R != Readers.end(); R++ )
		if ( R->unit -> attach_reader_device( R->fname.c_str(), R->is_looping,
						      R->kind, R->parm) ) {
			fprintf( stderr, "Failed to set up a reader for %s from \"%s\"\n",
				 R->unit->label, R->fname.c_str());
			return -1;
		}

	for ( vector<SParmReaderOrScheduler>::iterator R = Schedulers.begin(); R != Schedulers.end(); R++ )
		if ( R->unit -> attach_scheduler_device( R->fname.c_str(), R->is_looping,
							 R->kind, R->parm) ) {
			fprintf( stderr, "Failed to set up a scheduler for %s from \"%s\"\n",
				 R->unit->label, R->fname.c_str());
			return -1;
		}

	for ( vector<SParmValueRange>::iterator R = Ranges.begin(); R != Ranges.end(); R++ )
		R->unit -> attach_param_range( R->min, R->span, R->max,
					       R->kind, R->parm);
	return 0;
}

// EOF
