// ;-*-C++-*- *  Time-stamp: "2008-10-16 03:40:57 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-09-02
 *
 * Integrating method for CModel (TN's original description below)
 */

/*--------------------------------------------------------------------------
  Implementation of a 6-5 Runge-Kutta method with adaptive time step
  mostly taken from the book "The numerical analysis of ordinary differential
  equations - Runge-Kutta and general linear methods" by J.C. Butcher, Wiley,
  Chichester, 1987 and a free adaptation to a 6 order Runge Kutta method
  of an ODE system with additive white noise
--------------------------------------------------------------------------*/


#include "cnlib-integrate-rk65.h"
#include "cnlib-model.h"
#include <cfloat>



static double __Butchers_a[9][8] = {
	{ },
	{ 1./9 },
	{ .5/9,	.5/9 },
	{ 0.416666666666667e-1,	0., 0.125 },
	{ 1./6, 0., -0.5, 2./3},
	{ 0.1875e+1, 0., -0.7875e+1, 0.7e+1, -0.5},
	{ -0.4227272727272727e+1, 0., 0.176995738636364e+2, -0.142883522727273e+2, 0.522017045454545, 0.104403409090909e+1},
	{ 0.840622673179752e+1, 0., -0.337303717185049e+2, 0.271460231129622e+2, 0.342046929709216, -0.184653767923258e+1, 0.577349465373733},
	{ 0.128104575163399, 0., 0., -0.108433734939759, 0.669375, -0.146666666666667, 0.284444444444444, 0.173176381998583},
};


static double __Butchers_b[9] = {
	0.567119155354449e-1,
	0.,
	0.,
	0.210909572355356,
	0.141490384615385,
	0.202051282051282,
	0.253186813186813,
	0.843679809736684e-1,
	0.512820512820513e-1
};




void
CIntegrateRK65::prepare()
{
	for ( int i = 0; i < 9; i++ )
		Y[i].resize( model->_var_cnt), F[i].resize( model->_var_cnt);
	y5.resize( model->_var_cnt);
}


void
CIntegrateRK65::cycle()
{
	unsigned int i, j, k;

	double	aF;

      // calculate iterative terms rk65_Y[__i] and rk65_F[__i] (to sixth order)
	for ( i = 0; i < 9; i++ ) {
		for ( k = 0; k < model->_var_cnt; k++ ) {
			aF = 0.0;
			for ( j = 0; j < i; j++ )
				aF += __Butchers_a[i][j] * F[j][k];
			Y[i][k] = (*model->V)[k] + dt * aF;
		}
		model->derivative( Y[i], F[i]);
	}

      // sum up Y[i] and F[i] to build 5th order scheme -> y5
	for ( k = 0; k < model->_var_cnt; k++ ) {
		aF = 0.0;
		for ( j = 0; j < 8; j++ )
			aF += __Butchers_a[8][j] * F[j][k];
		y5[k] = (*model->V)[k] + dt * aF;
	}

      // sum up Y[i] and F[i] to build 6th order scheme -> W
	for ( k = 0; k < model->_var_cnt; k++ ) {
		aF = 0.0;
		for ( j = 0; j < 9; j++ )
			aF += __Butchers_b[j] * F[j][k];
		(*model->W)[k] = (*model->V)[k] + dt * aF;
	}

      // kinkiness in synapses causes dt to rocket
	double	dtx = min( dt_max, dt * dt_max_safeguard_factor);

      // determine minimal necessary new dt to get error < eps based on the
      // difference between results in y5 and W
	double try_eps, delta, try_dt;
      // exclude time (at index 0)
	for ( k = 1; k < model->_var_cnt; k++ ) {
		try_eps = max (eps_abs, min (eps, abs(eps_rel * (*model->W)[k])));
		delta = abs ((*model->W)[k] - y5[k]);
		if ( delta > DBL_EPSILON * y5[k] ) {
			try_dt = exp( (log(try_eps) - log(delta)) / 6) * dt;
			if ( try_dt < dtx )
				dtx = try_dt;
		}
	}

      // make sure we don't grind to a halt
	if ( dtx < dt_min )
		dtx = dt_min;

      // set the new step
	dt = dtx;

      // `fixate'
	vector<double> *_ = model->V;
	model->V = model->W, model->W = _;
}


// EOF
