// ;-*-C++-*- *  Time-stamp: "2008-10-16 03:40:57 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-09-02
 *
 * CModel household members
 */

#include "cnlib-model.h"

#include <iostream>
#include <algorithm>
#include <ctime>

using namespace std;


CModel::CModel( const char *inname, CIntegrate_base *inintegrator, int instatus)
{
	name = inname;
	_cycle = 0, _unit_cnt = 0;
	_var_cnt = 1; // reserve [0] for model_time

	V = new vector<double>, W = new vector<double>;
	V->resize( _var_cnt), W->resize( _var_cnt);
	(*V)[0] = 0.;

	(_integrator = inintegrator) -> model = this;

	_status = instatus;
	_dt_logger = NULL;

	spike_thresh = CNM_SPIKETHRESH, spike_lapse = CNM_SPIKELAPSE;

	verbosely = 1;
}


CModel::~CModel()
{
	list<CNeuron*>::iterator N;
	for_all_neurons (N)
		delete *N;

	list<CSynapse*>::iterator Y;
	for_all_synapses (Y)
		delete *Y;

	delete V;
	delete W;

	if ( _integrator->is_owned )
		delete _integrator;

	if ( _dt_logger )
		delete _dt_logger;
}


void
CModel::reset()
{
	list<CNeuron*>::iterator N;
	for ( N = neu_list.begin(); N != neu_list.end(); N++ )
		delete *N;

	list<CSynapse*>::iterator Y;
	for ( Y = syn_list.begin(); Y != syn_list.end(); Y++ )
		delete *Y;

	V->resize( 1);
	W->resize( 1);

	_var_cnt = 1, _unit_cnt = 0;
	_cycle = 0, (*V)[0] = 0.;
}









CNeuron*
CModel::neuron_by_label( const char *inlabel)
{
	list<CNeuron*>::iterator N;
	for_all_neurons (N)
		if ( strcmp( (*N)->label, inlabel) == 0 )
			return *N;
	return NULL;
}

CSynapse*
CModel::synapse_by_label( const char *inlabel)
{
	list<CSynapse*>::iterator Y;
	for_all_synapses (Y)
		if ( strcmp( (*Y)->label, inlabel) == 0 )
			return *Y;
	return NULL;
}







int
CModel::add_neuron( CNeuron *n, bool is_last)
{
      // prevent multiple neurons at same coordinates
	list<CNeuron*>::iterator N;
	for_all_neurons (N)
		if ( (*N)->type() != NT_READER && n->type() != NT_READER &&
		     (*N) -> pos.too_close( n->pos) &&
		     verbosely > 1 )
			cerr << "Warning: neuron " << n->label << " at (" << n->pos.x << "," << n->pos.y << "," << n->pos.z
			     << ") too close to " << (*N)->label << " at (" << (*N)->pos.x << "," << (*N)->pos.y << "," << (*N)->pos.z << ")\n";

	neu_list.push_back( n);

      // assign idx
	n->idx = _var_cnt;
	n->_parent_model = this;

	_var_cnt += n->v_no();
	_unit_cnt++;

	if ( is_last ) {
		finalize_additions();
		n->reset_state();
	}

	return 0;
}


int
CModel::add_synapse( CSynapse* y, bool is_last)
{
      // make sure synapse's source and target exist in the model
	bool	src_found = false,
		tgt_found = false;
	list<CNeuron*>::iterator N;
	for_all_neurons (N) {
		if      ( *N == y->_source )
			src_found = true;
		else if ( *N == y->_target )
			tgt_found = true;

		if ( y->_type == YT_OSCILTUNER &&
		     (y->_target->_type != NT_POISSON &&
		      y->_target->_type != NT_LV &&
		      y->_target->_type != NT_VDPOL &&
		      y->_target->_type != NT_COLPITTS) ) {
			cerr << "A CSynapseOscilTuner connecting to a non-oscillator unit ("
			     << y->_target->unit_class() << ")\n";
			return -1;
		}

		if ( src_found && tgt_found ) {
			syn_list.push_back( y);

		      // assign idx
			y->idx = _var_cnt;
			y->_parent_model = this;

			_var_cnt += y->v_no();
			_unit_cnt++;

			if ( is_last ) {
				finalize_additions();
				y->reset_state();
			}

			return 0;
		}
	}

	if ( verbosely > 0 )
		cerr << "Serious warning: For synapse " << y->label << ", " << (!src_found ?"source" :"target") << " not found in the model\n";

	return -1;
}


void
CModel::finalize_additions()
{
	V->resize( _var_cnt), W->resize( _var_cnt);

	list<CNeuron*>::iterator N;
	for_all_neurons (N)
		(*N) -> reset_state();
	list<CSynapse*>::iterator Y;
	for_all_synapses (Y)
		(*Y) -> reset_state();

	if ( _status & CN_MDL_SORTUNITS ) {
		neu_list.sort( __CUnitCompareObject);
		syn_list.sort( __CUnitCompareObject);
	}

	if ( verbosely > 2 )
		cout << "fixate: " << _var_cnt << " variables\n";

	_integrator->prepare();
}









void
CModel::reset_all_units()
{
	list<CNeuron*>::iterator N;
	for_all_neurons (N)
		(*N) -> reset_state();
	list<CSynapse*>::iterator Y;
	for_all_synapses (Y)
		(*Y) -> reset_state();
}


// EOF
