// ;-*-C++-*- *  Time-stamp: "2008-10-16 03:40:57 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *         building on original work by Thomas Nowotny (tnowotny@ucsd.edu)
 *
 * License: GPLv2 only
 *
 * Initial version: 2008-08-02
 *
 * Base class for CNeuron and CSynapse, the two base classes for all
 * the individual units
 */


#ifndef CN_UNIT_H
#define CN_UNIT_H


#include <vector>
#include <list>
#include <iterator>
#include <fstream>
#include <iostream>
#include <cstring>
#include <cmath>

#include "cnlib-types.h"
#include "cnlib-unit.h"

using namespace std;



// forward decls
class CSynapse;
class CNeuron;
class CModel;




#define CN_UENABLED		(1 << 0)
#define CN_ULISTENING		(1 << 1)
#define CN_UPRIVATEPARAMS	(1 << 2)

#define CN_NCONSCIOUS		(1 << 3)  // capable of firing spontaneously
#define CN_RECORDINGSPIKES	(1 << 4)
#define CN_NFIRING		(1 << 5)  // firing now
#define CN_NREFRACT		(1 << 6)  // in refractory phase now

#define CN_SMPLTAPE_LOOP	(1 << 7)


#define CN_LABEL_SIZE 40



class CUnit {
    friend class CModel;

    protected:
	TUnitType	_type;  // will look up p, pno and vno from __CN_DescriptorTable using _type as index
	int		_status;
    public:
	const char     *unit_class()		{ return __CN_DescriptorTable[_type].name; }
	int		status()		{ return _status;				}
	char		label[CN_LABEL_SIZE];

	TUnitType	type()			{ return _type;					}

	bool		is_enabled()		{ return _status & CN_UENABLED;			}
	void enable( bool indeed = true)	{ if ( indeed )  _status |=  CN_UENABLED;
						  else           _status &= ~CN_UENABLED;	}
    protected:
      // parent model
	CModel	*_parent_model;

      // variables for units in the model are catenated on a single
      // vector<double>, as an essential optimization measure; each
      // unit knows its own set of variables by this idx:
	size_t idx;
      // the containing model provides idx on registering our unit
    public:
	const double& model_time();

//    protected:
//     // a private copy of params; otherwise, points to the global table[_type]
//	double* P;

    public:
      // number of vars
	size_t v_no()	{ return __CN_DescriptorTable[_type].vno; }
	size_t p_no()	{ return __CN_DescriptorTable[_type].pno; }

      // parameter access
	virtual double &param_value( size_t i) = 0;
	const char *const &param_name( size_t i)	{ return __CN_DescriptorTable[_type].stock_param_names[i]; }
	const char *const &param_sym( size_t i)		{ return __CN_DescriptorTable[_type].stock_param_syms[i]; }

      // this reverts all params to initial stock values
	virtual void reset_params() = 0;

      // this reverts all of V to initial stock values
	void reset_state(); // defined in cn-model.h

      // constructors, destructor
	CUnit( TUnitType intype, const char *inlabel)
		{
			_type  = intype;			_status = 0 | CN_UENABLED;
//			P = new double[p_no()];			reset_params();
			memset( label, 0, CN_LABEL_SIZE);	strncpy( label, inlabel, CN_LABEL_SIZE);
			_listener = NULL;
		}
	CUnit( TUnitType intype)
		{
			_type  = intype;			_status = 0 | CN_UENABLED;
//			P = new double[p_no()];			reset_params();
			memset( label, 0, CN_LABEL_SIZE);	snprintf( label, CN_LABEL_SIZE-1, "fafa%p", this);
			_listener = NULL;
		}
       ~CUnit()
		{	/* delete[] P;	*/ }

      // each derived neuron has its own code for these
      // it is only called by the container model
	virtual void derivative( vector<double>&, vector<double>&) = 0;

	virtual void noise() = 0;

    private:
      // where V is piped on each time step
	ofstream *_listener;
    public:
	bool is_listening()		{ return _status & CN_ULISTENING; }
	void start_listening();
	void stop_listening();
	void pause_listening()		{ _status &= ~CN_ULISTENING; }
	void resume_listening()		{ _status |=  CN_ULISTENING; }

	void tell();
};


extern double __cn_dummy_double;


struct CUnitCompare {
	bool operator() (CUnit *l, CUnit *r)	{ return strcmp( l->label, r->label) < 0; }
};

extern CUnitCompare __CUnitCompareObject;







#define CN_MIN_DISTANCE .1

struct SCoord {
	double x, y, z;

      // distance
	double operator - ( SCoord& p)
		{
			return sqrt( pow(x - p.x, 2) + pow(y - p.y, 2) + pow(z - p.z, 2));
		}

	bool operator == ( SCoord& p)
		{
			return x == p.x && y == p.y && z == p.z;
		}
	bool operator != ( SCoord& p)
		{
			return x != p.x || y != p.y || z != p.z;
		}
	bool too_close( SCoord& p, double mindist = CN_MIN_DISTANCE)
		{
			return	fabs(x - p.x) < mindist &&
				fabs(y - p.y) < mindist &&
				fabs(z - p.z) < mindist;
		}
};







class CNeuron
  : public CUnit {

    friend class CModel;

    public:
	SCoord	pos;

      // list of all incoming synapses
	vector<CSynapse*> den;

      // ctor and dtor
	CNeuron( TUnitType intype, const char *inlabel,
		  double inx, double iny, double inz)
	      : CUnit( intype, inlabel)
		{
			pos.x = inx, pos.y = iny, pos.z = inz;
			spike_cnt = 0;
			t_last_spike_start = t_last_spike_end =
				t_firing_started = t_firing_ended = -INFINITY;
//			cerr << " + [" << intype << "]" << inlabel << " @(" << inx << "," << iny << "," << inz << ")\n";
		}

       ~CNeuron();

	void reset_state()
		{
			CUnit::reset_state();
			spike_cnt = 0;
		}

	double&  E(); // (defined in cnlib-model.h)
	double&  E( vector<double>& base)	{ return base[idx+0]; }
	double& dE( vector<double>& base)	{ return base[idx+0]; }
      // others are specific to, and defined in, derived classes

      // the total current on incoming synapses
	double Isyn( vector<double>&);
	double Isyn(); // to be called outside integration routines

      // keeping track of
	double	t_last_spike_start,
		t_last_spike_end;
	size_t	spike_cnt;

	bool is_spiking();
	void set_spike_recording( bool indeed = true)
		{ if ( indeed ) _status |=  CN_RECORDINGSPIKES;
			else	_status &= ~CN_RECORDINGSPIKES; }

	void spike_detect();

      // for conscious neurons:
	double	t_firing_started,
		t_firing_ended;

	bool is_conscious()	{ return _status & CN_NCONSCIOUS; }

	virtual void possibly_fire() = 0;

};










class CSynapse
  : public CUnit {

    friend class CModel;
    friend class CNeuron;

    protected:
	CNeuron	*_source,
		*_target;

	double t_last_release_started;

    public:
	CNeuron	*source()	{  return _source;  }
	CNeuron	*target()	{  return _target;  }

      // constructors
	CSynapse( TUnitType intype, const char *inlabel,
		  CNeuron *insource, CNeuron *intarget)
	      : CUnit( intype, inlabel)
		{
			_source = insource, _target = intarget;
			_target->den.push_back( this);
		}

       ~CSynapse();

	double length()
		{	return _source->pos - _target->pos;	}

	virtual double Isyn( vector<double>&) = 0;

	virtual void derivative( vector<double>&, vector<double>&) = 0;

//	virtual void noise( double*, double*)
//		{}
};


#endif

// EOF
